/* equations.h */

/*****************************************************************
        Copyright by Rockefeller University,
can not be reproduced or distributed without written permission of
the copyright holder.  Version 1.1 of April 2006.

Written by Alexandre Morozov (contact person: morozov@edsb.rockefeller.edu),
and Eric Siggia.

The program DNABEND implements an algorithm for computing elastic energies
for DNA molecules of arbitrary length and sequence. DNA can be constrained
to follow an arbitrary 3D curve, such as a left-handed superhelix found
in nucleosomal core particles.

The DNABEND method and its applications are described in the following
paper:

******************************************************************/

#ifndef _equations_h_
#define _equations_h_

#include <gsl/gsl_eigen.h>
#include <gsl/gsl_linalg.h>
#include "xyzframe.h"

//#define PI 3.1415926535897323
//#define DegtoRad(angle) (angle*PI/180)
//#define RadtoDeg(angle) (angle*180/PI)

class Equations {
 private:
  gsl_matrix *A; // Ax = b
  gsl_vector *b;
  gsl_vector *x;
  gsl_vector *curv; // curvature value at each basestep
  gsl_matrix **derivs; // derivatives of xyz coords in global frame
  unsigned int Lseq; // number of dinucs in seq (ie seq length in **dinucs**)
  static const unsigned int Nprm = 6; // number of WO prms: [ Twist    Roll    Tilt    Shift    Slide    Rise ]
  unsigned int DIM; // = Lseq * Nprm
  static const double F0 = 1.0; // WO potential scaling factor: E_{bs}
  static const double F1 = 0.0; // constraint scaling factor: E_1 roll
  static const double F2 = 0.0; // constraint scaling factor: E_1 tilt
  static const double F3 = 0.0; // constraint scaling factor: E_2 twist
  static const double F4 = 100.0; // constraint scaling factor: E_3 "native" BStable
  static const double F5 = 0.0; // constraint scaling factor: E_4 global xyz
  static const double F6 = 1.0; // constraint scaling factor: E_5 global displacements
  static const double Omega_0 = 34.69599; // mean value around which twist is expanded (\Omega_{0})
  static const double Roll_0 = 2.472812; // same for roll & tilt
  static const double Tilt_0 = -0.1554501;
  static const double Phi_0 = 20.0; // arbitrary initial phase for roll & tilt periodic variation
  static const double kappa_roll = 4.53; // components of the ideal curvature 2D vector (roll_id tilt_id)
  static const double kappa_tilt = 4.53; // NOTE: all angles are in Deg in Ax = b !!
  //static const double kappa_roll = 0.0; // components of the ideal curvature 2D vector (roll_id tilt_id)
  //static const double kappa_tilt = 0.0;
  void setWOcoeff(Sequence::Sequence *seq, DNAparams::DNAparams &prms); // setting up Ax=b in steps
  void setCurvConstr(void);
  void setTwistConstr(void);
  void setBStableConstr(gsl_matrix *bst);
  void setGlobalConstr(gsl_matrix *bst);
  void setDisplacementConstr(std::vector<coord> &frames_ideal);
  void setWOcoeff_global(Sequence::Sequence *seq, DNAparams::DNAparams &prms,
			 Geom::Geom &ideal, Geom::Geom &init);// global version
  void ReadDerivs(const std::string filename);
  void GetXYZMatrix(gsl_matrix *bst, gsl_matrix *mat, unsigned int ind);
 public:
  Equations();
  Equations(Sequence::Sequence *seq, DNAparams::DNAparams &prms); // set up A and b given: seq and WO prms
  // Set up A and b given: seq, WO prms & BStable (as gsl_matrix bc of inheritance conflicts)
  Equations(Sequence::Sequence *seq, DNAparams::DNAparams &prms, gsl_matrix *bst, const std::string mode);
  Equations(Sequence::Sequence *seq, DNAparams::DNAparams &prms, Geom::Geom &ideal, Geom::Geom &init);
  ~Equations();
  void Solve(void); // solve Ax = b using LU 
  void SolveSVD(void); // solve Ax = b using SVD 
  void ReconstructWO(coord &init_frame); // return from global displacements to local WO prms
  void Eigen(gsl_matrix *mat); // find eigenvalues for REAL SYMMETRIC A
  void Eigen(void);
  void PrintResult(void); // print x
  void GetCurv(void); // find curvature at each basestep
  void PrintCurv(void); // print curvature at each basestep
  int length(void); // length of x
  double getx(int ind); // x[ind]
  void setx(int ind, double value); // x[ind] = value
  gsl_vector *result(void); // x
  double Energy(Sequence::Sequence *seq, DNAparams::DNAparams &prms);
  double Energy_global(Sequence::Sequence *seq, DNAparams::DNAparams &prms,
		       Geom::Geom &ideal, Geom::Geom &WO, Geom::Geom &init);
};

#endif

