/* geometry.h */

/*****************************************************************
        Copyright by Rockefeller University,
can not be reproduced or distributed without written permission of
the copyright holder.  Version 1.1 of April 2006.

Written by Alexandre Morozov (contact person: morozov@edsb.rockefeller.edu),
and Eric Siggia.

The program DNABEND implements an algorithm for computing elastic energies
for DNA molecules of arbitrary length and sequence. DNA can be constrained
to follow an arbitrary 3D curve, such as a left-handed superhelix found
in nucleosomal core particles.

The DNABEND method and its applications are described in the following
paper:

******************************************************************/

#ifndef geometry_h_
#define geometry_h_

#include <math.h>
#include <stdlib.h>
#include <iostream>
#include <assert.h>
//#include "molecule.h"
#include "vector.h"

#define PI 3.1415926535897323
#define EPS 0.00000001

#define Dot(a,b) (a.x*b.x+a.y*b.y+a.z*b.z)
#define Multiply(v,c) {v.x*=c;v.y*=c;v.z*=c;}
#define Subtract(a,b) {a.x-=b.x;a.y-=b.y;a.z-=b.z;}
#define Add(a,b) {a.x+=b.x;a.y+=b.y;a.z+=b.z;}
#define SqrLength(v) (Dot(v,v))

#define Flip(minusv,v) {minusv.x=-v.x;minusv.y=-v.y;minusv.z=-v.z;}
#define DegtoRad(angle) (angle*PI/180)
#define RadtoDeg(angle) (angle*180/PI)

//#define TETRAD_BY_INDEX

typedef struct Polar_{
  double phi,theta;
} Polar_;

typedef struct euler{
  double phi,theta,gamma;
} euler;

typedef struct tetrad {
#ifdef TETRAD_BY_INDEX
  int ia,ja,ib,jb,ic,jc;
#else
  Vector *a,*b,*c; // 3 known tetrad atoms
#endif
  double chi; // dihedral angle
  double ang_1,ang_3; // 2 bond angles
  double l1,l2,l3; // 3 bond lengths
} tetrad;

double Dihedral(Vector x1, Vector x2, Vector x3, Vector x4);
double Theta2d(Vector a, Vector b, Vector R);
Polar_ CartesianToPolar(Vector b);
Vector RotatePolar(Vector a,Polar_ ang);
Vector InversePolar(Vector a,Polar_ ang);
Vector RotateZ(Vector a,double phi);
Vector RotateY(Vector a,double the);
Vector RotateX(Vector a,double psi);
coord RotateMat(coord const & crd, double a1, double a2, double a3);
double Quadr(double a1,double a2,double a3,int hand_flag);
Vector AtomByLAA(struct tetrad tetr);
Vector AtomByLAD(struct tetrad tetr);
inline double Bangle(Vector a,Vector b,Vector c);
inline double Blength(const Vector &a,const Vector &b);
double SqrDist(Vector a, Vector b);
int WithinSqrDist(Vector a, Vector b, double dist);
Vector ApplyRotation(Vector v, Vector R, double theta);
tetrad AssignLAA(Vector *a, Vector *b, Vector *c, Vector *d);
tetrad AssignLAD(Vector *a, Vector *b, Vector *c, Vector *d);

//This function calculates bond length. It returns 0 length rather than terminates.
inline double Blength(const Vector &a,const Vector &b){
  return sqrt((a.x - b.x)*(a.x - b.x) +
	      (a.y - b.y)*(a.y - b.y) +
	      (a.z - b.z)*(a.z - b.z));
}

//This function calculates *cos* of bond angle.
//It terminates if the angle is undefined (one of the vectors is 0 length)
inline double Bangle(Vector a,Vector b,Vector c){
  double tmp;
  a -= b;
  c -= b;
  assert(a.Length()!=0);
  assert(c.Length()!=0);
  tmp=(a * c)/(a.Length()*c.Length());
  //debug
  //printf("Dot:%lf\n",Dot(a,c));
  //printf("tmp: %lf\n",tmp);
  return tmp;
}

#endif


