/* minimize.h */

/*****************************************************************
        Copyright by Rockefeller University,
can not be reproduced or distributed without written permission of
the copyright holder.  Version 1.1 of April 2006.

Written by Alexandre Morozov (contact person: morozov@edsb.rockefeller.edu),
and Eric Siggia.

The program DNABEND implements an algorithm for computing elastic energies
for DNA molecules of arbitrary length and sequence. DNA can be constrained
to follow an arbitrary 3D curve, such as a left-handed superhelix found
in nucleosomal core particles.

The DNABEND method and its applications are described in the following
paper:

******************************************************************/

#ifndef _minimize_h_
#define _minimize_h_

#include <gsl/gsl_eigen.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_multimin.h>
#include "xyzframe.h"
#include "options.h"

typedef struct M_params {
  class Minimizer *M;
  class DNAparams *DNA;
} M_params;

class Minimizer {
 private:
  unsigned int Lseq; // number of dinucs in seq (ie seq length in **dinucs**)
  static const unsigned int Nprm = 6; // number of WO prms: [ Twist    Roll    Tilt    Shift    Slide    Rise ]
  unsigned int DIM; // number of Degrees of Freedom = Lseq * Nprm
  gsl_matrix *A; // Ax = b
  gsl_vector *b;
  gsl_vector *x; // degrees of freedom in the objective function
  gsl_vector *x0; // initial degrees of freedom in the objective function (ie input guess)
  int *ind; // vector of indices corresponding to dinuc types (compute them once from seq & store)
  Vector *disp0; // a set of ideal superhelical radius-vectors in the global frame
  gsl_matrix *F_SH; // Superhelix-induced force constants in the global frame
  gsl_matrix *F_WO; // WO force constants transformed into the global frame
  gsl_vector *dispWO; // WO equil. prms transformed into the global frame
  //Geom *xyz_ideal; // a set of ideal superhelical frames
  //Geom *xyz_wt; // a set of native (pdb) superhelical frames
  double Ewo; // elastic (WO) energy
  gsl_vector *Ewo_vec; // vector of elastic (WO) energies for each dinuc  
  double rms; // RMS between final and ideal configurations
  double Etotal; // total weighted energy
  double F0; // WO scaling factor: E_{bs} WO energy
  double F3; // sums of angles constraint scaling factor
  double F6; // constraint scaling factor: E_5 global displacements
  double F7; // constraint scaling factor: E_6 deviations from the ideal set of frames
  static const double Omega_0 = 34.69599; // mean value around which twist is expanded (\Omega_{0})
  static const double Roll_0 = 2.472812; // same for roll & tilt
  static const double Tilt_0 = -0.1554501;
  void ImportTails(Geom::Geom &xyz_ideal, Geom::Geom &xyz_wt, unsigned int start, unsigned int end);
  void LocalizeWO(Geom::Geom &wt, DNAparams::DNAparams &prms);
  void GlobalizeWO(Geom::Geom &ideal, Geom::Geom &wt, DNAparams::DNAparams &prms);
  void setWOcoeff(void);
  void setWOcoeff(gsl_vector *bst_ideal);
  void setWOcoeff(Vector *diff0, gsl_vector *bst_ideal);
  void setDisplacementConstr(void);
  void setIdealConstr(gsl_vector *bst_ideal);
  void setSumConstr(void);
  void UnshiftX(Vector *diff0, gsl_vector *bst_ideal);
 public:
  friend double Etot(const gsl_vector *v, void *params);
  friend void Etot_df(const gsl_vector *v, void *params, gsl_vector *df);
  friend void Etot_fdf (const gsl_vector *v, void *params, double *f, gsl_vector *df);
  Minimizer();
  Minimizer(Sequence::Sequence *seq, DNAparams::DNAparams &prms, Geom::Geom &xyz_ideal, Geom::Geom &xyz_wt,
	    struct Options *opt, std::string const &pdbout, std::string const &geomout, std::string const &runmode);
  ~Minimizer();
  double getx(int ind); // x[ind]
  void setx(int ind, double value); // x[ind] = value
  gsl_vector *getresult(void); // get x
  void setresult(const gsl_vector *v); // set x to v
  double E_WO(DNAparams::DNAparams &prms);
  double E_WO(DNAparams::DNAparams &prms, const double weight);
  double E_WO(void);
  //double E_constr_glob(void);
  double E_constr(void);
  void E_WO_df(gsl_vector *g);
  void E_constr_df(gsl_vector *g);
  int Run(void); // simplex
  int RunG(void); // conjugate gradient
  void Solve(void); // solve Ax = b
  void ReconstructWO(Geom::Geom &xyz_ideal); // reconstruct original WO displacements using ideal frames
  void PDB(const std::string filename, Vector orig); // write out a PDB file using global displacements
  double getEwo(void);
  double getEtotal(void);
  double getRMS(void);
};

#endif

