/* xyzframe.h */

/*****************************************************************
        Copyright by Rockefeller University,
can not be reproduced or distributed without written permission of
the copyright holder.  Version 1.1 of April 2006.

Written by Alexandre Morozov (contact person: morozov@edsb.rockefeller.edu),
and Eric Siggia.

The program DNABEND implements an algorithm for computing elastic energies
for DNA molecules of arbitrary length and sequence. DNA can be constrained
to follow an arbitrary 3D curve, such as a left-handed superhelix found
in nucleosomal core particles.

The DNABEND method and its applications are described in the following
paper:

******************************************************************/

#ifndef _xyzframe_h_
#define _xyzframe_h_

#include <gsl/gsl_deriv.h>
#include "bstable.h"

/*
// Frame struc:
typedef struct coord {
  Vector a,b,c; // x,y,z unit vectors
  Vector orig; // origin of the coord system
} coord;
*/

// Struc for computing derivs:
typedef struct Frame_params {
  unsigned int ind; // xyz frame index
  unsigned int crd; // 0,1,2 for x,y,z; other values not allowed
  unsigned int c1,c2; // coords in the table of WO prms
  class Geom *GM;
} Frame_params;

class Geom {
 private:
  void ReconstructGeom(coord & crd);
  void ReconstructGeom(coord & crd, unsigned int Find);
  gsl_matrix *ConnectivityCoeff(void);
  coord InitFirstFrame();
  void ReadFrames(std::string const & filename, std::vector<coord> &frm);
  Matrix3 T_deriv(double Omega, double r, double t, int i);
  Matrix3 Tmst_deriv(double Omega, double r, double t, int i);
  Matrix3 T(double Omega, double r, double t);
  Matrix3 Tmst(double Omega, double r, double t);
  gsl_matrix *m; // local copy of input WO prms used in xyz reconstruction
  coord crd_init; // first frame from which all others are built
  Matrix3 **derivs; // Lseq x 3*Lseq table of 3x3 matrix derivatives wrt [ Twist    Roll    Tilt ]
  gsl_matrix *F_curv; // Force constants (ie curvatures) induced in the **quadratic potential**
                      // by expanding displacements to 1st order in the global frame
  unsigned int Lseq; // number of sets of input WO prms (each set has 6 prms)
  static const unsigned int Nprm = 6; // number of WO prms: [ Twist    Roll    Tilt    Shift    Slide    Rise ]
 public:
  friend double GetFrame(double x, void *params);
  std::vector<coord> frames; // basepair frames: origin & 3 unit vecs
  std::vector<coord> frames_mst; // MST frames: origin & 3 unit vecs
  Geom(BStable::BStable & bst);
  Geom(BStable::BStable &bst, coord crd);
  Geom(gsl_matrix *bst_mat, coord crd);
  Geom(const gsl_vector *bst_vec, coord crd);
  Geom(std::string const & fbp, std::string const & fmst);
  ~Geom();
  void Write(const std::string mode);
  void PDB(const std::string filename, const char mode = 's');
  void PDBunitvecs(const std::string filename, const char mode = 's');
  double RMSD(Geom::Geom &); // compute RMSD with another set of frames (ie from pdb_seq)
  unsigned int length(); // vector length
  void BuildDerivs();
  Matrix3 NumDeriv(unsigned int bs_ind, unsigned int WO_ind, unsigned int angle_ind, double diff);
  Matrix3 GetDeriv(unsigned int bs_ind, unsigned int WO_ind, unsigned int angle_ind);
  void ComputeCurv();
  void WriteCurv(const std::string filename);
  void ReadCurv(const std::string filename);
  gsl_matrix *GetCurv();
  void WriteConnectivityCoeff(const std::string filename);
  //void Derivs(); // compute derivatives for all bp frames wrt all WO prms
  //void WriteDerivs(const std::string filename); // write computed derivs
  void CompareFrames(Geom::Geom &, const std::string mode); // compare how close 2 sets of bp+mst frames are to each other
  void Transform(Geom::Geom &gm);
  void Transform(Geom::Geom &gm, unsigned int begin, unsigned int end);
  double getm(unsigned int i1, unsigned int i2);
  void TranslateGeom(void);
  gsl_vector *GetGeom(void);
  Vector *disp(void);
  Vector *global_disp(void);
};

#endif

