#!/usr/bin/perl -w

use strict;

# This aux. script is used to output an ideal helix in PDB format.
# Both basepair & MST frames are output. Note that the chainID is I
# for the basepair, and J for MST. Also, the origin of frame MST_i
# is NOT shifted relative to the frame bp_i: origins of MST frames are
# not used anywhere afterwards.

my $PI = 3.1415926535897323;

my $P = 25.9; # pitch [A]
my $R = 41.9; # radius [A]
my $alpha = 1.84; # number of superhelical turns
my $B = $alpha * $P; # length of the helix [A]
my $N = 147; # number of basepairs
my $Reff = sqrt($R**2 + ($P/(2*$PI))**2);
my $Omega_global = $PI/180.0 * 34.69599;

# argv
my %opts = &getCommandLineOptions();
my $WOtable = $opts{t}; # table with (native) WO prms
my @WOprms = (); # (N-1) x 6 table of WO prms: [ twist   roll   tilt   shift   slide   rise ]
ReadWOTable($WOtable,\@WOprms) if (defined $WOtable);

my $out = "ideal_helix.pdb";
open(I,">$out") or die "Can't open file: $!\n";

my $out_mst = "ideal_helix_mst.pdb";
open(II,">$out_mst") or die "Can't open file: $!\n";

#for my $i (0 .. $N-1) {
#  my $ti = $B*$i/($N-1);
#  my $xi = $R*cos(2*$PI*$ti/$P);
#  my $yi = $R*sin(2*$PI*$ti/$P);
#  my $zi = -$ti;
#  printf I "ATOM   %4d %4s %3s %1s %3d    %8.3f%8.3f%8.3f  1.00  0.00\n",
#    $i,"O  ","ALA","I",$i,$xi,$yi,$zi;
#}
#print I "TER\n";
#close I;

my $atom_cnt = 1;
my $atom_cnt_mst = 1;
my $cumul_twist = 0.0;
for my $i (0 .. $N-1) {
  my $Omega;
  if (defined $opts{realtwist}) {
    $Omega = $PI/180.0 * $WOprms[$i][0] if ($i < $N-1); # redefine Omega to mask the global value
  }else{
    $Omega = $Omega_global;
  }
  my $ti = $B*$i/($N-1);
  my $si = (2*$PI*$Reff)/$P * $ti;
  my ($rx_i,$ry_i,$rz_i,$tx_i,$ty_i,$tz_i,$nx_i,$ny_i,$nz_i,$bx_i,$by_i,$bz_i) = frame($si);
  my $ti2 = $B*($i+0.5)/($N-1);
  my $si2 = (2*$PI*$Reff)/$P * $ti2;
  # NOTE: ($rx_mst,$ry_mst,$rz_mst) are NOT used
  my ($rx_mst,$ry_mst,$rz_mst,$tx_mst,$ty_mst,$tz_mst,$nx_mst,$ny_mst,$nz_mst,$bx_mst,$by_mst,$bz_mst) = frame($si2);
  # Rotate n and b of frame i to follow the helical twist:
  ($nx_i,$ny_i,$nz_i) = Rotate($nx_i,$ny_i,$nz_i,$tx_i,$ty_i,$tz_i,$cumul_twist);
  ($bx_i,$by_i,$bz_i) = Rotate($bx_i,$by_i,$bz_i,$tx_i,$ty_i,$tz_i,$cumul_twist);
  # Rotate n and b of MST frame to follow the helical twist:
  ($nx_mst,$ny_mst,$nz_mst) = Rotate($nx_mst,$ny_mst,$nz_mst,$tx_mst,$ty_mst,$tz_mst,($cumul_twist+0.5*$Omega) );
  ($bx_mst,$by_mst,$bz_mst) = Rotate($bx_mst,$by_mst,$bz_mst,$tx_mst,$ty_mst,$tz_mst,($cumul_twist+0.5*$Omega) );
  $cumul_twist += $Omega;
  if (defined $opts{longPDB}) {
    printf I "ATOM   %4d %4s %3s %1s %3d    %10.5f%10.5f%10.5f  1.00  0.00\n",
      $atom_cnt++,"O  ","ALA","I",$i+1,$rx_i,$ry_i,$rz_i;
    printf I "ATOM   %4d %4s %3s %1s %3d    %10.5f%10.5f%10.5f  1.00  0.00\n",
      $atom_cnt++,"N1 ","ALA","I",$i+1,$rx_i+$nx_i,$ry_i+$ny_i,$rz_i+$nz_i;
    printf I "ATOM   %4d %4s %3s %1s %3d    %10.5f%10.5f%10.5f  1.00  0.00\n",
      $atom_cnt++,"N2 ","ALA","I",$i+1,$rx_i+$bx_i,$ry_i+$by_i,$rz_i+$bz_i;
    printf I "ATOM   %4d %4s %3s %1s %3d    %10.5f%10.5f%10.5f  1.00  0.00\n",
      $atom_cnt++,"N3 ","ALA","I",$i+1,$rx_i+$tx_i,$ry_i+$ty_i,$rz_i+$tz_i;
    if ($i < $N-1) { # skip the last frame for which MST is not defined
      printf II "ATOM   %4d %4s %3s %1s %3d    %10.5f%10.5f%10.5f  1.00  0.00\n",
	$atom_cnt_mst++,"O  ","ALA","J",$i+1,$rx_i,$ry_i,$rz_i;
      printf II "ATOM   %4d %4s %3s %1s %3d    %10.5f%10.5f%10.5f  1.00  0.00\n",
	$atom_cnt_mst++,"N1 ","ALA","J",$i+1,$rx_i+$nx_mst,$ry_i+$ny_mst,$rz_i+$nz_mst;
      printf II "ATOM   %4d %4s %3s %1s %3d    %10.5f%10.5f%10.5f  1.00  0.00\n",
	$atom_cnt_mst++,"N2 ","ALA","J",$i+1,$rx_i+$bx_mst,$ry_i+$by_mst,$rz_i+$bz_mst;
      printf II "ATOM   %4d %4s %3s %1s %3d    %10.5f%10.5f%10.5f  1.00  0.00\n",
	$atom_cnt_mst++,"N3 ","ALA","J",$i+1,$rx_i+$tx_mst,$ry_i+$ty_mst,$rz_i+$tz_mst;
    }
  }else{
    printf I "ATOM   %4d %4s %3s %1s %3d    %8.3f%8.3f%8.3f  1.00  0.00\n",
      $atom_cnt++,"O  ","ALA","I",$i+1,$rx_i,$ry_i,$rz_i;
    printf I "ATOM   %4d %4s %3s %1s %3d    %8.3f%8.3f%8.3f  1.00  0.00\n",
      $atom_cnt++,"N1 ","ALA","I",$i+1,$rx_i+$nx_i,$ry_i+$ny_i,$rz_i+$nz_i;
    printf I "ATOM   %4d %4s %3s %1s %3d    %8.3f%8.3f%8.3f  1.00  0.00\n",
      $atom_cnt++,"N2 ","ALA","I",$i+1,$rx_i+$bx_i,$ry_i+$by_i,$rz_i+$bz_i;
    printf I "ATOM   %4d %4s %3s %1s %3d    %8.3f%8.3f%8.3f  1.00  0.00\n",
      $atom_cnt++,"N3 ","ALA","I",$i+1,$rx_i+$tx_i,$ry_i+$ty_i,$rz_i+$tz_i;
    if ($i < $N-1) { # skip the last frame for which MST is not defined
      printf II "ATOM   %4d %4s %3s %1s %3d    %8.3f%8.3f%8.3f  1.00  0.00\n",
	$atom_cnt_mst++,"O  ","ALA","J",$i+1,$rx_i,$ry_i,$rz_i;
      printf II "ATOM   %4d %4s %3s %1s %3d    %8.3f%8.3f%8.3f  1.00  0.00\n",
	$atom_cnt_mst++,"N1 ","ALA","J",$i+1,$rx_i+$nx_mst,$ry_i+$ny_mst,$rz_i+$nz_mst;
      printf II "ATOM   %4d %4s %3s %1s %3d    %8.3f%8.3f%8.3f  1.00  0.00\n",
	$atom_cnt_mst++,"N2 ","ALA","J",$i+1,$rx_i+$bx_mst,$ry_i+$by_mst,$rz_i+$bz_mst;
      printf II "ATOM   %4d %4s %3s %1s %3d    %8.3f%8.3f%8.3f  1.00  0.00\n",
	$atom_cnt_mst++,"N3 ","ALA","J",$i+1,$rx_i+$tx_mst,$ry_i+$ty_mst,$rz_i+$tz_mst;
    }
  }
}
print I "TER\n";
print II "TER\n";
close I;
close II;

#################
## Functions
#################

sub getCommandLineOptions {
  use Getopt::Long;
  my $usage = qq{usage: $0 <-help [-t WO.table -realtwist] -longPDB>\n};
  # Get args
  my %opts = ();
  &GetOptions(\%opts, "t=s", "help", "realtwist", "longPDB");
  if (defined $opts{help}) {
    print STDERR "$usage\n";
    exit -1;
  }
  if (defined $opts{realtwist} && !defined $opts{t}) {
    print STDERR "ERROR: -realtwist option requires a table specified through -t option!\n";
    exit -1;
  }
  &checkExist('f', $opts{t}) if (defined $opts{t});
  return %opts;
}

# checkExist()
#
sub checkExist {
  my ($type, $path) = @_;
  if ($type eq 'd') {
    if (! -d $path) {
      warn("dir not found: $path\n");
      exit -3;
    }
  }
  elsif ($type eq 'f') {
    if (! -f $path) {
      warn("file not found: $path\n");
      exit -3;
    }
    elsif (! -s $path) {
      warn("empty file: $path\n");
      exit -3;
    }
  }
}

# Reads a table of WO prms.
sub ReadWOTable {
  my ($table,$WOprms) = @_;
  #Check refs:
  if (ref($WOprms) ne 'ARRAY') {
    die "Expected array ref(s) in ReadWOTable!\nError in function args!\n";
  }
  open(II,"$table") or die "Can't open file: $!\n";
  my @lines = (<II>);
  close II;
  for my $i (1 .. $#lines) { # skip header
    chomp $lines[$i];
    my @tt = split ' ',$lines[$i];
    next if (@tt!=8);
    push @$WOprms, [ @tt[2 .. 7] ];
  }
  #debug
  #print "Lseq = ",$#$WOprms+1,"\n";
  # Check table length:
  die "Length of WO table is unequal to the expected number of basesteps!\n"
    unless (($N-1)==($#$WOprms+1) );
}

# Cross product": b x v
sub cross {
  my ($bx,$by,$bz,$vx,$vy,$vz) = @_;
  
  my $tx = $by*$vz-$bz*$vy;
  my $ty = $bz*$vx-$bx*$vz;
  my $tz = $bx*$vy-$by*$vx;
  
  return ($tx,$ty,$tz);
}

# In-plane rotation of a 3D vector around the z-axis:
sub RotateZ {
  my ($ax,$ay,$az,$phi) = @_;
  
  my $bx = $ax*cos($phi)+$ay*sin($phi);
  my $by = -$ax*sin($phi)+$ay*cos($phi);
  my $bz = $az;
  
  return ($bx,$by,$bz);
}

# This function rotates the Vector v around the axis R by theta radians.
# NOTE: Vector R must be normalized!!
sub Rotate {
  my ($tempx,$tempy,$tempz,$Rx,$Ry,$Rz,$theta) = @_;
  my ($posx,$posy,$posz,$A1x,$A1y,$A1z,$A2x,$A2y,$A2z,$A3x,$A3y,$A3z);
  my ($cost,$sint,$d,$Rx2,$Ry2,$Rz2);
  
  $cost=cos($theta);
  $sint=-sin($theta);
  #  This matrix was originally computed by Ed Thayer, and double-checked
  #  by me.  It is the product of a basis transformation into a basis in which
  #  R is one of the coordinate axes, a simple rotation around R, and
  #  a basis transformation back again.
  $Rx2=$Rx*$Rx;  $Ry2=$Ry*$Ry;  $Rz2=$Rz*$Rz;
  $d=$Rx2+$Ry2;
  if ($Rx2 + $Ry2 > 0) {
    $A1x=$Rx2+($Ry2+$Rx2*$Rz2)*$cost/$d;
    $A1y=$Rx*$Ry*(1-$cost)+$Rz*$sint;
    $A1z=$Rx*$Rz*(1-$cost)-$Ry*$sint;
    $A2x=$Rx*$Ry*(1-$cost)-$Rz*$sint;
    $A2y=$Ry2+($Rx2+$Ry2*$Rz2)*$cost/$d;
    $A2z=$Ry*$Rz*(1-$cost)+$Rx*$sint;
    $A3x=$Rx*$Rz*(1-$cost)+$Ry*$sint;
    $A3y=$Ry*$Rz*(1-$cost)-$Rx*$sint;
    $A3z=$Rz2+$d*$cost;
  }else{
    $A1x = $cost     ; $A1y = $Rz*$sint; $A1z = 0;
    $A2x = -$Rz*$sint; $A2y = $cost    ; $A2z = 0;
    $A3x =         0; $A3y =        0; $A3z = $Rz2;
  }
  #  This is just an obscure way of writing the product of matrix A
  #  with Vector v.
  my $x = $A1x * $tempx + $A1y * $tempy + $A1z * $tempz;
  my $y = $A2x * $tempx + $A2y * $tempy + $A2z * $tempz;
  my $z = $A3x * $tempx + $A3y * $tempy + $A3z * $tempz;
  
  return ($x,$y,$z);
}

sub frame {
  my ($si) = @_;
  # \vec{r}(s)
  my $rx_i = $R*cos($si/$Reff);
  my $ry_i = $R*sin($si/$Reff);
  my $rz_i = -($P/(2*$PI*$Reff) ) * $si;
  # \vec{t}(s) [ tangent ]
  my $tx_i = -($R/$Reff) * sin($si/$Reff);
  my $ty_i = ($R/$Reff) * cos($si/$Reff);
  my $tz_i = -$P/(2*$PI*$Reff);
  # \vec{n}(s) [ normal ]
  my $nx_i = -cos($si/$Reff);
  my $ny_i = -sin($si/$Reff);
  my $nz_i = 0;
  # \vec{b}(s) [ bi-normal ]
  my ($bx_i,$by_i,$bz_i) = cross($tx_i,$ty_i,$tz_i,$nx_i,$ny_i,$nz_i); # t x n
  
  return ($rx_i,$ry_i,$rz_i,$tx_i,$ty_i,$tz_i,$nx_i,$ny_i,$nz_i,$bx_i,$by_i,$bz_i); # t,n,b vectors
}

