#!/usr/bin/perl -w

use strict;

# This script is used to run DNABEND given a FASTA seq or a set of seqs, such as those made by ./make_mutated_fasta.pl.
# This is useful if someone wants to scan the seq or a set of seqs in order to compute the free energy, energy spectrum,
# or simply to make a set of files for building a dinuc etable.

my $HOME = $ENV{HOME};
my $EMINIM = $HOME."/DNABEND_1.1";

my $MAXL = 147; # max length of the nucleosome site
my $RT = 1; # (arbitary temp)

# argv
my %opts = &getCommandLineOptions();
my $fasta  = $opts{f}; # input FASTA seqs
my $Wsh    = $opts{Wsh}; # weight for the superhelical constraint
my $Wel    = $opts{Wel}; # weight for the elastic energy
my $out    = $opts{out}; # output file with energies etc. (for all input seqs)
my $prefix = $opts{prefix}; # prefix used to construct the output files (not needed unless -dumpfiles is on)
my $begin  = $opts{begin}; # one-based start of the nucleosomal site [ default = 1 ]
my $length = $opts{length}; # length of the nucleosomal site in basepairs [ default = 147 ]: ie nucleosomal site is [b,b+l-1]
my $step   = $opts{step}; # steps in which the seq is traversed (usually 1 to compute E @ all possible positions)
$Wsh = 0.1 if (!defined $Wsh); # use best "straight" one if undefined explicitly
$Wel = 1.0 if (!defined $Wel); # use default value of 1.0 if undefined explicitly
$out = "summary.dat" if (!defined $out);
my @bits_out = split /\./,$out; # process input name to create unique names for each seq
my $summary_prefix = $bits_out[0];
my $summary_postfix = "";
if (@bits_out > 1) {
  $summary_postfix = join ".",@bits_out[1 .. $#bits_out];
}
$prefix = "1kx5" if (!defined $prefix);
$begin = 1 if (!defined $begin);
$length = 147 if (!defined $length);
$step = 1 if (!defined $step);
my $end = $begin+$length-1; # b+l-1
# Check step etc.:
if ($length <= 0 || $step <= 0 || $begin <= 0) {
  die "ERROR: [begin,length,step] = [$begin,$length,$step] are not all positive!\n";
}
if ($end > $MAXL) {
  die "[begin,end=begin+length-1] must be with [1,$MAXL]!\n";
}
# Check opts compatibility:
if ($opts{link} && !$opts{dumpfiles}) {
  warn "-link option is undefined in the absence of -dumpfiles option and will be ignored!\n";
}
# Echo command to stdout:
print "\# $0 -f $fasta -Wsh $Wsh -Wel $Wel -out $out -prefix $prefix -begin $begin -length $length -step $step";
print " -dumpfiles" if ($opts{dumpfiles});
print " -link" if ($opts{link});
print " -rc" if ($opts{rc});
print "\n";
#########################

# Previously made files with ideal sets of frames:
my $ideal_helix_file = "ideal_helix_fitted.long.pdb";
my $ideal_helix_mst_file = "ideal_helix_mst_fitted.long.pdb";
my $bstable_wt = "WOprms_wt.dat";
my $bstable_ideal = "WOprms_ideal.dat";
my $curv_file = "curvature_ideal.dat";
&checkExist('d', "$EMINIM/data");
&checkExist('f', "$EMINIM/data/".$ideal_helix_file);
&checkExist('f', "$EMINIM/data/".$ideal_helix_mst_file);
&checkExist('f', "$EMINIM/data/".$bstable_wt);
&checkExist('f', "$EMINIM/data/".$bstable_ideal);
&checkExist('f', "$EMINIM/data/".$curv_file);
# Link them into current dir (even though not all are used by esolve):
`ln -s $EMINIM/data/$ideal_helix_file` unless (-e $ideal_helix_file);
`ln -s $EMINIM/data/$ideal_helix_mst_file` unless (-e $ideal_helix_mst_file);
`ln -s $EMINIM/data/$bstable_wt` unless (-e $bstable_wt);
`ln -s $EMINIM/data/$bstable_ideal` unless (-e $bstable_ideal);
`ln -s $EMINIM/data/$curv_file` unless (-e $curv_file);
# Link in the force field files:
my $Fij = "Fij_AM.dat";
my $Equil_mean = "Equil_AM.mean.dat";
my $Equil_stdev = "Equil_AM.stddev.dat";
my $Fdir = "$EMINIM/data";
&checkExist('d', $Fdir);
&checkExist('f', $Fdir."/".$Fij);
&checkExist('f', $Fdir."/".$Equil_mean);
&checkExist('f', $Fdir."/".$Equil_stdev);
`rm $Fij` if (-e $Fij);
`rm $Equil_mean` if (-e $Equil_mean);
`rm $Equil_stdev` if (-e $Equil_stdev);
`ln -s $Fdir/$Fij`;
`ln -s $Fdir/$Equil_mean`;
`ln -s $Fdir/$Equil_stdev`;

# Prepare input files for DNABEND according to the requested nucleosomal footprint:
my $tag = $begin."_".$end;
#######
my @aa = split /\./,$ideal_helix_file;
@aa = ($aa[0],$tag,@aa[1 .. $#aa]);
my $ideal_helix_file_trunc = join ".",@aa;
unless (-s $ideal_helix_file_trunc) {
  truncate_helix_file($ideal_helix_file,$ideal_helix_file_trunc,$begin,$end);
}
#######
my @bb = split /\./,$ideal_helix_mst_file;
@bb = ($bb[0],$tag,@bb[1 .. $#bb]);
my $ideal_helix_mst_file_trunc = join ".",@bb;
unless (-s $ideal_helix_mst_file_trunc) {
  truncate_helix_file($ideal_helix_mst_file,$ideal_helix_mst_file_trunc,$begin,$end-1); # one less for MST frames
}
#######
my $bstable_wt_trunc = $bstable_wt.".".$tag;
unless (-s $bstable_wt_trunc) {
  truncate_bstable($bstable_wt,$bstable_wt_trunc,$begin,$end-1);
}
#######
my $curv_file_trunc = $curv_file.".".$tag;
unless (-s $curv_file_trunc) {
  if ($tag eq "1_147") { # no need to recompute curvature coeffs for the "full" 1-147 case
    `ln -s $curv_file $curv_file_trunc`;
  }else{
    system("$EMINIM/bin/esolve -geom_wt $bstable_wt_trunc -PDBid_bp $ideal_helix_file_trunc -PDBid_mst $ideal_helix_mst_file_trunc -curv $curv_file_trunc -compute_curv");
  }
}

# Read in all the input seqs:
my @seqs = read_fasta_noBio($fasta);
# Files can be dumped only for a single FASTA seq, otherwise PDB and geom files will be rewritten with each new seq
# (this is done intentionally for back-compatibility with ROSETTA output processing scripts):
if (@seqs > 1 && $opts{dumpfiles}) {
  die "Files can be dumped only for a single FASTA seq!\n";
}

my $cnt = 1;
print "seqN   E_WO_min    E_WO_ave  E_tot_min  E_tot_ave   F_WO    Ftot   ind_min_WO  ind_min_tot  dir_WO  dir_tot\n"; # header for stdout
foreach my $seq (@seqs) {
  chomp $seq;
  my $logfile = "logfile.seq".$cnt.".dat";
  my $logfile_err = "logfile.seq".$cnt.".err";
  # Open a new summary file for each seq:
  my $cur_out = $summary_prefix.".seq".$cnt.".".$summary_postfix;
  open(XI,">$cur_out") or die "Can't open file: $!\n";
  print XI "seqN    r1    r2    Ewo      Econ     Etot     RMSD\n"; # header
  # Make a FASTA file from each seq given step + nucleosome footprint: [b,b+l-1]
  my $tmp_fasta = "parsed_seq".$cnt.".fna";
  my @ind_begin = (); # one-based position of first nuc for parsed seqs
  my @ind_end = (); # one-based position of last nuc for parsed seqs
  my ($numseqs,$remove_flag) = (undef,undef);
  if ($opts{restart} && -s $logfile) { # logfile exists and may contain partial run, examine it
    # my @known_seq_line = `tail -5 $logfile|head -1`; # old format
    my @known_seq_line = `tail -6 $logfile|head -1`; # new format
    if (@known_seq_line==1) { # putative seq found in the logfile, attempt to match
      chomp $known_seq_line[0];
      ($numseqs,$remove_flag) = parse_fasta($tmp_fasta,$seq,$step,$length,\@ind_begin,\@ind_end,$opts{rc},$known_seq_line[0]);
    }else{ # no seq found
      ($numseqs,$remove_flag) = parse_fasta($tmp_fasta,$seq,$step,$length,\@ind_begin,\@ind_end,$opts{rc});
    }
  }else{ # no logfile found, or -restart is not given
    ($numseqs,$remove_flag) = parse_fasta($tmp_fasta,$seq,$step,$length,\@ind_begin,\@ind_end,$opts{rc});
  }
  if (defined $remove_flag) { # logfile was not found useful, remove it
    `/bin/rm $logfile` if (-e $logfile);
  }
  #debug
  #exit;
  # Prepare a list of output PDB and geometry files if required:
  my $tmp_pdblist = "pdbout.dat";
  my $tmp_geomlist = "geomout.dat";
  my @pdbs = (); # list of output PDB files
  my @geoms = (); # list of output BStable geoms (which are appended below to the output PDB files)
  if ($opts{dumpfiles}) {
    make_filelist($prefix,$numseqs,$tmp_pdblist,$tmp_geomlist,\@pdbs,\@geoms);
  }
  # Run the main code:
  if ($opts{dumpfiles}) {
    `$EMINIM/bin/esolve -fasta $tmp_fasta -geom_wt $bstable_wt_trunc -PDBid_bp $ideal_helix_file_trunc -PDBid_mst $ideal_helix_mst_file_trunc -fpdb $tmp_pdblist -fgeom $tmp_geomlist -curv $curv_file_trunc -w_sh $Wsh -w_el $Wel >> $logfile 2> $logfile_err`;
    # Merge PDB & geom files:
    merge_files(\@pdbs,\@geoms);
    # Link up reference files in a specific way reminiscent of ROSETTA runs (if requested):
    if ($opts{link}) {
      link_pdb_files(\@pdbs,0);
    }
  }else{
    `$EMINIM/bin/esolve -fasta $tmp_fasta -geom_wt $bstable_wt_trunc -PDBid_bp $ideal_helix_file_trunc -PDBid_mst $ideal_helix_mst_file_trunc -curv $curv_file_trunc -w_sh $Wsh -w_el $Wel >> $logfile 2> $logfile_err`;
  }
  # Parse the output file and output table with energies, RMSDs etc.:
  my @Ewo = (); # elastic energy: Ewo
  my @Econstr = (); # constraint energy: Esh
  my @Etot = (); # total energy: w1*Ewo + w2*Esh
  my @rmsd = (); # rmsds to initial ideal struc
  parse_esolve_output($logfile,\@Ewo,\@Econstr,\@Etot,\@rmsd);
  if ($#ind_begin != $#Ewo) { # must have [Ewo,Etot,RMSD] for each [r1,r2]
    die "Could not find a unique set of [Ewo,Etot,RMSD] for every seq!\n";
  }
  for my $i (0 .. $#Ewo) {
    printf XI "%4d  %4d  %4d %8.3f %8.3f %8.3f %8.3f\n",$cnt,$ind_begin[$i],$ind_end[$i],$Ewo[$i],$Econstr[$i],$Etot[$i],$rmsd[$i];
  }
  # Print minE, free energy (if applicable) etc.:
  my ($minE_WO,$minind_WO,$strand_WO) = (undef,undef,"+");
  my ($minEtot,$minind_tot,$strand_tot) = (undef,undef,"+");
  if (!defined $opts{rc}) {
    ($minE_WO,$minind_WO) = min(\@Ewo);
    ($minEtot,$minind_tot) = min(\@Etot);
  }else{
    ($minE_WO,$minind_WO,$strand_WO) = min_rc(\@Ewo,\@ind_begin,\@ind_end);
    ($minEtot,$minind_tot,$strand_tot) = min_rc(\@Etot,\@ind_begin,\@ind_end);
  }
  printf "%4d   %8.3f   %8.3f  %8.3f  %8.3f  %8.3f  %8.3f   %5d     %5d      %1s      %1s\n",
    $cnt,$minE_WO,ave(\@Ewo),$minEtot,ave(\@Etot),Fen(\@Ewo),Fen(\@Etot),$minind_WO,$minind_tot,$strand_WO,$strand_tot;
  # Clean up:
  #`/bin/rm $tmp_fasta` if (-e $tmp_fasta);
  `/bin/rm $tmp_pdblist` if (-e $tmp_pdblist);
  `/bin/rm $tmp_geomlist` if (-e $tmp_geomlist);
  $cnt++;
  close XI;
}

#################
## Functions
#################

sub getCommandLineOptions {
  use Getopt::Long;
  my $usage = qq{usage: $0 -f seqs.fna <-Wsh weight_sh -Wel weight_el -out summary.dat -prefix 1kx5 -begin begin -length length -step step -dumpfiles -link -rc -restart>\n};
  # Get args
  my %opts = ();
  &GetOptions(\%opts, "f=s", "Wsh=f", "Wel=f", "out=s", "prefix=s", "begin=i", "length=i", "step=i", "dumpfiles", "link", "rc", "restart");
  if (!defined $opts{f}) {
    print STDERR "$usage\n";
    exit -1;
  }
  &checkExist('f', $opts{f});
  return %opts;
}

# checkExist()
#
sub checkExist {
  my ($type, $path) = @_;
  if ($type eq 'd') {
    if (! -d $path) {
      warn("dir not found: $path\n");
      exit -3;
    }
  }
  elsif ($type eq 'f') {
    if (! -f $path) {
      warn("file not found: $path\n");
      exit -3;
    }
    elsif (! -s $path) {
      warn("empty file: $path\n");
      exit -3;
    }
  }
}

# This sub loads fasta seqs into an array WITHOUT BioPerl
# functions. It can optionally perform various checks on
# the seqs it has read.
sub read_fasta_noBio {
  my ($fasta_file,$mode) = @_;
  my @seqs = ();
  my $cnt = -1;
  open(II,"$fasta_file") or die "Can't open file: $!\n";
  while (<II>) {
    chomp;
    my @tt = split ' ';
    next if (@tt==0);
    if (substr($tt[0],0,1) eq '>') {
      $cnt++;
      next;
    }
    die "Unrecognized FASTA seq!\n" unless (@tt==1);
    if (!defined $seqs[$cnt]) {
      $seqs[$cnt] = uc($tt[0]); # make everything uppercase!!!
    }else{
      $seqs[$cnt] .= uc($tt[0]); # make everything uppercase!!!
    }
  }
  close II;
  if (defined $mode) {
    if ($mode eq "check_width") { # checks if all seqs are of the same length
      for my $i (1 .. $#seqs) {
	die "Seqs read from $fasta_file are not of the same length!\n"
	  if ((length($seqs[$i]) - length($seqs[$i-1]) ) != 0);
      }
    }
  }
  return @seqs;
}

# This sub parses a single FASTA seq into multiple fixed length seqs according to the requested step
# and the nucleosomal footprint.
sub parse_fasta {
  my ($fasta_file,$seq,$step,$length,$ind_begin,$ind_end,$revcom,$last_known_seq) = @_;
  # Check refs:
  if (ref($ind_begin) ne 'ARRAY' || ref($ind_end) ne 'ARRAY') {
    die "Expected array ref(s) in parse_fasta!\nError in function args!\n";
  }
  #debug
  #print $last_known_seq,"\n" if (defined $last_known_seq);
  my @dna_letters = split //,$seq;
  # Make all seq pieces:
  my @all_seqs = (); # seq strings
  my @all_seqIDs = (); # seqIDs
  my $piece_num = 0;
  for (my $j=0; $j<=($#dna_letters+1-$length); $j+=$step) {
    my $seqID = "seq.".($piece_num+1);
    my $curseq = join "",@dna_letters[$j .. $j+$length-1];
    push @all_seqIDs, $seqID;
    push @all_seqs, $curseq;
    push @$ind_begin, $j+1; # one-based!
    push @$ind_end, $j+$length; # one-based!
    $piece_num++;
    if (defined $revcom) { # rc requested
      push @all_seqIDs, $seqID."_rc";
      push @all_seqs, rc($curseq);
      push @$ind_begin, $j+$length; # one-based!
      push @$ind_end, $j+1; # one-based!
      $piece_num++;
    }
  }
  # Resize arrays if the last_known_seq exists and matches one of the seqs in our array:
  my $remove_flag = 1; # remove previous logfile by default
  if (defined $last_known_seq) {
    for my $i (0 .. $#all_seqs) {
      if ($all_seqs[$i] eq $last_known_seq) { # always keep seqs starting from the first match
	$remove_flag = undef; # do not remove previous logfile ..
	@all_seqs = @all_seqs[$i+1 .. $#all_seqs];
	@all_seqIDs = @all_seqIDs[$i+1 .. $#all_seqIDs];
	#@$ind_begin = $ind_begin->[$i+1 .. $#$ind_begin];
	#@$ind_end = $ind_end->[$i+1 .. $#$ind_end];
	last;
      }
    }
  }
  # Output all seqs:
  open(I,">$fasta_file") or die "Can't open file: $!\n";
  for my $i (0 .. $#all_seqs) {
    printf I ">%s\n",$all_seqIDs[$i];
    printf I "%s\n",$all_seqs[$i];
  }
  close I;
  return ($piece_num,$remove_flag);
}

#This aux. sub returns a reverse complement of the input
#sequence.
sub rc {
  my ($seq) = @_;
  my %pairs = ();
  $pairs{'A'} = 'T';
  $pairs{'T'} = 'A';
  $pairs{'C'} = 'G';
  $pairs{'G'} = 'C';
  my @bits_r = reverse split('',$seq);
  my @bits_rc = ();
  foreach my $bit (@bits_r) {
    if (defined $pairs{$bit}) {
      push @bits_rc,$pairs{$bit};
    }else{
      die "Unknown base in rc: $bit!\n";
    }
  }
  return join '',@bits_rc;
}

# This sub truncates PDB format files (including "long" format) for ideal helices.
sub truncate_helix_file {
  my ($filein,$fileout,$b,$e) = @_;
  open(V,"$filein") or die "Can't open file: $!\n";
  my @lines = (<V>);
  close V;
  open(VI,">$fileout") or die "Can't open file: $!\n";
  foreach my $line (@lines) {
    chomp $line;
    next if (length($line)==0);
    if ($line =~ /ATOM/) {
      my $resnum = substr($line,22,4);
      if ($resnum >= $b && $resnum <= $e) {
	print VI $line,"\n";
      }
    }else{
      print VI $line,"\n"; # copy all non-ATOM lines
    }
  }
  close VI;
}

# This sub truncates BStable files.
sub truncate_bstable {
  my ($filein,$fileout,$b,$e) = @_;
  open(V,"$filein") or die "Can't open file: $!\n";
  my @lines = (<V>);
  close V;
  open(VI,">$fileout") or die "Can't open file: $!\n";
  foreach my $line (@lines) {
    chomp $line;
    my @tt = split ' ',$line;
    next if (@tt==0);
    if ($tt[0] eq "res") {
      print VI $line,"\n";
    }else{
      my $resnum = $tt[0];
      if ($resnum >= $b && $resnum <= $e) {
	print VI $line,"\n";
      }
      last if ($resnum > $e);
    }
  }
  close VI;
}

# This sub makes lists of PDB and geom files used as input by DNABEND.
sub make_filelist {
  my ($prefix,$numseqs,$pdblist,$geomlist,$pdbs,$geoms) = @_;
  # Check refs:
  if (ref($pdbs) ne 'ARRAY' || ref($geoms) ne 'ARRAY') {
    die "Expected array ref(s) in make_filelist!\nError in function args!\n";
  }
  my $max_bits = length($numseqs);
  open(I,">$pdblist") or die "Can't open file: $!\n";
  open(II,">$geomlist") or die "Can't open file: $!\n";
  for my $j (0 .. $numseqs-1) {
    my $tag = ('0' x ($max_bits-length($j))).$j;
    my $curpdb = $prefix.".".$tag."_0001.pdb";
    my $curgeom = $prefix.".".$tag."_0001.geom";
    printf I "%s\n",$curpdb;
    printf II "%s\n",$curgeom;
    push @$pdbs, $curpdb;
    push @$geoms, $curgeom;
  }
  close I;
  close II;
}

# This sub appends geom files to the end of PDB files.
sub merge_files {
  my ($pdbs,$geoms) = @_;
  # Check refs:
  if (ref($pdbs) ne 'ARRAY' || ref($geoms) ne 'ARRAY') {
    die "Expected array ref(s) in merge_files!\nError in function args!\n";
  }
  # Check dims:
  if ($#$pdbs != $#$geoms) {
    die "PDB and geom filelists must have the same dims!\n";
  }
  for my $i (0 .. $#$pdbs) {
    my $curpdb = $pdbs->[$i];
    my $curgeom = $geoms->[$i];
    &checkExist('f', $curpdb);
    &checkExist('f', $curgeom);
    `cat $curgeom >> $curpdb`;
    `/bin/rm $curgeom`;
  }
}

# This sub creates a set of softlinks to a PDB specified by its $ind thereby
# making a set of reference files for other scripts.
sub link_pdb_files {
  my ($pdbs,$ind) = @_;
  # Check refs:
  if (ref($pdbs) ne 'ARRAY') {
    die "Expected array ref(s) in link_pdb_files!\nError in function args!\n";
  }
  # Check index:
  if ($ind < 0 || $ind > $#$pdbs) {
    die "Reference file index is out of range!\n";
  }
  my $basepdb = $pdbs->[$ind];
  # Create links:
  foreach my $pdb (@$pdbs) {
    chomp $pdb;
    my @aa = split /\./,$pdb;
    die "Unknown filename format: $pdb!\n" unless (@aa==3);
    my @bb = split /\_/,$aa[1];
    die "Unknown filename format: $pdb!\n" unless (@bb==2);
    my $refpdb = join ".", ($aa[0],$bb[0],"ref_0001",$aa[2]);
    `ln -s $basepdb $refpdb`;
  }
}

# This sub parses DNABEND output for energies and RMSDs.
sub parse_esolve_output {
  my ($logfile,$Ewo,$Econstr,$Etot,$rmsd) = @_;
  # Check refs:
  if (ref($Ewo) ne 'ARRAY' || ref($Econstr) ne 'ARRAY' ||
      ref($Etot) ne 'ARRAY' || ref($rmsd) ne 'ARRAY') {
    die "Expected array ref(s) in parse_esolve_output!\nError in function args!\n";
  }
  my @Ewo_lines = `grep E_WO $logfile`;
  my @Econ_lines = `grep Econstr $logfile`;
  my @Etot_lines = `grep Etot $logfile`;
  my @rmsd_lines = `grep RMSD $logfile`;
  die "Unequal dims in [Ewo,Etot,RMSD] arrays!\n"
    if ($#Econ_lines != $#Etot_lines || $#Ewo_lines != $#Etot_lines || $#Ewo_lines != $#rmsd_lines);
  for my $j (0 .. $#Ewo_lines) {
    chomp $Ewo_lines[$j];
    chomp $Econ_lines[$j];
    chomp $Etot_lines[$j];
    chomp $rmsd_lines[$j];
    my @aa = split ' ',$Ewo_lines[$j];
    my @dd = split ' ',$Econ_lines[$j];
    my @bb = split ' ',$Etot_lines[$j];
    my @cc = split ' ',$rmsd_lines[$j];
    die "Unknown logfile line format!\n" if (@aa!=3 || @dd!=3 || @bb!=3 || @cc!=3);
    push @$Ewo, $aa[2];
    push @$Econstr, $dd[2];
    push @$Etot, $bb[2];
    push @$rmsd, $cc[2];
  }
}

# This sub is specific to the "rc" mode: given an array of energies and [begin,end] indices
# (which help discriminate between "+" and "-" directions of seq scans), it returns the min
# energy + the index of the *original* seq and its direction (+/-).
sub min_rc {
  my ($Eref,$ind_begin,$ind_end) = @_;
  #Check refs:
  if (ref($Eref) ne 'ARRAY' || ref($ind_begin) ne 'ARRAY' || ref($ind_end) ne 'ARRAY') {
    die "Expected array ref(s) in min_rc!\nError in function args!\n";
  }
  #Check dims:
  if ($#$Eref != $#$ind_begin || $#$Eref != $#$ind_end) {
    die "Unequal array dims in min_rc!\n";
  }
  my ($minE,$minind) = min($Eref);
  my $strand;
  if ($ind_begin->[$minind] < $ind_end->[$minind]) { # '+'
    $strand = "+";
  }else{
    $strand = "-";
  }
  $minind = int($minind/2);
  return ($minE,$minind,$strand);
}

# This function simply finds a minimum among all values in an input array,
# and returns it along with the array index.
sub min {
  my ($aref) = @_;
  #Check refs:
  if (ref($aref) ne 'ARRAY') {
    die "Expected array ref(s) in min!\nError in function args!\n";
  }
  my $min = 10000000000;
  my $minind = undef;
  for my $j (0 .. $#$aref) {
    if ($aref->[$j] < $min) {
      $min = $aref->[$j];
      $minind = $j; # 0-based index
    }
  }
  die "Could not find array min!\n" unless (defined $minind);
  return ($min,$minind);
  #return $min;
}

# This function simply finds an average value for all values in an input array.
sub ave {
  my ($aref) = @_;
  #Check refs:
  if (ref($aref) ne 'ARRAY') {
    die "Expected array ref(s) in ave!\nError in function args!\n";
  }
  my $ave = 0.0;
  if (scalar(@$aref) > 0) {
    for my $j (0 .. $#$aref) {
      $ave += $aref->[$j];
    }
    $ave /= ($#$aref+1);
  }
  return $ave;
}

# This sub computes free energy F for a set of energies.
# NOTE: F may not be well defined if step > 1 (ie not all possible seqs are sampled).
# NOTE2: this definition of F shifts all energies by <E> (ie by the average energy
# of the sequence). In order to restore the original def add <E> back in: F + <E> is
# the proper free energy. Otherwise seqs cannot be compared with each other unless
# <E> happens to be very close for all seqs!!
sub Fen {
  my ($Eref) = @_;
  #Check refs:
  if (ref($Eref) ne 'ARRAY') {
    die "Expected array ref(s) in Fen!\nError in function args!\n";
  }
  my $meanE = ave($Eref); # choose <E> as ref point
  my $Z = 0.0; # partition function
  for my $i (0 .. $#$Eref) { # over each energy
    my $curE = $Eref->[$i] - $meanE;
    $Z += exp(-$curE/$RT);
  }
  return -$RT*log($Z);
}

