#!/usr/local/perl/bin/perl -w

use strict;
use POSIX qw(ceil floor);

# This aux. script sets up a CONDOR run starting from several FASTA files in the current dir.
# The FASTA files are listed in the input file.

my $HOME = $ENV{HOME}; # if you run from home dir
#my $HOME = "/exports/morozov"; # if you run on the cluster
my $EXE = "$HOME/DNABEND_1.1/scripts/RunEminim.pl";

# argv
my %opts  = &getCommandLineOptions();
my $file  = $opts{l}; # list of FASTA files (one per line)
my $Nruns = $opts{n}; # number of pieces to split an exe file into (for DAVID only, condor file stays the same!)
$Nruns = 1 if (!defined $Nruns);
die "Illegal Nruns: $Nruns!\n" if ($Nruns < 0);

# Read file info:
my @fi = read_list($file);
my $cnt = 1;
my @dirs = ();
foreach my $ff (@fi) {
  chomp $ff;
  &checkExist('f', $ff);
  # Name dir using seqIDs (which become filenames if split_fasta.pl was used to create FASTA files):
  #my $curdir = "run".$cnt++;
  my @aa = split /\./, $ff;
  $#aa--;
  my $curdir = join ".",@aa;
  `mkdir $curdir` unless (-d $curdir);
  #`cp $ff $curdir`; # do not copy, make softlinks instead
  chdir $curdir or die "Can't cd: $!\n";
  `ln -s ../$ff` unless (-e $ff);
  chdir '..' or die "Can't cd: $!\n";
  push @dirs, $curdir;
}

my $condor_file = "condor.run";
make_condor_file($condor_file,\@fi,\@dirs);
my $exe_file = "dna_bend.pl"; # actual exe_files will be named dna_bend.$cnt.pl ..
make_exe_file($exe_file,\@fi,\@dirs,$Nruns);

####################
###  Functions  ####
####################

sub getCommandLineOptions {
  use Getopt::Long;
  my $usage = qq{usage: $0 -l list_of_files <-dumpfiles -n Nruns>\n};
  # Get args
  my %opts = ();
  &GetOptions(\%opts, "l=s", "dumpfiles", "n=i");
  if (!defined $opts{l}) {
    print STDERR "$usage\n";
    exit -1;
  }
  &checkExist('f', $opts{l});
  return %opts;
}

# checkExist()
#
sub checkExist {
  my ($type, $path) = @_;
  if ($type eq 'd') {
    if (! -d $path) {
      warn("dir not found: $path\n");
      exit -3;
    }
  }
  elsif ($type eq 'f') {
    if (! -f $path) {
      warn("file not found: $path\n");
      exit -3;
    }
    elsif (! -s $path) {
      warn("empty file: $path\n");
      exit -3;
    }
  }
}

# This function reads in a list of items from a file (one item per line, rest of line ignored;
# '#' comment symbols recognized).
sub read_list {
  my ($file) = @_;
  my @tokens = ();
  my $cnt = 0;
  open(I,"$file") or die "Can't open file: $!\n";
  while (<I>) {
    chomp;
    my @tt = split ' ';
    next if (@tt==0);
    next if (substr($tt[0],0,1) eq '#');
    my $token = $tt[0];
    push @tokens,$token;
    $cnt++;
  }
  close I;
  print "Read $cnt token(s) from file: $file ..\n";
  return @tokens;
}

# This sub outputs a condor file.
sub make_condor_file {
  my ($condor_file,$fi,$dirs) = @_;
  #Check refs:
  if (ref($fi) ne 'ARRAY' || ref($dirs) ne 'ARRAY') {
    die "Expected array/hash ref(s) in make_condor_file!\nError in function args!\n";
  }
  my $init_dir = `pwd`;
  chomp($init_dir);
  open(II,">$condor_file") or die "Can't open file: $!\n";
  print II "#!/bin/tcsh\n";
  print II "# Condor script file\n";
  print II "# Tanja Kortemme and Alex Morozov\n\n";
  print II "########\n# setup\n########\n\n";
  print II "universe     = vanilla\nRequirements = Arch == \"INTEL\"\nGetEnv       = True\n\n";
  print II "Notify_user  = morozov\@u.washington.edu\nnotification = Error\n\n";
  print II "Log          = condor.log\n\n";
  print II "Executable   = $EXE\n";
  print II "Arguments    = \$(FlagFasta) \$(Fasta) \$(FlagWsh) \$(Wsh) \$(Flagbegin) \$(begin) \$(Flaglength) \$(length) \$(Flagstep) \$(step) \$(Flagrc) \$(Restart)";
  print II " \$(Flagdump)" if ($opts{dumpfiles});
  print II "\n";
  print II "FlagFasta    = -f\n";
  print II "FlagWsh      = -Wsh\n";
  print II "Flagbegin    = -begin\n";
  print II "Flaglength   = -length\n";
  print II "Flagstep     = -step\n";
  print II "Flagrc       = -rc\n";
  print II "Restart      = -restart\n";
  print II "Flagdump     = -dumpfiles\n" if ($opts{dumpfiles});
  printf II "Wsh          = %f\n",0.1;
  printf II "begin        = %d\n",1;
  printf II "length       = %d\n",147;
  printf II "step         = %d\n\n",1;
  print II "###########################\n# here are the actual runs\n###########################\n\n";
  for my $j (0 .. $#$fi) {
    printf II "InitialDir  = %s/%s\n",$init_dir,$dirs->[$j];
    printf II "Fasta       = %s\n",$fi->[$j];
    my @aa = split /\./,$fi->[$j];
    my @bb = @aa;
    $aa[$#aa] = "out";
    $bb[$#bb] = "err";
    my $out = join ".",@aa;
    my $err = join ".",@bb;
    printf II "Output      = %s\n",$out;
    printf II "Error       = %s\n",$err;
    print II "Queue\n\n";
  }
  close II;
  `chmod +x $condor_file`;
}

# This sub outputs an exe file (suitable to launch on the local [RU] cluster).
sub make_exe_file {
  my ($exe_file,$fi,$dirs,$Nruns) = @_;
  #Check refs:
  if (ref($fi) ne 'ARRAY' || ref($dirs) ne 'ARRAY') {
    die "Expected array/hash ref(s) in make_exe_file!\nError in function args!\n";
  }
  my $step = ceil(@$fi/$Nruns);
  my $cnt = 1;
  my @exe_bits = split /\./,$exe_file;
  for (my $i=0; $i<=$#$fi; $i+=$step) {
    my $exe_file_cur = join(".",(@exe_bits[0 .. $#exe_bits-1],$cnt,$exe_bits[$#exe_bits]) );
    $cnt++;
    my $lastind = $i+$step-1;
    if ($i+$step-1 > $#$fi) {
      $lastind = $#$fi;
    }
    open(III,">$exe_file_cur") or die "Can't open file: $!\n";
    #print III "#!/usr/bin/perl -w\n\n";
    print III "#!/usr/local/perl/bin/perl -w\n\n";
    print III "use strict;\n\n";
    for my $j ($i .. $lastind) {
      my @aa = split /\./,$fi->[$j];
      my @bb = @aa;
      $aa[$#aa] = "out";
      $bb[$#bb] = "err";
      my $out = join ".",@aa;
      my $err = join ".",@bb;
      printf III "chdir \"%s\" or die \"Can\'t cd: \$\!\\n\";\n",$dirs->[$j];
      printf III "\`%s -f %s -Wsh %f -begin %d -length %d -step %d -rc -restart",$EXE,$fi->[$j],0.1,1,147,1;
      print III " -dumpfiles" if ($opts{dumpfiles});
      printf III " > %s 2> %s \&\`;\n",$out,$err;
      print III "chdir \'..\' or die \"Can\'t cd: \$\!\\n\";\n";
      printf III "sleep %d;\n\n",5;
    }
    close III;
    `chmod +x $exe_file_cur`;
  }
  printf "Created %d exe_files ..\n",$cnt-1;
}

