#!/usr/local/perl/bin/perl -w

use strict;

# This aux. script combines results of DNABEND runs in several directories, and produces
# an overall output file in the following format:
# seqN    Ewo      Econ     Etot     RMSD    Ewo_rc      Econ_rc     Etot_rc     RMSD_rc
# Here seqN is the absolute one-based coordinate along the chromosome.

my $Lseq = 147; # seq length used in the run

# argv
my %opts = &getCommandLineOptions();
my $dir   = $opts{prefix}; # prefix for all dirs in which output files reside
my $list  = $opts{l}; # OR list of all dirs (one per line)
my $fasta = $opts{f};
my $out   = $opts{out}; # output table
$out = "default.out" if (!defined $out);

my @dirs = ();
if (defined $list) {
  @dirs = read_list($list);
  die "No dirs found in file: $list!\n" if (@dirs==0);
}elsif (defined $dir) {
  @dirs = `ls -d $dir*`;
  die "No dirs with prefix: $dir found!\n" if (@dirs==0);
}

# Sort dirs by names (IMPORTANT: that's what tells the script the absolute number for each run):
# All dirs should have names prefixN, where N is the relative number of the run.
# If a dir_list is provided thru the -l option, it is up to the user to ensure the correct assembly order!!
my %nums = (); # hash of dir names by run numbers (inferred from dir names!)
if (defined $dir) {
  foreach my $dd (@dirs) {
    chomp $dd;
    &checkExist('d', $dd);
    my @aa = split /$dir/, $dd;
    $nums{$aa[1]} = $dd;
  }
}elsif (defined $list) {
  my $dir_cnt = 0;
  foreach my $dd (@dirs) {
    chomp $dd;
    &checkExist('d', $dd);
    $nums{$dir_cnt} = $dd;
    $dir_cnt++;
  }
}

my $seq_cnt = 1; # one-based seq position along the chromosome
open(I,">$out") or die "Can't open file: $!\n";
my $header = sprintf("%8s  %8s  %8s  %8s  %8s  %8s  %8s  %8s  %8s",
		     "seqN","Ewo","Econ","Etot","RMSD","Ewo_rc","Econ_rc","Etot_rc","RMSD_rc");
printf I "%s\n",$header;
foreach my $key (reverse sort numerically keys %nums) {
  my $curdir = $nums{$key};
  chdir $curdir or die "Can't cd: $!\n";
  print "Working in $curdir ..\n";
  my $summary = "summary.seq1.dat";
  open(II,"$summary") or die "Can't open file: $!\n";
  my @lines = (<II>);
  close II;
  for (my $i=1; $i<$#lines; $i+=2) {
    chomp $lines[$i];
    chomp $lines[$i+1];
    my @aa = split ' ',$lines[$i];
    my @bb = split ' ',$lines[$i+1];
    next unless (@aa==7 && @bb==7);
    my $scoreline = sprintf("%8.3f  %8.3f  %8.3f  %8.3f  %8.3f  %8.3f  %8.3f  %8.3f",
			    $aa[3],$aa[4],$aa[5],$aa[6],$bb[3],$bb[4],$bb[5],$bb[6]);
    printf I "%8d  %s\n",$seq_cnt,$scoreline;
    $seq_cnt++;
  }
  #$seq_cnt += ($Lseq - 1); # this accounts for "gaps" between seqs in adjacent files which were not sampled
  chdir '..' or die "Can't cd: $!\n";
}
close I;

# Optionally, carry out a check using a file with the original FASTA seq (ie the chromosome):
if (defined $fasta) {
  my @seqs = read_fasta_noBio($fasta);
  my $char_cnt = 0;
  foreach my $seq (@seqs) {
    chomp $seq;
    my @seq_bits = split //,$seq;
    foreach my $bit (@seq_bits) {
      $char_cnt++;
    }
  }
  my $expected_num_seqs = $char_cnt - ($Lseq - 1); # last Lseq-1 [=146] positions are illegal
  my $actual_num_seqs = $seq_cnt - 1; # seq_cnt was one-based
  printf "Number of nucleotides in seq = %d\n",$char_cnt;
  printf "Number of seqs found in DNABEND output = %d\n",$actual_num_seqs;
  warn "WARNING: Mismatch between expected ($expected_num_seqs) and actual ($actual_num_seqs) number of seqs!\n"
    if ($actual_num_seqs != $expected_num_seqs);
}

####################
###  Functions  ####
####################

sub getCommandLineOptions {
  use Getopt::Long;
  my $usage = qq{usage: $0 [-prefix dir_prefix OR -l dir_list] <-f fasta.fna -out default.out>\n};
  # Get args
  my %opts = ();
  &GetOptions(\%opts, "prefix=s", "f=s", "out=s", "l=s");
  # Check for legal invocation
  if ((!defined $opts{prefix} && !defined $opts{l}) ||
      (defined $opts{prefix} && defined $opts{l}) ) {
    print STDERR "$usage\n";
    exit -1;
  }
  &checkExist('f', $opts{f}) if (defined $opts{f});
  &checkExist('f', $opts{l}) if (defined $opts{l});
  return %opts;
}

# checkExist()
#
sub checkExist {
  my ($type, $path) = @_;
  if ($type eq 'd') {
    if (! -d $path) {
      warn("dir not found: $path\n");
      exit -3;
    }
  }
  elsif ($type eq 'f') {
    if (! -f $path) {
      warn("file not found: $path\n");
      exit -3;
    }
    elsif (! -s $path) {
      warn("empty file: $path\n");
      exit -3;
    }
  }
}

# This function reads in a list of items from a file (one item per line, rest of line ignored;
# '#' comment symbols recognized).
sub read_list {
  my ($file) = @_;
  my @tokens = ();
  my $cnt = 0;
  open(I,"$file") or die "Can't open file: $!\n";
  while (<I>) {
    chomp;
    my @tt = split ' ';
    next if (@tt==0);
    next if (substr($tt[0],0,1) eq '#');
    my $token = $tt[0];
    push @tokens,$token;
    $cnt++;
  }
  close I;
  print "Read $cnt token(s) from file: $file ..\n";
  return @tokens;
}

#This sub loads fasta seqs into an array WITHOUT BioPerl
#functions. It can optionally perform various checks on
#the seqs it read.
sub read_fasta_noBio {
  my ($fasta_file,$mode) = @_;
  my @seqs = ();
  my $cnt = -1;
  open(II,"$fasta_file") or die "Can't open file: $!\n";
  while (<II>) {
    chomp;
    my @tt = split ' ';
    next if (@tt==0);
    if (substr($tt[0],0,1) eq '>') {
      $cnt++;
      next;
    }
    die "Unrecognized FASTA seq!\n" unless (@tt==1);
    if (!defined $seqs[$cnt]) {
      $seqs[$cnt] = uc($tt[0]); # make everything uppercase!!!
    }else{
      $seqs[$cnt] .= uc($tt[0]); # make everything uppercase!!!
    }
  }
  close II;
  if (defined $mode) {
    if ($mode eq "check_width") { # checks if all seqs are of the same length
      for my $i (1 .. $#seqs) {
	die "Seqs read from $fasta_file are not of the same length!\n"
	  if ((length($seqs[$i]) - length($seqs[$i-1]) ) != 0);
      }
    }
  }
  return @seqs;
}

sub numerically { $b <=> $a; }

