#!/usr/bin/perl -w

use strict;
use Bio::Perl;
use Bio::SeqIO;
use POSIX qw(ceil floor);

# This small script splits a single long FASTA seq into multiple seqs.
# ie >seqID description
#    line1
#    line2
#    ...
# into
# >seqID.1
#  line1
# >seqID.2
#  line2
#  ...

# Usage for creating input seqs in nucleosome runs [32 = 2 * number_of_processors]:
# ./split_single_seq.pl -f chrom.fa -n 32 -tail 146

# Usage for splitting input seq into nucleosome binding sites:
# ./split_single_seq.pl -f chrom.fa -tail 146 -forceL 147

# argv
my %opts  = &getCommandLineOptions();
my $fasta = $opts{f}; # input file
my $n     = $opts{n}; # number of seqs into which to split the input seq
my $tail  = $opts{tail}; # overlap between seqs [i-1,i], eg AACCGGGT GGTCAAAAA has overlap=3 (GGT)
my $Lseq  = $opts{forceL}; # keep seq pieces with only this value of L (useful if we want to split into equal pieces with step=1)
my $out   = $opts{out}; # output file
$tail = 0 if (!defined $tail);
#$n = 1 if (!defined $n);
$out = $fasta.".split" if (!defined $out);

my $fasta_Fh = Bio::SeqIO->newFh(-file => $fasta,
				 -format => 'fasta');

# NOTE: "tail" of the long seq is not lost if the step is not a multiple
#       of seqL - it should be in the last file
my $fasta_out_Fh = Bio::SeqIO->newFh(-file => ">$out",
				     -format => 'fasta');

# Determine step_size and check the validity of input seq:
my @seqs = (<$fasta_Fh>);
die "Empty FASTA file: $fasta!\n" if (@seqs == 0);
die "Multiple FASTA seqs are not allowed!\n" if (@seqs > 1);
my @seq_bits = split //,$seqs[0]->seq;
################################
# NOTE: we set n here if not defined
################################
$n = scalar(@seq_bits) if (!defined $n);
my $step = ceil(@seq_bits/$n);

my $cnt = 1;
for (my $i=0; $i<=$#seq_bits; $i+=$step) {
  my $curID = $seqs[0]->id.".".$cnt;
  $cnt++;
  my $lastind = $i+$step-1+$tail;
  if ($i+$step-1+$tail > $#seq_bits) {
    $lastind = $#seq_bits;
  }
  my $cur_desc = sprintf(" [%d,%d]",$i,$lastind);
  my $curseq = join "",@seq_bits[$i .. $lastind];
  next if (defined $Lseq && length($curseq) != $Lseq);
  my $seq = Bio::Seq->new(-seq              => $curseq,
			  -desc             => $cur_desc,
			  -id               => $curID);
  print $fasta_out_Fh $seq;
}

####################
###  Functions  ####
####################

sub getCommandLineOptions {
  use Getopt::Long;
  my $usage = qq{usage: $0 -f seqs.fna <-n Nseqs -tail overlap -forceL Lseq -out seqs_out.fna>\n};
  # Get args
  my %opts = ();
  &GetOptions(\%opts, "f=s", "n=i", "out=s", "tail=i", "forceL=i");
  if (!defined $opts{f}) {
    print STDERR "$usage\n";
    exit -1;
  }
  &checkExist('f', $opts{f});
  return %opts;
}

# checkExist()
#
sub checkExist {
  my ($type, $path) = @_;
  if ($type eq 'd') {
    if (! -d $path) {
      warn("dir not found: $path\n");
      exit -3;
    }
  }
  elsif ($type eq 'f') {
    if (! -f $path) {
      warn("file not found: $path\n");
      exit -3;
    }
    elsif (! -s $path) {
      warn("empty file: $path\n");
      exit -3;
    }
  }
}

