/* anneal.c */

/*****************************************************************
        Copyright by Rockefeller University,
can not be reproduced or distributed without written permission of
the copyright holder.  Version 1.1 of April 2006.

Written by Alexandre Morozov (contact person: morozov@edsb.rockefeller.edu),
and Eric Siggia.

The program DNABEND implements an algorithm for computing elastic energies
for DNA molecules of arbitrary length and sequence. DNA can be constrained
to follow an arbitrary 3D curve, such as a left-handed superhelix found
in nucleosomal core particles.

The DNABEND method and its applications are described in the following
paper:

******************************************************************/

#include "anneal.h"

//====  Simulated Annealing Functions ====//

gsl_siman_params_t SArun_params = {N_TRIES, ITERS_FIXED_T, STEP_SIZE,
				   K, T_INITIAL, MU_T, T_MIN};

void Anneal(Sequence::Sequence *seq, DNAparams::DNAparams &prms, Geom::Geom &xyz_ideal,
	    Geom::Geom &xyz_wt, struct Options *opt) {
  // Check runmode:
  std::string runmode = opt->minmax;
  if (runmode != "min" && runmode != "max") {
    std::cerr << "ERROR in Anneal => invalid runmode: "<<runmode<<"!\n";
    exit(EXIT_FAILURE);
  }
  
  // Initialize parameters:
  SA_params conf;
  conf.ss = seq;
  conf.DNA = &prms;
  conf.ideal = &xyz_ideal;
  conf.wt = &xyz_wt;
  conf.op = opt;
  conf.start = opt->start;
  conf.stop = opt->stop;
  
  // Check and set the [start,stop] coords:
  if (conf.start == -1 || conf.stop == -1) {
    conf.start = 1;
    conf.stop = seq->Length();
    //fprintf(stderr,"Both -start and -stop options must be specified to override the default of no flanking seq!\n");
    fprintf(stderr,"Using default values of [start,stop] = [%d,%d] ..\n",conf.start,conf.stop);
  }else{
    fprintf(stderr,"Using user-provided values of [start,stop] = [%d,%d] ..\n",conf.start,conf.stop);
  }
  if (conf.start < 1 || conf.start > seq->Length() ||
      conf.stop < 1 || conf.stop > seq->Length() ||
      conf.start >= conf.stop) {
    fprintf(stderr,"[start,stop] = [%d,%d] are invalid!\nExiting ..\n",conf.start,conf.stop);
    exit(EXIT_FAILURE);
  }
  
  // Initialize random numbers:
  gsl_rng_env_setup();
  
  const gsl_rng_type *T = gsl_rng_default;
  gsl_rng *r = gsl_rng_alloc(T);
  
  // Report siman settings, initial seq and its energy:
  printf("*********** Starting simulated annealing: ***********\n");
  printf("runmode = %s\n",runmode.c_str() );
  printf("N_TRIES = %d, ITERS_FIXED_T = %d, STEP_SIZE = %d, K = %g, T_INITIAL = %g, MU_T = %g, T_MIN = %g\n",
	 N_TRIES,ITERS_FIXED_T,STEP_SIZE,K,T_INITIAL,MU_T,T_MIN);
  // The number of T steps can be computed analytically using: N_T = log(T/T_{min}) / log(\mu).
  // Then the total number of energy evaluations is: N_T * ITERS_FIXED_T
  int NT = (int) (log(T_INITIAL/T_MIN) / log(MU_T) + 1);
  printf("N_T_steps = %d, total_N_steps = %d\n",NT,NT*ITERS_FIXED_T);
  printf("*****************************************************\n");
  printf("Initial seq: E = %.3f\n",E1(&conf) );
  conf.ss->Print(0,conf.ss->Length()-1); // print initial seq
  
  // Run simulated annealing:
  if (runmode == "min") { // minimize the energy
    //gsl_siman_solve(r, &conf, E1, S1, M1, P1, NULL, NULL, NULL, sizeof(SA_params), SArun_params);
    gsl_siman_solve(r, &conf, E1, S1, M1, P1, copy1, copy_construct1, destroy1, 0, SArun_params);
    printf("Highest affinity seq: E = %.3f\n",E1(&conf) );
  }else if (runmode == "max") { // maximize the energy
    //gsl_siman_solve(r, &conf, E2, S1, M1, P1, NULL, NULL, NULL, sizeof(SA_params), SArun_params);
    gsl_siman_solve(r, &conf, E2, S1, M1, P1, copy1, copy_construct1, destroy1, 0, SArun_params);
    printf("Lowest affinity seq: E = %.3f\n",E1(&conf) );
  }
  conf.ss->Print(0,conf.ss->Length()-1); // print final seq
}

// Return the <average> energy of a configuration (stored in xp)
double E1(void *xp) {
  SA_params *conf = (SA_params *) xp;
  Sequence *curseq = conf->ss;
  //Options *curopt = conf->op;
  DNAparams *DNAprms = conf->DNA;
  Geom *ideal_frm = conf->ideal;
  Geom *wt_frm = conf->wt;
  int seqL = conf->stop - conf->start + 1; // length of the variable subseq
  char *subseq_str = new char[seqL+1];
  long double *Ewo_all = new long double[curseq->Length() - seqL + 1]; // all energies along the seq
  long double Ewo_ave = 0.0; // average energy over the whole seq (which is >= variable subseq)
  int Ecnt = 0; // energy computation counter
  for (int j=0; j<curseq->Length() - seqL + 1; ++j) {
    // Make a subsequence:
    for (int k=0; k<seqL; ++k) {
      subseq_str[k] = curseq->CharAt(j+k);
    }
    subseq_str[seqL] = '\0';
    Sequence *subseq = new Sequence(subseq_str, seqL, "dummy");
    Minimizer min(subseq, *DNAprms, *ideal_frm, *wt_frm, conf->op, "none", "none", "silent");
    Ewo_all[j] = min.getEwo();
    Ewo_ave += Ewo_all[j];
    Ecnt++;
    delete subseq;
  }
  if (Ecnt > 0) Ewo_ave /= Ecnt;
  // Compute free energy:
  long double Z = 0.0;
  for (int j=0; j<curseq->Length() - seqL + 1; ++j) {
    Z += expl(-(Ewo_all[j] - Ewo_ave) ); // note that RT = 1
  }
  delete [] subseq_str;
  delete [] Ewo_all;
  return (double) (Ewo_ave - logl(Z) );
}

// Return "minus" the <average> energy of a configuration (stored in xp)
double E2(void *xp) {
  double En = E1(xp);
  return -En;
}

// Modify current configuration using random steps taken from the random number generator 
void S1(const gsl_rng *r, void *xp, double step_size) {
  SA_params *conf = (SA_params *) xp;
  Sequence *curseq = conf->ss;
  Options *curopt = conf->op;
  int i1;
  //int i2;
  int seqL;
  if (curopt->mf==true) {
    seqL = curseq->Length(); // all seq mutated
    i1 = 0; // 0-based beginning
    //i2 = seqL - 1; // 0-based end
  }else{
    seqL = conf->stop - conf->start + 1; // length of the variable subseq; only subseq mutated
    i1 = conf->start - 1; // 0-based beginning
    //i2 = conf->stop - 1; // 0-based end
  }
  
  int ulimit = (int) fabs(step_size); // this is always = STEP_SIZE
  //debug
  //printf("ulimit = %d\n",ulimit);
  for (int i=0; i<=ulimit; ++i) { // step_size determines how many muts we make in the current seq
    double u1 = gsl_rng_uniform(r) * seqL * 10; // two random numbers
    double u2 = gsl_rng_uniform(r) * 4 * 10;
    
    int nt_ind = (int) fabs(((int) u1)%seqL) + i1;
    char nt_type = (int) fabs(((int) u2)%4);
    
    curseq->Mutate(nt_ind,nt_type); // this makes a point mut in seq pointed to within conf
    //debug
    //printf("%f %f\n",u1,u2);
    //printf("%d %d\n",nt_ind,nt_type);
    //curseq->Print(0,curseq->Length()-1);
  }
}

// Return the distance between two configurations xp and yp
// (in this case, a number of mutations between two seqs).
// Currently (gsl-1.7) this function is not used at all by the GSL simulated
// annealing routine I am using.
double M1(void *xp, void *yp) {
  SA_params *confx = (SA_params *) xp;
  SA_params *confy = (SA_params *) yp;
  
  Sequence *s1 = confx->ss;
  Sequence *s2 = confy->ss;
  
  assert(s1->Length()==s2->Length() );
  
  int Hdist = 0; // Hamming distance
  for (int i=0; i<s1->Length(); ++i) {
    if (s1->IndexOfCharAt(i) != s2->IndexOfCharAt(i) ) ++Hdist;
  }
  
  return (double) Hdist;
}

// Print the contents of the configuration xp (ie print current seq)
void P1(void *xp) {
  SA_params *conf = (SA_params *) xp;
  Sequence *curs = conf->ss;
  char *seqstring = curs->SeqString();
  printf("   %s ",seqstring);
  //curs->Print(0,curs->Length()-1);
  delete [] seqstring;
}

// The three functions below are constructor, destructor and copyfunc that
// explicitly manipulate contents of SA_params (rather than relying on internal
// siman.c machinery such as memcpy).

// This function type should copy the configuration source into dest:
void copy1 (void *source, void *dest) {
  SA_params *conf_source = (SA_params *) source;
  SA_params *conf_dest = (SA_params *) dest;
  // Copy pointers to all objects that do not change during the run:
  conf_dest->DNA = conf_source->DNA;
  conf_dest->ideal = conf_source->ideal;
  conf_dest->wt = conf_source->wt;
  conf_dest->op = conf_source->op;
  // Copy [start,stop] coords:
  conf_dest->start = conf_source->start;
  conf_dest->stop = conf_source->stop;
  // Create a copy of the sequence stored in source:
  Sequence *seq_src = conf_source->ss;
  conf_dest->ss = new Sequence(*seq_src);
}

// This function type should create a new copy of the configuration xp:
void *copy_construct1 (void *xp) {
  SA_params *yp = new SA_params;
  SA_params *conf_source = (SA_params *) xp;
  // Copy pointers to all objects that do not change during the run:
  yp->DNA = conf_source->DNA;
  yp->ideal = conf_source->ideal;
  yp->wt = conf_source->wt;
  yp->op = conf_source->op;
  // Copy [start,stop] coords:
  yp->start = conf_source->start;
  yp->stop = conf_source->stop;
  // Create a copy of the sequence stored in conf_source:
  Sequence *seq_conf_source = conf_source->ss;
  yp->ss = new Sequence(*seq_conf_source);
  
  return yp;
}

// This function type should destroy the configuration xp, freeing its memory:
void destroy1 (void *xp) {
  SA_params *conf_source = (SA_params *) xp;
  delete conf_source->ss; // delete seq
  delete conf_source; // delete pointers to constant objects
}

