/* bstable.c */

/*****************************************************************
        Copyright by Rockefeller University,
can not be reproduced or distributed without written permission of
the copyright holder.  Version 1.1 of April 2006.

Written by Alexandre Morozov (contact person: morozov@edsb.rockefeller.edu),
and Eric Siggia.

The program DNABEND implements an algorithm for computing elastic energies
for DNA molecules of arbitrary length and sequence. DNA can be constrained
to follow an arbitrary 3D curve, such as a left-handed superhelix found
in nucleosomal core particles.

The DNABEND method and its applications are described in the following
paper:

******************************************************************/

#include "bstable.h"

BStable::BStable(void) {
  params = NULL;
  dinuc_labels = NULL;
  dinuc_nums = NULL;
}

// NOTE: all angles in BStable are in deg NOT rad!!
BStable::BStable(std::string const & filename) {
  ReadBStable(filename);
}

// Construct WO prm table from the solution to the system of
// linear eqs + the seq which was used to construct this system.
// NOTE: all angles in BStable are in deg NOT rad!!
BStable::BStable(gsl_vector *eq, Sequence::Sequence *seq) {
  unsigned int xlength = eq->size;
  if (xlength%Nprm != 0) { // the table is not a multiple of 6 - unclear how to proceed
    std::cerr << "Unexpected dims of the solution to system of linear eqs - not a multiple of Nprm=6!\nExiting ..\n";
    exit(EXIT_FAILURE);
  }
  Lseq = xlength/Nprm; // length of seq in **dinucs**
  if ((unsigned int) seq->Length() != Lseq + 1) {
    std::cerr << "Mismatch between seq length and dims of the system of linear eqs!\nExiting ..\n";
    exit(EXIT_FAILURE);
  }
  // Everything checks out, so allocate memory for the table and load its values from x:
  dinuc_nums = new int[Lseq];
  params = new double*[Lseq];
  dinuc_labels = new char*[Lseq];
  for (unsigned int i=0; i<Lseq; i++) {
    params[i] = new double[Nprm];
    dinuc_labels[i] = new char[3];
  }
  for (unsigned int i=0; i<Lseq; ++i) {
    dinuc_nums[i] = i+1;
    char ch1 = seq->CharAt(i);
    char ch2 = seq->CharAt(i+1);
    dinuc_labels[i][0] = ch1;
    dinuc_labels[i][1] = ch2;
    dinuc_labels[i][2] = '\0'; // make sure C string terminates normally
    for (unsigned int j=0; j<Nprm; ++j) {
      int linind = i*Nprm + j;
      params[i][j] = gsl_vector_get(eq,linind); // [ Twist(deg)  Roll(deg)  Tilt(deg)  Shift(A)  Slide(A)  Rise(A) ]
    }
  }
}

// Construct WO prm table from the default set of WO prms for a given seq.
// NOTE: all angles in BStable are in deg NOT rad!!
BStable::BStable(DNAparams::DNAparams &prms, Sequence::Sequence *seq) {
  Lseq = seq->Length() - 1;
  // Allocate memory for the table and lookup its defaut values:
  dinuc_nums = new int[Lseq];
  params = new double*[Lseq];
  dinuc_labels = new char*[Lseq];
  for (unsigned int i=0; i<Lseq; i++) {
    params[i] = new double[Nprm];
    dinuc_labels[i] = new char[3];
  }
  for (unsigned int i=0; i<Lseq; ++i) {
    dinuc_nums[i] = i+1;
    char ch1 = seq->CharAt(i);
    char ch2 = seq->CharAt(i+1);
    dinuc_labels[i][0] = ch1;
    dinuc_labels[i][1] = ch2;
    dinuc_labels[i][2] = '\0'; // make sure C string terminates normally
    int ind = get_basestep_index(ch1,ch2);
    for (unsigned int j=0; j<Nprm; ++j) {
      params[i][j] = prms.means_bs[ind][j];
    }
  }
}

BStable::~BStable() {
  delete [] dinuc_nums;
  for (unsigned int i=0; i<Lseq; i++) {
    delete [] dinuc_labels[i];
    delete [] params[i];
  }
  delete [] dinuc_labels;
  delete [] params;
}

void BStable::ReadBStable(std::string const & filename) {
  using std::cout;
  using std::cerr;
  using std::endl;
  using std::ifstream;
  
  ifstream file(filename.c_str() );
  
  char buffer[512];
  
  if (file.fail()) {
    cerr << "Could not open file: "<<filename<<" for reading!\nExiting..\n";
    exit(EXIT_FAILURE);
  }
  
  // Count basesteps in file:
  int cnt = 0; // basestep counter
  double tmp[8];
  char tmp_string[3];
  int tmp_int;
  while (!file.eof()) {
    file.getline(buffer,512,'\n');
    if (file.fail() && !file.eof() ) file.clear();
    if (sscanf(buffer,"%d %s %lf %lf %lf %lf %lf %lf",
	       &tmp_int,tmp_string,&tmp[0],&tmp[1],&tmp[2],&tmp[3],&tmp[4],&tmp[5]) == 8) {
      ++cnt;
    }
  }
  file.clear(); // forget we hit the end of file
  file.seekg(0, std::ios::beg); // move to the start of the file
  
  // Allocate arrays:
  Lseq = cnt;
  dinuc_nums = new int[Lseq];
  params = new double*[Lseq];
  dinuc_labels = new char*[Lseq];
  for (unsigned int i=0; i<Lseq; i++) {
    params[i] = new double[Nprm];
    dinuc_labels[i] = new char[3];
  }
  //debug
  //printf("Lseq = %d\n",Lseq);
  
  cnt = -1; // reset basestep counter
  while (!file.eof()) {
    file.getline(buffer,512,'\n');
    if (file.fail() && !file.eof() ) file.clear();
    if (cnt==-1) {
      ++cnt;
      continue; // skip header
    }
    if (sscanf(buffer,"%d %s %lf %lf %lf %lf %lf %lf",&dinuc_nums[cnt],dinuc_labels[cnt],
	       &params[cnt][0],&params[cnt][1],&params[cnt][2],&params[cnt][3],&params[cnt][4],&params[cnt][5]) == 8) {
      dinuc_labels[cnt][2] = '\0'; // make sure C string terminates normally 
      ++cnt;
    }
  }
  file.close();
  /*
  //debug
  for (int i=0; i<Lseq; ++i) {
    for (int j=0; j<Nprm; ++j) {
      printf("%6.3f ",params[i][j]);
    }
    printf("\n");
  }
  */
}

void BStable::WriteBStable(std::string const & filename) {
  FILE *fp = fopen(filename.c_str(),"w");
  if (fp == NULL) {
    printf("Error reading file %s\n",filename.c_str() );
    exit(1);
  }
  //debug
  //printf("Lseq = %d\n",Lseq);
  fprintf(fp," res name   twist   roll   tilt   shift   slide   rise\n"); // header
  for (unsigned int i=0; i<Lseq; ++i) {
    fprintf(fp,"%4d %2s   ",dinuc_nums[i],dinuc_labels[i]);
    for (unsigned int j=0; j<Nprm; ++j) {
      fprintf(fp,"%9.5f  ",params[i][j]);
    }
    fprintf(fp,"\n");
  }
  fclose(fp);
}

const unsigned int BStable::length(void) {
  return Lseq;
}

const double BStable::get(int i, int j) {
  return params[i][j];
}

gsl_matrix* BStable::Export(void) {
  gsl_matrix *m = gsl_matrix_calloc(Lseq,Nprm);
  for (unsigned int i=0; i<Lseq; ++i) {
    for (unsigned int j=0; j<Nprm; ++j) {
      gsl_matrix_set(m,i,j,params[i][j]);
    }
  }
  return m;
}

// This aux. function is a lookup table for dinuc indices in the table.
int get_basestep_index(char l1, char l2) {
  
  int ind = 0;
  
  if (l1=='A') {
    ind = 0;
  }else if (l1=='C') {
    ind = 4;
  }else if (l1=='G') {
    ind = 8;
  }else if (l1=='T') {
    ind = 12;
  }else{
    std::cerr << "Unknown first DNA base label: "<<l1<<"!\nExiting ..\n";
    exit(EXIT_FAILURE);
  }
  
  if (l2=='A') {
    return (ind+0);
  }else if (l2=='C') {
    return (ind+1);
  }else if (l2=='G') {
    return (ind+2);
  }else if (l2=='T') {
    return (ind+3);
  }else{
    std::cerr << "Unknown second DNA base label: "<<l2<<"!\nExiting ..\n";
    exit(EXIT_FAILURE);
  }
  
  return -1;
}

