/* edesign.c */

/*****************************************************************
        Copyright by Rockefeller University,
can not be reproduced or distributed without written permission of
the copyright holder.  Version 1.1 of April 2006.

Written by Alexandre Morozov (contact person: morozov@edsb.rockefeller.edu),
and Eric Siggia.

The program DNABEND implements an algorithm for computing elastic energies
for DNA molecules of arbitrary length and sequence. DNA can be constrained
to follow an arbitrary 3D curve, such as a left-handed superhelix found
in nucleosomal core particles.

The DNABEND method and its applications are described in the following
paper:

******************************************************************/

#include "options.h"
#include "fastafile.h"
#include "sequence.h"
#include "equations.h"
#include "dna_parameters.h"
#include "bstable.h"
#include "xyzframe.h"
#include "anneal.h"

// This main function does seq design by simulated annealing. Initial seq (usually 71 bp) has to be chosen
// by the user. Both high and low affinity seqs can be created.

int main(int argc, char **argv) {
  // (A) ==== Read in command-line options: ==== //
  if (argc < 2) {
    fprintf(stderr,"usage: %s [-fasta <Fasta file of *single* input seq>] [-geom_wt <WOprms_wt.dat>] [-geom_id <WOprms_ideal.dat>] [-PDBid_bp <ideal_helix_fitted.long.pdb>] [-PDBid_mst <ideal_helix_mst_fitted.long.pdb>] [-curv <curvature_ideal.dat>] [-fpdb <pdblist.dat>] [-fgeom <geomlist.dat>] [-w_el <elastic energy weight>] [-w_sh <superhelical constraint weight>] [-w_sum <sum of angles constraint weight>] [-w_dev <deviations from ideal frames constraint weight>] [-compute_curv] [-minmax <min/max>] [-start <mutated_region_start>] [-stop <mutated_region_end>] [-mf]\n",argv[0]);
    exit(1);
  }
  int argbase = 1;
  struct Options *opt = ReadArguments(argbase, argc, argv);
  // (B) ==== Read DNA parameters (moments + force constants): ==== //
  DNAparams DNAprms;
  // (C) ==== Compute & store a set of xyz frames corresponding to the wt (i.e. pdb) sequence: ==== //
  // Table of wt WO prms (read in from file or obtained thru solving Ax=b for an input seq): 
  BStable bst_wt(opt->geom_wt);
  //bst_wt.WriteBStable("WOprms_wt.dat.copy");
  //gsl_matrix *m = bst_wt.Export();
  // Set of wt frames:
  Geom xyz_wt(bst_wt);
  //xyz_wt.PDBunitvecs("WOprms_wt.long.pdb",'l');
  //xyz_wt.PDB("junk_wt.pdb");
  //exit(0);
  // (D) ==== Read in & store a set of ideal superhelical frames superimposed onto the wt conformation: ==== //
  Geom *xyz_ideal_ptr;
  if (opt->geom_id != "none") { // user provided ideal BStable: use it rather than the 2 PDB files
    std::cerr << "Using a BStable: "<< opt->geom_id <<" to construct the ideal curve ..\n";
    BStable bst_ideal(opt->geom_id);
    xyz_ideal_ptr = new Geom(bst_ideal);
    //xyz_ideal_ptr = new Geom(bst_wt);
  }else{ // either default or user-provided PDB files with bp and MST frames
    std::cerr << "Using PDB files: ("<< opt->PDBid_bp <<","<< opt->PDBid_mst <<") to construct the ideal curve ..\n";
    xyz_ideal_ptr = new Geom(opt->PDBid_bp,opt->PDBid_mst);
  }
  Geom xyz_ideal = *xyz_ideal_ptr;
  //xyz_ideal.PDB("junk_ideal.pdb");
  //exit(0);
  // Compute & write derivs & curvatures for the ideal frames if requested:
  if (opt->compute_curv) {
    std::cerr << "Computing derivatives and curvature coeffs for ideal frames ..\n";
    xyz_ideal.BuildDerivs();
    // debug: check derivs
    //Matrix3 D = xyz_ideal.NumDeriv(140, 10, 1, 1.0);
    //D.Print();
    //Matrix3 D2 = xyz_ideal.GetDeriv(140, 10, 1);
    //D2.Print();
    //exit(0);
    xyz_ideal.ComputeCurv();
    xyz_ideal.WriteCurv(opt->curv);
    std::cerr << "Computed curvature coeffs and saved them in: "<<opt->curv<<" ..\n";
    exit(EXIT_SUCCESS);
  }else{
    xyz_ideal.ReadCurv(opt->curv);
  }
  // =========================================================== //
  // (E) ==== Read in the initial sequence from input file: ==== //
  char *fastafile = opt->fasta;
  FastaFile sequences;
  sequences.ReadFasta(fastafile);
  if (sequences.Size() != 1) {
    std::cerr << "Expected a single starting seq but read "<<sequences.Size()<<" seq(s) from file: "<<fastafile<<"!\nExiting ..\n";
    exit(EXIT_FAILURE);
  }
  Sequence *seq = sequences[0];
  std::cout << "Initial seq of length L = "<<seq->Length()<<":"<<std::endl;
  seq->Print(0,seq->Length()-1);
  // (F) ==== Design a seq with min (max) DNA energy by simulated annealing, starting from the initial seq: ==== //
  if (opt->fpdb != "none") {
    std::cerr << "List of PDB files is irrelevant in seq design mode - ignoring ..\n";
  }
  if (opt->fgeom != "none") {
    std::cerr << "List of geom files is irrelevant in seq design mode - ignoring ..\n";
  }
  Anneal(seq, DNAprms, xyz_ideal, xyz_wt, opt);
  //gsl_matrix_free(m);
  delete opt;
}

