/* equations.c */

/*****************************************************************
        Copyright by Rockefeller University,
can not be reproduced or distributed without written permission of
the copyright holder.  Version 1.1 of April 2006.

Written by Alexandre Morozov (contact person: morozov@edsb.rockefeller.edu),
and Eric Siggia.

The program DNABEND implements an algorithm for computing elastic energies
for DNA molecules of arbitrary length and sequence. DNA can be constrained
to follow an arbitrary 3D curve, such as a left-handed superhelix found
in nucleosomal core particles.

The DNABEND method and its applications are described in the following
paper:

******************************************************************/

#include "equations.h"

// This is Equations class which sets up A and b in the system of linear algebraic
// equations: Ax = b. It takes a Sequence (usually read in from input FASTA file)
// as input. The linear system is solved, and x is stored for subsequent analysis.
// NOTE: all angles are stored in degrees here!!

Equations::Equations(Sequence::Sequence *seq, DNAparams::DNAparams &prms) {
  // Set dimensions:
  Lseq = seq->Length()-1; // seq length **!! in dinucs !!**
  DIM = Lseq * Nprm;
  
  // Allocate memory:
  A = gsl_matrix_calloc(DIM,DIM);
  b = gsl_vector_calloc(DIM);
  x = gsl_vector_calloc(DIM);
  curv = gsl_vector_calloc(Lseq);
  derivs = NULL;
  
  // Set up coeffs for WO potential:
  setWOcoeff(seq,prms);
  
  /*
  //debug
  printf("======= Printing b =======\n");
  gsl_vector_fprintf(stdout, b, "%6.3f");
  printf("======= Printing A =======\n");
  //gsl_matrix_fprintf(stdout, A, "%6.3f"); // writes it all in one column
  for (int i=0; i<DIM; ++i) {
    for (int j=0; j<DIM; ++j) {
      printf("%6.3f ",gsl_matrix_get(A,i,j) );
    }
    printf("\n");
  }
  */
  
  // Add curvature constraint:
  setCurvConstr();
  
  // Add twist constraint (total sum of twist angles is constrained):
  setTwistConstr();
  
  /*
  // Add curvature constraint:
  double curv_constr[DIM][DIM];
  for (int m1=0; m1<DIM; ++m1) // init array
    for (int m2=0; m2<DIM; ++m2)
      curv_constr[m1][m2] = 0.0;
  */
  /*
  //debug
  printf("======= Printing b =======\n");
  gsl_vector_fprintf(stdout, b, "%6.3f");
  printf("======= Printing A =======\n");
  //gsl_matrix_fprintf(stdout, A, "%6.3f"); // writes it all in one column
  for (int i=0; i<DIM; ++i) {
    for (int j=0; j<DIM; ++j) {
      printf("%6.3f ",gsl_matrix_get(A,i,j) );
    }
    printf("\n");
  }
  */
}

// This constructor uses a coordinate transformation from local to global displacements, in order to make
// setting superhelical constraints easier.
// "ideal" set of frames: fixed set of unit vecs used to carry out the coordinate transformation.
// "init" set of frames: used as a starting point (must not be too far from the superhelix!).
Equations::Equations(Sequence::Sequence *seq, DNAparams::DNAparams &prms, Geom::Geom &ideal, Geom::Geom &init) {
  // Set dimensions:
  Lseq = seq->Length()-1; // seq length **!! in dinucs !!**
  DIM = Lseq * Nprm;
  
  // Allocate memory:
  A = gsl_matrix_calloc(DIM,DIM);
  b = gsl_vector_calloc(DIM);
  x = gsl_vector_calloc(DIM);
  curv = gsl_vector_calloc(Lseq);
  derivs = NULL;
  
  // Set up coeffs for *global* WO potential:
  setWOcoeff_global(seq, prms, ideal, init);
  // Set up penalty for deviations from the superhelilcal path:
  setDisplacementConstr(ideal.frames);
}

// == OUTDATED == //
// Mode determines whether constraints around native values of WO prms or xyz constraints in global
// frame are to be imposed. 
Equations::Equations(Sequence::Sequence *seq, DNAparams::DNAparams &prms, gsl_matrix *bst, const std::string mode) {
  // Set dimensions:
  Lseq = seq->Length()-1; // seq length **!! in dinucs !!**
  DIM = Lseq * Nprm;
  
  // Check dimensions:
  if (Lseq != bst->size1) {
    std::cerr << "Mismatch between Sequence length and BStable length!\nExiting ..\n";
    exit(EXIT_FAILURE);
  }
  
  // Allocate memory:
  A = gsl_matrix_calloc(DIM,DIM);
  b = gsl_vector_calloc(DIM);
  x = gsl_vector_calloc(DIM);
  curv = gsl_vector_calloc(Lseq);
  derivs = NULL;
  
  // Set up coeffs for WO potential:
  setWOcoeff(seq,prms);
  
  // Mode-dependent constraints:
  if (mode == "bstable") {
    // Add a constraint based on the input BStable (ie a set of WO prms from Xtal struc):
    setBStableConstr(bst);
  }else if (mode == "xyzspace") {
    // Read the derivs table from file:
    ReadDerivs("derivs.dat");
    // Add a constraint in global xyz space:
    setGlobalConstr(bst);
  }else{
    std::cerr << "Unknown mode: "<<mode<<" in Equations constructor!\nExiting ..\n";
    exit(EXIT_FAILURE);
  }
}

Equations::Equations() {
  A = NULL;
  b = NULL;
  x = NULL;
  curv = NULL;
  derivs = NULL;  
}

Equations::~Equations() {
  if (A != NULL) gsl_matrix_free(A);
  if (b != NULL) gsl_vector_free(b);
  if (x != NULL) gsl_vector_free(x);
  if (curv != NULL) gsl_vector_free(curv);
  if (derivs != NULL) {
    for (unsigned int j=0; j<3; ++j) {
      gsl_matrix_free(derivs[j]);
    }
    delete [] derivs;
  }
}

// This function computes WO energy for a given solution of Ax = b.
double Equations::Energy(Sequence::Sequence *seq, DNAparams::DNAparams &prms) {
  double E = 0.0; // total energy of the conformation
  
  for (unsigned int i=0; i<Lseq; ++i) {
    char ch1 = seq->CharAt(i);
    char ch2 = seq->CharAt(i+1);
    int ind = get_basestep_index(ch1,ch2);
    for (unsigned int j1=0; j1<Nprm; ++j1) {
      int linind1 = i*Nprm + j1;
      for (unsigned int j2=0; j2<Nprm; ++j2) {
	int linind2 = i*Nprm + j2;
	E += F0 * 0.5 * prms.Fij[ind][j1][j2] * (getx(linind1) - prms.means_bs[ind][j1]) * (getx(linind2) - prms.means_bs[ind][j2]);
      }
    }
  }
  
  return E;
}

// This function adds WO prms to Ax = b.
void Equations::setWOcoeff(Sequence::Sequence *seq, DNAparams::DNAparams &prms) {
  // Set up coeffs for WO potential:
  for (unsigned int i=0; i<Lseq; ++i) {
    char ch1 = seq->CharAt(i);
    char ch2 = seq->CharAt(i+1);
    int ind = get_basestep_index(ch1,ch2);
    for (unsigned int j1=0; j1<Nprm; ++j1) {
      int linind1 = i*Nprm + j1;
      double tmp = 0.0;
      for (unsigned int j2=0; j2<Nprm; ++j2) {
	int linind2 = i*Nprm + j2;
	double Aold = gsl_matrix_get(A,linind1,linind2);
	gsl_matrix_set(A,linind1,linind2,F0*prms.Fij[ind][j1][j2] + Aold);
	tmp += F0 * prms.Fij[ind][j1][j2] * prms.means_bs[ind][j2];
      }
      double bold = gsl_vector_get(b,linind1);
      gsl_vector_set(b,linind1,tmp + bold);
    }
  }
  /*
  //debug
  //printf("======= Printing b =======\n");
  //gsl_vector_fprintf(stdout, b, "%6.3f");
  printf("======= Printing A =======\n");
  //gsl_matrix_fprintf(stdout, A, "%6.3f"); // writes it all in one column
  for (unsigned int i=0; i<DIM; ++i) {
    printf("i=%3d: ",i);
    for (unsigned int j=0; j<DIM; ++j) {
      if (gsl_matrix_get(A,i,j) == 0.0) continue;
      printf("%8.5f ",gsl_matrix_get(A,i,j) );
    }
    printf("\n");
  }
  */
}

// This function adds transformed WO prms to Ax = b. Two sets of frames are passed in:
// a) for ideal superhelix; b) for "straight" DNA corresponding to input (average) WO prms.
void Equations::setWOcoeff_global(Sequence::Sequence *seq, DNAparams::DNAparams &prms,
				  Geom::Geom &ideal, Geom::Geom &init) {
  // Check input frames:
  if (ideal.frames_mst.size() != init.frames_mst.size() || ideal.frames.size() != init.frames.size() ||
      ideal.frames.size()-1 != ideal.frames_mst.size() || init.frames.size()-1 != init.frames_mst.size() ) {
    std::cerr << "ERROR in setWOcoeff_global: mismatch in WO/ideal input frames!Exiting ..\n";
    exit(EXIT_FAILURE);
  }
  // Set up *global* coeffs for WO potential using a set of ideal MST frames:
  for (unsigned int i=0; i<Lseq; ++i) {
    char ch1 = seq->CharAt(i);
    char ch2 = seq->CharAt(i+1);
    //debug
    //printf("i=%d %c%c\n",i,ch1,ch2);
    int ind = get_basestep_index(ch1,ch2);
    // Construct 4 3x3 matrices with original WO coeffs & then transform them into the global frame:
    Matrix3 AA,AD,DA,DD; // angle-angle, angle-diplacement, displacement-angle etc.
    for (unsigned int j1=0; j1<3; ++j1) {
      for (unsigned int j2=0; j2<3; ++j2) {
	AA.set(j1, j2, F0*prms.Fij[ind][j1][j2]);
      }
    }
    for (unsigned int j1=0; j1<3; ++j1) {
      for (unsigned int j2=3; j2<6; ++j2) {
	AD.set(j1, j2-3, F0*prms.Fij[ind][j1][j2]);
      }
    }
    for (unsigned int j1=3; j1<6; ++j1) {
      for (unsigned int j2=0; j2<3; ++j2) {
	DA.set(j1-3, j2, F0*prms.Fij[ind][j1][j2]);
      }
    }
    for (unsigned int j1=3; j1<6; ++j1) {
      for (unsigned int j2=3; j2<6; ++j2) {
	DD.set(j1-3, j2-3, F0*prms.Fij[ind][j1][j2]);
      }
    }
    // MST frames:
    Matrix3 R_DA(ideal.frames_mst.at(i).a, ideal.frames_mst.at(i).b, ideal.frames_mst.at(i).c);
    Matrix3 R_DD(R_DA),Rt(R_DA),R_WO(R_DA);
    Rt.transpose();
    // Matrix multiplication:
    AD.RegMultiply(Rt); // FR^{-1}
    R_DA.RegMultiply(DA); // RF
    DD.RegMultiply(Rt);
    R_DD.RegMultiply(DD); // RFR^{-1}
    //debug
    //AD.Print();
    //exit(0);
    // find RHS in the global frame:
    Vector v0 = init.frames.at(i+1).orig - init.frames.at(i).orig; // global displacements from the initial conformation
    //Vector WO(prms.means_bs[ind][3],prms.means_bs[ind][4],prms.means_bs[ind][5]); // vector of local WO prms
    //Vector v0 = R_WO.VecMultiply(WO);
    //Vector WO(init.getm(ind,3),init.getm(ind,4),init.getm(ind,5) );
    //Vector v0 = R_WO.VecMultiply(WO);
    //debug
    //printf("%d: v0 = [%.2f  %.2f  %.2f]\n",i+1,v0.x,v0.y,v0.z);
    // Insert 4 matrices into A, and update the RHS:
    for (unsigned int j1=0; j1<Nprm; ++j1) {
      int linind1 = i*Nprm + j1;
      double tmp = 0.0;
      for (unsigned int j2=0; j2<Nprm; ++j2) {
	int linind2 = i*Nprm + j2;
	double coeff;
	if (j1<3 && j2<3) { // AA
	  coeff = AA.value(j1,j2);
	}else if (j1<3 && j2>=3) { // AD
	  coeff = AD.value(j1,j2-3);
	}else if (j1>=3 && j2<3) { // DA
	  coeff = R_DA.value(j1-3,j2);
	}else if (j1>=3 && j2>=3) { // DD
	  coeff = R_DD.value(j1-3,j2-3);
	}
	double Aold = gsl_matrix_get(A,linind1,linind2);
	gsl_matrix_set(A,linind1,linind2,coeff + Aold);
	if (j2<3) {
	  //tmp += coeff * prms.means_bs[ind][j2];
	  tmp += coeff * init.getm(i,j2);
	}else if (j2==3) {
	  tmp += coeff * v0.x;
	}else if (j2==4) {
	  tmp += coeff * v0.y;
	}else if (j2==5) {
	  tmp += coeff * v0.z;
	}
      }
      double bold = gsl_vector_get(b,linind1);
      gsl_vector_set(b,linind1,tmp + bold);
    }
  }
  /*
  //debug
  //printf("======= Printing b =======\n");
  //gsl_vector_fprintf(stdout, b, "%6.3f");
  printf("======= Printing A =======\n");
  //gsl_matrix_fprintf(stdout, A, "%6.3f"); // writes it all in one column
  for (unsigned int i=0; i<30; ++i) {
    printf("i=%3d: ",i);
    for (unsigned int j=0; j<30; ++j) {
      //if (gsl_matrix_get(A,i,j) == 0.0) continue;
      printf("%8.5f ",gsl_matrix_get(A,i,j) );
    }
    printf("\n");
  }
  */
}

// This function is mostly for testing - it checks if the energy stayed the same after
// rotation into the global frame.
// ideal: frames to use in the rotation (no WO prms); init: frames (+WO prms) for which the energy is
// computed; WO: straight DNA (+WO prms)
double Equations::Energy_global(Sequence::Sequence *seq, DNAparams::DNAparams &prms,
				Geom::Geom &ideal, Geom::Geom &WO, Geom::Geom &init) {
  double E1 = 0.0, E2 = 0.0; // total energies: E1 => WO; E2 => global constraint
  
  // Check input frames:
  if (ideal.frames_mst.size() != init.frames_mst.size() || ideal.frames.size() != init.frames.size() ||
      ideal.frames_mst.size() != WO.frames_mst.size() || ideal.frames.size() != WO.frames.size() ||
      ideal.frames.size()-1 != ideal.frames_mst.size() || init.frames.size()-1 != init.frames_mst.size() ||
      WO.frames.size()-1 != WO.frames_mst.size() ) {
    std::cerr << "ERROR in setWOcoeff_global: mismatch in WO/ideal input frames!Exiting ..\n";
    exit(EXIT_FAILURE);
  }
  
  // Compute energy in the **global** frame:
  for (unsigned int i=0; i<Lseq; ++i) {
    char ch1 = seq->CharAt(i);
    char ch2 = seq->CharAt(i+1);
    int ind = get_basestep_index(ch1,ch2);
    // Construct 4 3x3 matrices with original WO coeffs & then transform them into the global frame:
    Matrix3 AA,AD,DA,DD; // angle-angle, angle-diplacement, displacement-angle etc.
    for (unsigned int j1=0; j1<3; ++j1) {
      for (unsigned int j2=0; j2<3; ++j2) {
	AA.set(j1, j2, F0*prms.Fij[ind][j1][j2]);
      }
    }
    for (unsigned int j1=0; j1<3; ++j1) {
      for (unsigned int j2=3; j2<6; ++j2) {
	AD.set(j1, j2-3, F0*prms.Fij[ind][j1][j2]);
      }
    }
    for (unsigned int j1=3; j1<6; ++j1) {
      for (unsigned int j2=0; j2<3; ++j2) {
	DA.set(j1-3, j2, F0*prms.Fij[ind][j1][j2]);
      }
    }
    for (unsigned int j1=3; j1<6; ++j1) {
      for (unsigned int j2=3; j2<6; ++j2) {
	DD.set(j1-3, j2-3, F0*prms.Fij[ind][j1][j2]);
      }
    }
    // MST frames:
    Matrix3 R_DA(ideal.frames_mst.at(i).a, ideal.frames_mst.at(i).b, ideal.frames_mst.at(i).c);
    Matrix3 R_DD(R_DA),Rt(R_DA),R_WO(R_DA);
    Rt.transpose();
    // Matrix multiplication:
    AD.RegMultiply(Rt); // FR^{-1}
    R_DA.RegMultiply(DA); // RF
    DD.RegMultiply(Rt);
    R_DD.RegMultiply(DD); // RFR^{-1}
    //debug
    //AD.Print();
    //exit(0);
    // find RHS in the global frame:
    //Vector v0 = init.frames.at(i+1).orig - init.frames.at(i).orig; // global displacements from the initial conformation
    Vector v0(init.getm(i,3),init.getm(i,4),init.getm(i,5) );
    v0 = R_WO.VecMultiply(v0);
    //Vector WO_global = WO.frames.at(i+1).orig - WO.frames.at(i).orig;
    Vector WO_local(WO.getm(i,3),WO.getm(i,4),WO.getm(i,5) ); // vector of local WO prms
    Vector WO_global = R_WO.VecMultiply(WO_local);
    v0 -= WO_global;
    // Compute global constraint E:
    Vector v_ideal = ideal.frames.at(i+1).orig - ideal.frames.at(i).orig;
    v_ideal -= (v0 + WO_global);
    double E2cur = F6 * (v_ideal * v_ideal);
    E2 += E2cur;
    // Compute "global" WO E:
    double E1cur = 0.0;
    for (unsigned int j1=0; j1<Nprm; ++j1) {
      // \delta \theta 1
      double delta_1,delta_2,coeff;
      if (j1<3) {
	//delta_1 = init.getm(i,j1) - prms.means_bs[ind][j1];
	delta_1 = init.getm(i,j1) - WO.getm(i,j1);
      }else if (j1==3) {
	delta_1 = v0.x;
      }else if (j1==4) {
	delta_1 = v0.y;
      }else if (j1==5) {
	delta_1 = v0.z;
      }
      for (unsigned int j2=0; j2<Nprm; ++j2) {
	// F_{12}
	if (j1<3 && j2<3) { // AA
	  coeff = AA.value(j1,j2);
	}else if (j1<3 && j2>=3) { // AD
	  coeff = AD.value(j1,j2-3);
	}else if (j1>=3 && j2<3) { // DA
	  coeff = R_DA.value(j1-3,j2);
	}else if (j1>=3 && j2>=3) { // DD
	  coeff = R_DD.value(j1-3,j2-3);
	}
	// \delta \theta 2
	if (j2<3) {
	  //delta_2 = init.getm(i,j2) - prms.means_bs[ind][j2];
	  delta_2 = init.getm(i,j2) - WO.getm(i,j2);
	}else if (j2==3) {
	  delta_2 = v0.x;
	}else if (j2==4) {
	  delta_2 = v0.y;
	}else if (j2==5) {
	  delta_2 = v0.z;
	}
	E1cur += 0.5 * coeff * delta_1 * delta_2;
      }
    }
    //debug
    printf("%d: E1 = %.3f; E2 = %.3f; |diff| = %.3f\n",i+1,E1cur,E2cur,sqrt(v_ideal*v_ideal) );
    E1 += E1cur;
  }
  //debug
  printf("Total: E1 = %.3f; E2 = %.3f\n",E1,E2);
  return E1 + E2;
}

// Given a set of WO angles + global displacements (typically from a solution of Ax=b after coord
// transform), go back to the standard set of 6 WO prms.
// init_frame is fixed and is used to build the rest of the structure.
// Here code from Geom is reused to compute shift, slide, rise.
void Equations::ReconstructWO(coord &init_frame) {
  std::vector<coord> frames,frames_mst;
  
  frames.push_back(init_frame); // save 1st frame
  
  for (unsigned int i=0; i<Lseq; ++i) {
    // Compute Gamma & phi:
    double twi = DegtoRad(getx(i*Nprm+0) ); // twist roll tilt (in rad!!)
    double r = DegtoRad(getx(i*Nprm+1) );
    double t = DegtoRad(getx(i*Nprm+2) );
    double Gamma = sqrt(r*r + t*t);
    double phi = atan(t/r);
    if (r < 0.0) phi += PI; // phi > PI/2 if r < 0
    // previous frame:
    std::vector<coord>::reverse_iterator p = frames.rbegin();
    // Construct [T_{i}^{g}]:
    Matrix3 Tig1(p->a,p->b,p->c); // vectors become matrix columns
    Matrix3 Tig2(p->a,p->b,p->c); // need 2 copies because we'll rotate them differently
    // Init. unit vector frame (in which rotations happen):
    coord funit;
    funit.a.set(1.0,0.0,0.0);
    funit.b.set(0.0,1.0,0.0);
    funit.c.set(0.0,0.0,1.0);
    funit.orig.set(0.0,0.0,0.0);
    // Get MST and T_{i+1}^{i} in the local frame:
    coord mst = RotateMat(funit, -0.5*twi + phi, 0.5*Gamma, -phi);
    coord Tnext = RotateMat(funit, -0.5*twi + phi, Gamma, -0.5*twi - phi);
    // Construct T_{MST} and T_{i+1}^{i} in the global frame:
    Matrix3 Tmst(mst.a,mst.b,mst.c); // vectors become matrix columns
    Matrix3 Tij(Tnext.a,Tnext.b,Tnext.c); // vectors become matrix columns
    Tig1.RegMultiply(Tij); // T_{i+1}^{i} in the global frame
    Tig2.RegMultiply(Tmst); // T_{MST} in the global frame
    Tig1.loadvecs(Tnext.a,Tnext.b,Tnext.c); // load matrix into 3 vectors by columns
    Tig2.loadvecs(mst.a,mst.b,mst.c); // load matrix into 3 vectors by columns
    // Global displacements:
    Vector xglob(getx(i*Nprm+3),getx(i*Nprm+4),getx(i*Nprm+5) );
    //debug
    //printf("%d: xglob = [%.2f  %.2f  %.2f]\n",i+1,xglob.x,xglob.y,xglob.z);
    //printf("%d: |xglob| = %.2f\n",i+1,xglob.Length() );
    //xglob += (frames_ideal.at(i+1).orig - frames_ideal.at(i).orig);
    setx(i*Nprm+3,xglob * mst.a); // shift
    setx(i*Nprm+4,xglob * mst.b); // slide
    setx(i*Nprm+5,xglob * mst.c); // rise
    // ================================ //
    Tnext.orig = p->orig + xglob;
    frames.push_back(Tnext);
    mst.orig = p->orig;
    frames_mst.push_back(mst); // note that MST origin is the same as that for frame T_{i}!
  }
}

// This function adds Bstable constraints to Ax = b.
void Equations::setBStableConstr(gsl_matrix *bst) {
  for (unsigned int i=0; i<Lseq; ++i) {
    for (unsigned int j=0; j<Nprm; ++j) {
      int linind = i*Nprm + j;
      double Aold = gsl_matrix_get(A,linind,linind);
      gsl_matrix_set(A,linind,linind,F4 + Aold);
      double bold = gsl_vector_get(b,linind);
      gsl_vector_set(b,linind,F4*gsl_matrix_get(bst,i,j) + bold);
    }
  }
}

// This function adds a constraint on global displacements.
// This function adds Bstable constraints to Ax = b.
void Equations::setDisplacementConstr(std::vector<coord> &frames_ideal) {
  for (unsigned int i=0; i<Lseq; ++i) {
    Vector xid = frames_ideal.at(i+1).orig - frames_ideal.at(i).orig;
    for (unsigned int j=0; j<Nprm; ++j) {
      int linind = i*Nprm + j;
      if (j >= 3) { // 3,4,5 are the displacements
	double Aold = gsl_matrix_get(A,linind,linind);
	gsl_matrix_set(A,linind,linind,2*F6 + Aold);
      }
      double bold = gsl_vector_get(b,linind);
      if (j==3) {
	gsl_vector_set(b,linind,2*F6*xid.x + bold);
      }else if (j==4) {
	gsl_vector_set(b,linind,2*F6*xid.y + bold);
      }else if (j==5) {
	gsl_vector_set(b,linind,2*F6*xid.z + bold);
      }
      /*
      if (j==0) { // twist
	double bold = gsl_vector_get(b,linind);
	gsl_vector_set(b,linind,2*F6*Omega_0 + bold);
      }
      if (j==1) { // roll
	double bold = gsl_vector_get(b,linind);
	gsl_vector_set(b,linind,2*F6*Roll_0 + bold);
      }
      if (j==2) { // tilt
	double bold = gsl_vector_get(b,linind);
	gsl_vector_set(b,linind,2*F6*Tilt_0 + bold);
      }
      */
    }
  }
}

/*
// This function adds global "xyz" constraints to Ax = b.
void Equations::setGlobalConstr(gsl_matrix *bst) {
  gsl_matrix *con = gsl_matrix_calloc(3,6);
  
  for (unsigned int i=0; i<Lseq; ++i) {
    GetXYZMatrix(bst,con,i); // get a 3x6 matrix of Jacobian coeffs
    for (unsigned int j1=0; j1<Nprm; ++j1) {
      int linind1 = i*Nprm + j1;
      double tmp = 0.0;
      for (unsigned int j2=0; j2<Nprm; ++j2) {
	int linind2 = i*Nprm + j2;
	double B_ijk = 0.0;
	for (unsigned int l=0; l<3; ++l) {
	  double a_ljk_1 = gsl_matrix_get(con,l,j1);
	  //if (j1 < 3) a_ljk_1 = RadtoDeg(a_ljk_1);
	  //if (j1 < 3) a_ljk_1 *= 30;
	  //double a_ljk_1 = 1.0;
	  double a_ljk_2 = gsl_matrix_get(con,l,j2);
	  //if (j2 < 3) a_ljk_2 = RadtoDeg(a_ljk_2);
	  //if (j2 < 3) a_ljk_2 *= 30;
	  //double a_ljk_2 = 1.0;
	  B_ijk += (2 * F5 * a_ljk_2 * a_ljk_1);
	  //debug
	  //printf("j1=%d, j2=%d: a_ljk_1 * a_ljk_2 = %g\n",j1,j2,a_ljk_1*a_ljk_2);
	}
	//debug
	//printf("B_ijk = %10.5f; bst = %8.3f\n",B_ijk,gsl_matrix_get(bst,i,j2) );
	double Aold = gsl_matrix_get(A,linind1,linind2);
	gsl_matrix_set(A,linind1,linind2,B_ijk + Aold);
	tmp += (B_ijk * gsl_matrix_get(bst,i,j2) );
      }
      double bold = gsl_vector_get(b,linind1);
      gsl_vector_set(b,linind1,tmp + bold);
    }
  }
  
  //debug
  printf("======= Printing b =======\n");
  gsl_vector_fprintf(stdout, b, "%6.3f");
  printf("======= Printing A =======\n");
  //gsl_matrix_fprintf(stdout, A, "%6.3f"); // writes it all in one column
  for (unsigned int i=0; i<20; ++i) {
    printf("i=%3d: ",i);
    for (unsigned int j=0; j<20; ++j) {
      //if (gsl_matrix_get(A,i,j) == 0.0) continue;
      printf("%8.5f ",gsl_matrix_get(A,i,j) );
    }
    printf("\n");
  }
  //exit(0);
  
  gsl_matrix_free(con);
}
*/

// This function is used to compute a 3x6 matrix of Jacobian coeffs which are used to
// assess deviations of \vec{r_i} from its ideal values.
void Equations::GetXYZMatrix(gsl_matrix *bst, gsl_matrix *mat, unsigned int ind) {
  // Compute WO prms:
  double twi = DegtoRad(gsl_matrix_get(bst,ind,0) ); // twist roll tilt (in rad!!)
  double r = DegtoRad(gsl_matrix_get(bst,ind,1) );
  double t = DegtoRad(gsl_matrix_get(bst,ind,2) );
  double shift = gsl_matrix_get(bst,ind,3);
  double slide = gsl_matrix_get(bst,ind,4);
  double rise = gsl_matrix_get(bst,ind,5);
  // Precompute sin/cos:
  double sinm = sin(0.5*twi);
  double cosm = cos(0.5*twi);
  // Compute Jacobian:
  // x
  gsl_matrix_set(mat,0,0, DegtoRad(-0.5 * sinm * (0.5 * rise * r + shift) + 0.5 * cosm * (0.5 * rise * t - slide) ) );
  gsl_matrix_set(mat,0,1, DegtoRad(0.5 * rise * cosm) );
  gsl_matrix_set(mat,0,2, DegtoRad(0.5 * rise * sinm) );
  gsl_matrix_set(mat,0,3, cosm);
  gsl_matrix_set(mat,0,4, -sinm);
  gsl_matrix_set(mat,0,5, 0.5 * r * cosm + 0.5 * t * sinm);
  // y
  gsl_matrix_set(mat,1,0, DegtoRad(0.5 * sinm * (0.5 * rise * t - slide) + 0.5 * cosm * (0.5 * rise * r + shift) ) );
  gsl_matrix_set(mat,1,1, DegtoRad(0.5 * rise * sinm) );
  gsl_matrix_set(mat,1,2, DegtoRad(-0.5 * rise * cosm) );
  gsl_matrix_set(mat,1,3, sinm);
  gsl_matrix_set(mat,1,4, cosm);
  gsl_matrix_set(mat,1,5, 0.5 * r * sinm - 0.5 * t * cosm);
  // z
  gsl_matrix_set(mat,2,0, DegtoRad(0.0) );
  gsl_matrix_set(mat,2,1, DegtoRad(-0.5 * shift) );
  gsl_matrix_set(mat,2,2, DegtoRad(0.5 * slide) );
  gsl_matrix_set(mat,2,3, -0.5 * r);
  gsl_matrix_set(mat,2,4, 0.5 * t);
  gsl_matrix_set(mat,2,5, 1.0);
}

// This function adds global "xyz" constraints to Ax = b.
void Equations::setGlobalConstr(gsl_matrix *bst) {
  for (unsigned int i1=0; i1<Lseq; ++i1) {
    for (unsigned int j1=0; j1<Nprm; ++j1) {
      int linind1 = i1*Nprm + j1;
      double tmp = 0.0;
      for (unsigned int i2=0; i2<Lseq; ++i2) {
	for (unsigned int j2=0; j2<Nprm; ++j2) {
	  int linind2 = i2*Nprm + j2;
	  double B_ijkl = 0.0;
	  for (unsigned int l=0; l<3; ++l) {
	    double a_ljk_1 = gsl_matrix_get(derivs[l],i1,linind1);
	    double a_ljk_2 = gsl_matrix_get(derivs[l],i2,linind2);
	    B_ijkl += (2 * F5 * a_ljk_2 * a_ljk_1);
	    //debug
	    //printf("j1=%d, j2=%d: a_ljk_1 * a_ljk_2 = %g\n",j1,j2,a_ljk_1*a_ljk_2);
	  }
	  //debug
	  //printf("B_ijkl = %10.5f; bst = %8.3f\n",B_ijkl,gsl_matrix_get(bst,i,j2) );
	  double Aold = gsl_matrix_get(A,linind1,linind2);
	  gsl_matrix_set(A,linind1,linind2,B_ijkl + Aold);
	  tmp += (B_ijkl * gsl_matrix_get(bst,i2,j2) );
	}
      }
      double bold = gsl_vector_get(b,linind1);
      gsl_vector_set(b,linind1,tmp + bold);
    }
  }
  /*
  //debug
  //printf("======= Printing b =======\n");
  //gsl_vector_fprintf(stdout, b, "%6.3f");
  printf("======= Printing A =======\n");
  //gsl_matrix_fprintf(stdout, A, "%6.3f"); // writes it all in one column
  for (unsigned int i=0; i<DIM; ++i) {
    printf("i=%3d: ",i);
    for (unsigned int j=0; j<DIM; ++j) {
      if (gsl_matrix_get(A,i,j) == 0.0) continue;
      printf("%8.5f ",gsl_matrix_get(A,i,j) );
    }
    printf("\n");
  }
  //exit(0);
  */
}

/*
// This function adds global "xyz" constraints to Ax = b.
void Equations::setGlobalConstr(gsl_matrix *bst) {
  double C[DIM][DIM];
  double d[DIM];
  for (unsigned int i=0; i<DIM; ++i) {
    d[i] = 0.0;
    for (unsigned int j=0; j<DIM; ++j) {
      C[i][j] = 0.0;
    }
  }
  
  for (unsigned int i1=0; i1<Lseq; ++i1) {
    for (unsigned int j1=0; j1<Nprm; ++j1) {
      int linind1 = i1*Nprm + j1;
      for (unsigned int j2=0; j2<Nprm; ++j2) {
	int linind2 = i1*Nprm + j2;
	double B_ijk = 0.0;
	for (unsigned int l=0; l<3; ++l) {
	  double a_ljk_1 = gsl_matrix_get(derivs[l],i1,linind1);
	  //double a_ljk_1 = 1.0;
	  double a_ljk_2 = gsl_matrix_get(derivs[l],i1,linind2);
	  //double a_ljk_2 = 1.0;
	  B_ijk += (2 * F5 * a_ljk_2 * a_ljk_1);
	}
	C[linind1][linind2] = B_ijk;
	d[linind1] += (B_ijk * gsl_matrix_get(bst,i1,j2) );
	//debug
	//printf("i1 = %d; j2 = %d: bst = %8.3f\n",i1,j2,gsl_matrix_get(bst,i1,j2) );
	//printf("C[%d][%d] = %8.3f\n",linind1,linind2,C[linind1][linind2]);
      }
      //debug
      //printf("d[%d] = %8.3f\n",linind1,d[linind1]);      
      //exit(0);
    }
  }

  //debug
  printf("======= Printing d =======\n");
  for (unsigned int i=0; i<DIM; ++i) {
    printf("%8.5f\n",d[i]);
  }
  printf("======= Printing C =======\n");
  for (unsigned int i=0; i<30; ++i) {
    for (unsigned int j=0; j<30; ++j) {
      printf("%8.5f ",C[i][j]);
    }
    printf("\n");
  }


  for (unsigned int i1=0; i1<Lseq; ++i1) {
    for (unsigned int j1=0; j1<Nprm; ++j1) {
      int linind1 = i1*Nprm + j1;
      for (unsigned int l=0; l<3; ++l) {
	double a_ljk = gsl_matrix_get(derivs[l],i1,linind1);
	//double a_ljk = 1.0;
	C[linind1][linind1] += (2 * F5 * a_ljk * a_ljk);
	d[linind1] += (2 * F5 * a_ljk * a_ljk * gsl_matrix_get(bst,i1,j1) );
	//debug
	//printf("i1 = %d; j1 = %d: bst = %8.3f\n",i1,j1,gsl_matrix_get(bst,i1,j1) );
      }
    }
    //debug
    //printf("C[%d][%d] = %8.3f\n",linind1,linind1,C[linind1][linind1]);
    //printf("d[%d] = %8.3f\n",linind1,d[linind1]);
    //exit(0);
  }

  // Add the constraints to A & b:
  for (unsigned int i=0; i<DIM; ++i) {
    for (unsigned int j=0; j<DIM; ++j) {
      double Aold = gsl_matrix_get(A,i,j);
      gsl_matrix_set(A,i,j,Aold + C[i][j]);
    }
    double bold = gsl_vector_get(b,i);
    gsl_vector_set(b,i,bold + d[i]);
  }
}
*/

// This function adds a curvature constraint to Ax = b.
void Equations::setCurvConstr(void) {
  for (unsigned int i=0; i<Lseq; ++i) {
    for (unsigned int j=1; j<=2; ++j) { // j=1: roll; j=2: tilt
      int linind = i*Nprm + j;
      double F = 0.0;
      double ideal = 0.0;
      if (j==1) { // choose between F1 & F2, and between ideal roll & tilt
	F = F1;
	ideal = kappa_roll * cos(i*DegtoRad(Omega_0) + DegtoRad(Phi_0) );
      }else if (j==2) {
	F = F2;
	ideal = kappa_tilt * sin(i*DegtoRad(Omega_0) + DegtoRad(Phi_0) );
      }
      double Aupdated = 2 * F + gsl_matrix_get(A,linind,linind); // note that each old value of A must be invoked only once in summation!!
      gsl_matrix_set(A,linind,linind,Aupdated);
      double bupdated = 2 * F * ideal + gsl_vector_get(b,linind); // note that each old value of b must be invoked only once in summation!!
      gsl_vector_set(b,linind,bupdated);
    }
  }
}

// This function adds a curvature constraint to Ax = b.
void Equations::setTwistConstr(void) {
  for (unsigned int i1=0; i1<Lseq; ++i1) {
    int linind_twist_1 = i1*Nprm; // j=0: twist
    double bupdated = gsl_vector_get(b,linind_twist_1);
    for (unsigned int i2=0; i2<Lseq; ++i2) {
      int linind_twist_2 = i2*Nprm; // j=0: twist
      double Aupdated = 2 * F3 + gsl_matrix_get(A,linind_twist_1,linind_twist_2);
      gsl_matrix_set(A,linind_twist_1,linind_twist_2,Aupdated);
      bupdated += 2 * F3 * Omega_0;
    }
    gsl_vector_set(b,linind_twist_1,bupdated);
  }
}

// This function solves Ax = b (A & b must be set!)
void Equations::Solve(void) {
  int s;
  gsl_permutation *p = gsl_permutation_alloc(DIM);
  gsl_linalg_LU_decomp(A, p, &s);
  gsl_linalg_LU_solve(A, p, b, x);
  //debug
  //printf ("x = \n");
  //gsl_vector_fprintf (stdout, x, "%g");
  
  gsl_permutation_free (p);
}

// This function solves Ax = b using SVD (A & b must be set!)
void Equations::SolveSVD(void) {
  
  gsl_matrix *Acpy = gsl_matrix_calloc(DIM,DIM);
  gsl_matrix_memcpy(Acpy,A);
  gsl_matrix *V = gsl_matrix_calloc(DIM,DIM);
  gsl_vector *S = gsl_vector_calloc(DIM);
  gsl_vector *work = gsl_vector_calloc(DIM);
  
  //gsl_linalg_SV_decomp_jacobi(Acpy,V,S); // Acpy becomes U; V is V
  gsl_linalg_SV_decomp(Acpy,V,S,work);
  gsl_linalg_SV_solve(Acpy,V,S,b,x);
  
  gsl_matrix_free(Acpy);
  gsl_matrix_free(V);
  gsl_vector_free(S);
  gsl_vector_free(work);
  /*
  //debug
  gsl_matrix *Acpy = gsl_matrix_calloc(6,6);
  gsl_vector *bcpy = gsl_vector_calloc(6);
  gsl_vector *xcpy = gsl_vector_calloc(6);
  for (int i=0; i<6; ++i) {
    gsl_vector_set(bcpy,i,gsl_vector_get(b,i) );
    for (int j=0; j<6; ++j) {
      gsl_matrix_set(Acpy,i,j,gsl_matrix_get(A,i,j) );
    }
  }
  
  gsl_matrix *V = gsl_matrix_calloc(6,6);
  gsl_vector *S = gsl_vector_calloc(6);
  gsl_vector *work = gsl_vector_calloc(6);
  
  //debug
  printf("======= Printing Acpy =======\n");
  for (unsigned int i=0; i<6; ++i) {
    printf("i=%3d: ",i);
    for (unsigned int j=0; j<6; ++j) {
      printf("%8.5f ",gsl_matrix_get(Acpy,i,j) );
    }
    printf("\n");
  }
  
  //gsl_linalg_SV_decomp_jacobi(Acpy,V,S); // Acpy becomes U; V is V
  gsl_linalg_SV_decomp(Acpy,V,S,work);
  gsl_linalg_SV_solve(Acpy,V,S,bcpy,xcpy);
  
  //debug
  printf("======= Printing bcpy =======\n");
  gsl_vector_fprintf(stdout, bcpy, "%6.3f");
  printf("======= Printing U =======\n");
  for (unsigned int i=0; i<6; ++i) {
    printf("i=%3d: ",i);
    for (unsigned int j=0; j<6; ++j) {
      printf("%8.5f ",gsl_matrix_get(Acpy,i,j) );
    }
    printf("\n");
  }
  printf("======= Printing V =======\n");
  for (unsigned int i=0; i<6; ++i) {
    printf("i=%3d: ",i);
    for (unsigned int j=0; j<6; ++j) {
      printf("%8.5f ",gsl_matrix_get(V,i,j) );
    }
    printf("\n");
  }
  printf("======= Printing S =======\n");
  gsl_vector_fprintf(stdout, S, "%6.3f");
  printf("======= Printing xcpy =======\n");
  gsl_vector_fprintf(stdout, xcpy, "%6.3f");
  exit(0);
  */
}

void Equations::Eigen(void) {
  Eigen(A);
}

// This [ diagnostic ] function finds eigenvectors & eigenvalues for a
// REAL SYMMETRIC matrix A.
void Equations::Eigen(gsl_matrix *mat) {
  // Check symmetry:
  for (unsigned int i=0; i<DIM; ++i) {
    for (unsigned int j=0; j<DIM; ++j) {
      if (j<=i) continue;
      if (fabs(gsl_matrix_get(mat,i,j) - gsl_matrix_get(mat,j,i) ) > 1e-04) { // mat must be symmetric!!
	std::cerr << "Matrix not symmetric - skipping diagonalization!\n";
	return;
      }
    }
  }
  // Diagonalize:
  gsl_vector *eval = gsl_vector_alloc(DIM);
  gsl_eigen_symm_workspace *w = gsl_eigen_symm_alloc(DIM);
  
  gsl_eigen_symm(mat, eval, w); // main function
  gsl_eigen_symm_free(w);
  // Print result:
  for (unsigned int i = 0; i < DIM; ++i) {
    double eval_i = gsl_vector_get(eval, i);
    printf ("eigenvalue[%d] = %g\n",i,eval_i);
  }
  gsl_vector_free(eval);
}

// This function prints x in a convenient form. 
void Equations::PrintResult(void) {
  if (x==NULL) {
    fprintf(stderr,"No results to print - did you run Solve()?!\n");
    return;
  }
  fprintf(stdout," k    Twist    Roll    Tilt    Shift   Slide   Rise\n");
  for (unsigned int i=0; i<Lseq; ++i) {
    fprintf(stdout,"%3d",i+1);
    for (unsigned int j=0; j<Nprm; ++j) {
      int linind = i*Nprm + j;
      fprintf(stdout,"  %6.2f",gsl_vector_get(x,linind) );
    }
    fprintf(stdout,"\n");
  }
}

// This function returns the total length of x:
int Equations::length(void) {
  return DIM;
}

double Equations::getx(int ind) {
  return gsl_vector_get(x,ind);
}

void Equations::setx(int ind, double value) {
  gsl_vector_set(x,ind,value);
}

gsl_vector *Equations::result(void) {
  return x;
}

// This function computes (linearized) |curvature| at each basestep given x.
void Equations::GetCurv(void) {
  if (x==NULL) {
    fprintf(stderr,"Curvature cannot be computed - did you run Solve()?!\n");
    return;
  }
  for (unsigned int i=0; i<Lseq; ++i) {
    int linind_roll = i*Nprm + 1; // roll_i
    int linind_tilt = i*Nprm + 2; // tilt_i
    double d1 = gsl_vector_get(x,linind_roll);
    double d2 = gsl_vector_get(x,linind_tilt);
    gsl_vector_set(curv,i,sqrt(d1*d1 + d2*d2) ); // modulus of curvature vector (ie Gamma)
  }
}

// This function prints computed curvature.
void Equations::PrintCurv(void) {
  if (curv==NULL) {
    fprintf(stderr,"Curvature cannot be printed - did you run GetCurv()?!\n");
    return;
  }
  fprintf(stdout," k    Curv\n");
  for (unsigned int i=0; i<Lseq; ++i) {
    fprintf(stdout,"%3d  %6.2f\n",i+1,gsl_vector_get(curv,i) );
  }
}

// This function reads in a table of pre-computed derivatives.
void Equations::ReadDerivs(const std::string filename) {
  // Allocate memory:
  derivs = new gsl_matrix*[3];
  for (unsigned int j=0; j<3; ++j) {
    derivs[j] = gsl_matrix_calloc(Lseq,Lseq*Nprm); // allocate table for all derivs
  }
  
  using std::ifstream;
  
  ifstream file(filename.c_str() );
  
  char buffer[512];
  
  if (file.fail()) {
    std::cerr << "Could not open file: "<<filename<<" for reading!\nExiting..\n";
    exit(EXIT_FAILURE);
  }
  
  unsigned int ind1,ind2,ind3,ind4;
  double tmp;
  int cnt = 0; // line counter
  while (!file.eof()) {
    file.getline(buffer,512,'\n');
    if (file.fail() && !file.eof() ) file.clear();
    if (cnt==0) {
      if (sscanf(buffer,"%d %d %d %d",&ind1,&ind2,&ind3,&ind4) == 4) { // read in table dims
	if (ind1!=Lseq || ind2!=Lseq || ind3!=Nprm || ind4!=3) { // check table dims
	  std::cerr << "Incorrect table dims in: "<<filename<<"!\nExiting..\n";
	  exit(EXIT_FAILURE);
	}
      }else{
	std::cerr << "Unknown header format in: "<<filename<<"!\nExiting..\n";
	exit(EXIT_FAILURE);
      }
      ++cnt;
      continue; // get table dims from header
    }
    if (sscanf(buffer,"%d %d %d %d %lf",&ind1,&ind2,&ind3,&ind4,&tmp) == 5) {
      int linind = ind2*Nprm + ind3;
      gsl_matrix_set(derivs[ind4],ind1,linind,tmp);
      ++cnt;
    }
  }
  /*
  //debug
  for (unsigned int i1=0; i1<Lseq; ++i1) { // over all frames; frame 0 is independent of any WO prms
    for (unsigned int i2=0; i2<Lseq; ++i2) { // over all sets of WO prms
      for (unsigned int j2=0; j2<Nprm; ++j2) {
	int linind2 = i2*Nprm + j2;
	for (unsigned int j=0; j<3; ++j) { // over x,y,z components
	  fprintf(stdout,"%4d %4d %4d %4d %10.5f\n",
		  i1,i2,j2,j,gsl_matrix_get(derivs[j],i1,linind2) );
	}
      }
    }
  }
  */
  file.close();
}

