/* esolve.c */

/*****************************************************************
        Copyright by Rockefeller University,
can not be reproduced or distributed without written permission of
the copyright holder.  Version 1.1 of April 2006.

Written by Alexandre Morozov (contact person: morozov@edsb.rockefeller.edu),
and Eric Siggia.

The program DNABEND implements an algorithm for computing elastic energies
for DNA molecules of arbitrary length and sequence. DNA can be constrained
to follow an arbitrary 3D curve, such as a left-handed superhelix found
in nucleosomal core particles.

The DNABEND method and its applications are described in the following
paper:

******************************************************************/

#include "options.h"
#include "fastafile.h"
#include "sequence.h"
#include "equations.h"
#include "dna_parameters.h"
#include "bstable.h"
#include "xyzframe.h"
#include "minimize.h"

// This main function sets up and solves Ax=b, where A,b are given by WO potential + constraints,
// and x is a solution for a set of WO prms.
// It also translates any set of WO prms (obtained from x or otherwise) into global coords
// of xyz frames associated with basepairs or MSTs (so-called mid-step triads).

std::vector<std::string> ReadList(std::string const &filename);
// Diagnostic functions:
double Econstr(Geom::Geom &xyz_ideal, int bp_ind1, int dof_ind1, double val1, int bp_ind2, int dof_ind2, double val2);
double Ederiv(Geom::Geom &xyz_ideal, int bp_ind1, int dof_ind1, double val1, int bp_ind2, int dof_ind2, double val2);
Vector Dglob(Geom::Geom &xyz_ideal, int bp_ind, int dof_ind, double val, int ind_out);
Vector Ddiff(Geom::Geom &xyz_ideal, int bp_ind, int dof_ind, double val, int ind_out);

int main(int argc, char **argv) {
  // (A) ==== Read in command-line options: ==== //
  if (argc < 2) {
    fprintf(stderr,"usage: %s [-fasta <Fasta file of input seqs>] [-geom_wt <WOprms_wt.dat>] [-geom_id <WOprms_ideal.dat>] [-PDBid_bp <ideal_helix_fitted.long.pdb>] [-PDBid_mst <ideal_helix_mst_fitted.long.pdb>] [-curv <curvature_ideal.dat>] [-fpdb <pdblist.dat>] [-fgeom <geomlist.dat>] [-w_el <elastic energy weight>] [-w_sh <superhelical constraint weight>] [-w_sum <sum of angles constraint weight>] [-w_dev <deviations from ideal frames constraint weight>] [-compute_curv]\n",argv[0]);
    exit(1);
  }
  int argbase = 1;
  struct Options *opt = ReadArguments(argbase, argc, argv);
  // (B) ==== Read DNA parameters (moments + force constants): ==== //
  DNAparams DNAprms;
  // (C) ==== Compute & store a set of xyz frames corresponding to the wt (i.e. pdb) sequence: ==== //
  // Table of wt WO prms (read in from file or obtained thru solving Ax=b for an input seq): 
  BStable bst_wt(opt->geom_wt);
  //bst_wt.WriteBStable("WOprms_wt.dat.copy");
  //gsl_matrix *m = bst_wt.Export();
  // Set of wt frames:
  Geom xyz_wt(bst_wt);
  //xyz_wt.PDBunitvecs("WOprms_wt.long.pdb",'l');
  //xyz_wt.PDB("junk_wt.pdb");
  //exit(0);
  // (D) ==== Read in & store a set of ideal superhelical frames superimposed onto the wt conformation: ==== //
  Geom *xyz_ideal_ptr;
  if (opt->geom_id != "none") { // user provided ideal BStable: use it rather than the 2 PDB files
    std::cerr << "Using a BStable: "<< opt->geom_id <<" to construct the ideal curve ..\n";
    BStable bst_ideal(opt->geom_id);
    xyz_ideal_ptr = new Geom(bst_ideal);
    //xyz_ideal_ptr = new Geom(bst_wt);
  }else{ // either default or user-provided PDB files with bp and MST frames
    std::cerr << "Using PDB files: ("<< opt->PDBid_bp <<","<< opt->PDBid_mst <<") to construct the ideal curve ..\n";
    xyz_ideal_ptr = new Geom(opt->PDBid_bp,opt->PDBid_mst);
  }
  Geom xyz_ideal = *xyz_ideal_ptr;
  //xyz_ideal.PDB("junk_ideal.pdb");
  //exit(0);
  // Compute & write derivs & curvatures for the ideal frames if requested:
  if (opt->compute_curv) {
    std::cerr << "Computing derivatives and curvature coeffs for ideal frames ..\n";
    xyz_ideal.BuildDerivs();
    // debug: check derivs
    //Matrix3 D = xyz_ideal.NumDeriv(140, 10, 1, 1.0);
    //D.Print();
    //Matrix3 D2 = xyz_ideal.GetDeriv(140, 10, 1);
    //D2.Print();
    //exit(0);
    xyz_ideal.ComputeCurv();
    xyz_ideal.WriteCurv(opt->curv);
    std::cerr << "Computed curvature coeffs and saved them in: "<<opt->curv<<" ..\n";
    exit(EXIT_SUCCESS);
  }else{
    xyz_ideal.ReadCurv(opt->curv);
  }
  //xyz_ideal.WriteConnectivityCoeff("connectivity.dat"); // this is for testing only
  //////////////////////////////
  //xyz_ideal.CompareFrames(xyz_wt,"mst");
  //xyz_ideal.Write("mst");
  //========= debug =========//
  //double E1 = Econstr(xyz_ideal, 10, 1, 0.5, 30, 0, 0.5);
  //double E2 = Ederiv(xyz_ideal, 10, 1, 0.5, 30, 0, 0.5);
  //printf("Compare E1 = %.4f with E2 = %.4f ..\n",E1,E2);
  //Vector x1 = Dglob(xyz_ideal, 10, 1, 1.0, 100);
  //Vector x2 = Ddiff(xyz_ideal, 10, 1, 1.0, 100);
  //printf("x1 = [%.3f %.3f %.3f], |x1| = %.3f\n",x1.x,x1.y,x1.z,x1.Length() );
  //printf("x2 = [%.3f %.3f %.3f], |x2| = %.3f\n",x2.x,x2.y,x2.z,x2.Length() );
  //exit(0);
  //////////////////////
  // =========================================================== //
  //Geom xyz_wt2(bst_wt);
  //xyz_wt2.Transform(xyz_ideal,10,140);
  //std::cout << "RMSD = " << xyz_wt2.RMSD(xyz_ideal) << std::endl;
  //xyz_wt2.PDBunitvecs("junk_wt_unitv_transformed.pdb");
  //exit(0);
  // (E) ==== Read in the sequence(s) from input file: ==== //
  char *fastafile = opt->fasta;
  FastaFile sequences;
  sequences.ReadFasta(fastafile);
  std::cerr << "Read "<<sequences.Size()<<" seq(s) from file: "<<fastafile<<std::endl;
  if (sequences.Size() == -1) {
    std::cerr << "WARNING: fasta file "<<fastafile<<" is empty!\nExiting ..\n";
    exit(EXIT_SUCCESS);
  }
  // (F) ==== For each seq, minimize its DNA energy and compare resulting frames with wt: ==== //
  // Read in a list of output PDB filenames and BStable files (if any):
  std::vector<std::string> pdbsout, geomsout;
  if (opt->fpdb != "none") pdbsout = ReadList(opt->fpdb);
  if (opt->fgeom != "none") geomsout = ReadList(opt->fgeom);
  // Check dims of filename lists - output PDBs and BStables must be either both present or both absent!
  if ((pdbsout.size()==0 && geomsout.size() > 0) || (pdbsout.size() > 0 && geomsout.size()==0) ) {
    std::cerr << "ERROR: output PDB and BStable filenames must be either specified together or omitted!\n";
    exit(EXIT_FAILURE);
  }
  if (pdbsout.size() > 0 && geomsout.size() > 0 &&
      (pdbsout.size() != (unsigned int) sequences.Size() || geomsout.size() != (unsigned int) sequences.Size() ) ) {
    std::cerr << "ERROR: please provide one PDB and one BStable filename for each input sequence!\n";
    exit(EXIT_FAILURE);
  }
  for (int seqnum=0; seqnum < sequences.Size(); ++seqnum) {
    Sequence *seq = sequences[seqnum];
    seq->Print(0,seq->Length()-1);
    /**********************/
    if (pdbsout.size() > 0) {
      Minimizer min(seq, DNAprms, xyz_ideal, xyz_wt, opt, pdbsout.at(seqnum), geomsout.at(seqnum), "verbose");
    }else{
      Minimizer min(seq, DNAprms, xyz_ideal, xyz_wt, opt, "none", "none", "verbose");
    }
    //Equations eq(seq,DNAprms);
    //Equations eq(seq,DNAprms,m,"bstable");
    //Equations eq(seq,DNAprms,m,"xyzspace");
    //eq.Solve();
    //BStable bst(eq.result(),seq); // table of WO prms from x (angles in Deg!)
    //bst.WriteBStable("junk");
    //std::cout << "E = "<<eq.Energy(seq,DNAprms)<<std::endl;
    /**********************/
  }
  //gsl_matrix_free(m);
  delete opt;
}

// A generic function to read 1 item per line from a file and return a vector with read items.
// No header is expected.
std::vector<std::string> ReadList(std::string const &filename) {
  std::vector<std::string> items;
  
  using std::cout;
  using std::cerr;
  using std::endl;
  using std::ifstream;
  
  ifstream file(filename.c_str() );
  
  char buffer[512];
  char citem[512];
  
  if (file.fail()) {
    cerr << "Could not open file: "<<filename<<" for reading!\nExiting..\n";
    exit(EXIT_FAILURE);
  }
  
  int cnt = 0; // item counter
  while (!file.eof()) {
    file.getline(buffer,512,'\n');
    if (file.fail() && !file.eof() ) file.clear();
    if (sscanf(buffer,"%s",citem) == 1) {
      std::string item(citem);
      items.push_back(item);
      cnt++;
    }
  }
  file.close();
  cerr << "Read "<<cnt<< " item(s) from file: "<<filename<<" ..\n";
  
  return items;
}

// This testing function computes displacement constraint energy by "brute force":
// it constructs 2 sets of frames with one DoF slightly changed, and then computes
// the diff. of global radius-vectors. This diff. squared is essentially the constraint
// energy.
double Econstr(Geom::Geom &xyz_ideal, int bp_ind1, int dof_ind1, double val1, int bp_ind2, int dof_ind2, double val2) {
  double E = 0.0;
  static const unsigned int Nprm = 6;
  static const unsigned int Lseq = 146;
  gsl_vector *v1 = xyz_ideal.GetGeom(); // get WO prms from a set of ideal frames
  // User-changed parameters for 2 prms:
  double a_old = gsl_vector_get(v1,bp_ind1*Nprm + dof_ind1);
  gsl_vector_set(v1, bp_ind1*Nprm + dof_ind1, a_old + val1);
  double b_old = gsl_vector_get(v1,bp_ind2*Nprm + dof_ind2);
  gsl_vector_set(v1, bp_ind2*Nprm + dof_ind2, b_old + val2);
  // New sets of frames with changed prms:
  Geom xyz(v1,xyz_ideal.frames.at(0) );
  //xyz.PDBunitvecs("junk2.pdb");
  // Compute Econstr:
  for (unsigned int s=0; s<Lseq; ++s) {
    double Ediff = (xyz.frames.at(s+1).orig - xyz_ideal.frames.at(s+1).orig) *
      (xyz.frames.at(s+1).orig - xyz_ideal.frames.at(s+1).orig);
    //debug
    printf("s=%d: %.4f\n",s+1,Ediff);
    E += Ediff;
  }
  // Free memory:
  gsl_vector_free(v1);
  return E;
}

// This testing function computes displacement constraint energy given a set of pre-computed curvature coeffs.
double Ederiv(Geom::Geom &xyz_ideal, int bp_ind1, int dof_ind1, double val1, int bp_ind2, int dof_ind2, double val2) {
  double E = 0.0; // total WO energy of the conformation
  static const unsigned int Nprm = 6;
  //static const unsigned int Lseq = 146;
  gsl_matrix *c = xyz_ideal.GetCurv();
  
  // The energy is simply the force const coeff x product of user-defined deviations from the ideal values.
  int linind1 = bp_ind1*Nprm + dof_ind1;
  int linind2 = bp_ind2*Nprm + dof_ind2;
  E = (gsl_matrix_get(c,linind1,linind2) + gsl_matrix_get(c,linind2,linind1) ) * val1 * val2 +
    gsl_matrix_get(c,linind1,linind1) * val1 * val1 + gsl_matrix_get(c,linind2,linind2) * val2 * val2;
  
  return E;
}

// This testing function computes diff. displacement by using global frame reconstructions.
Vector Dglob(Geom::Geom &xyz_ideal, int bp_ind, int dof_ind, double val, int ind_out) {
  Vector x(0,0,0);
  static const int Nprm = 6;
  static const int Lseq = 146;
  assert(ind_out > 0 && ind_out <= Lseq);
  assert(bp_ind >=0 && bp_ind < Lseq);
  assert(dof_ind >=0 && dof_ind < Nprm);
  gsl_vector *v1 = xyz_ideal.GetGeom(); // get WO prms from a set of ideal frames
  // User-changed parameters for 2 prms:
  double a_old = gsl_vector_get(v1,bp_ind*Nprm + dof_ind);
  gsl_vector_set(v1, bp_ind*Nprm + dof_ind, a_old + val);
  // New set of frames with changed prms:
  Geom xyz(v1,xyz_ideal.frames.at(0) );
  x = xyz.frames.at(ind_out).orig - xyz_ideal.frames.at(ind_out).orig;
  std::cout << "RMSD = " << xyz_ideal.RMSD(xyz) << std::endl;
  //debug
  //xyz_ideal.PDBunitvecs("junk1.pdb");
  //xyz.PDBunitvecs("junk2.pdb");
  //exit(0);
  return x;
}

// This is also a testing function.
Vector Ddiff(Geom::Geom &xyz_ideal, int bp_ind, int dof_ind, double val, int ind_out) {
  Vector x(0,0,0);
  static const int Nprm = 6;
  static const int Lseq = 146;
  assert(dof_ind >=0 && dof_ind < Nprm);
  assert(ind_out >=0 && ind_out < Lseq);
  assert(bp_ind >=0 && bp_ind < Lseq);
  //==========
  gsl_matrix *b = gsl_matrix_calloc(3*Lseq,3*Lseq);
  FILE *fp = fopen("connectivity.dat","r");
  assert(gsl_matrix_fread(fp,b)==0); // 1st index: [j,beta]; 2nd index: [j1,alpha]
  fclose(fp);
  //==========
  if (dof_ind >= 3) {
    if (bp_ind <= ind_out) {
      if (dof_ind==3) {
	x.x = val;
      }else if (dof_ind==4) {
	x.y = val;
      }else if (dof_ind==5) {
	x.z = val;
      }
      coord mst = xyz_ideal.frames_mst.at(bp_ind);
      Matrix3 T(mst.a,mst.b,mst.c);
      x = T.VecMultiply(x);
    }
  }else{
    double tmp[3] = {0, 0, 0};
    for (int j=bp_ind; j<=ind_out; ++j) {
      for (int beta=0; beta<3; ++beta) {
	int linind1 = j*3 + beta;
	int linind2 = bp_ind*3 + dof_ind;
	tmp[beta] += (val * gsl_matrix_get(b,linind1,linind2) );
	//debug
	//printf("after: linind1 = %d, linind2 = %d: b = %.3f\n",linind1,linind2,gsl_matrix_get(b,linind1,linind2) );
      }
    }
    x.x = tmp[0];
    x.y = tmp[1];
    x.z = tmp[2];
  }
  gsl_matrix_free(b);
  return x;
}

