/* geometry.c */

/*****************************************************************
        Copyright by Rockefeller University,
can not be reproduced or distributed without written permission of
copyright holder.  Version of April 2006.

Written by Alexandre Morozov (contact person: morozov@edsb.rockefeller.edu),
and Eric Siggia.

The program DNABEND implements an algorithm for computing elastic energies
for DNA molecules of arbitrary length and sequence. DNA can be constrained
to follow an arbitrary 3D curve, such as a left-handed superhelix found
in nucleosomal core particles.

The DNABEND method and its applications are described in the following
paper:

******************************************************************/

#include "geometry.h"


// Returns polar coords of the vector x; discards its length.
Polar_ CartesianToPolar(Vector b){
  Polar_ a;
  if (b.Length()!=0) {
    a.theta=acos(b.z/b.Length());
    if (fabs(a.theta)>EPS && fabs(a.theta-PI)>EPS) {
      a.phi=asin(b.x/(b.Length()*sin(a.theta)));
    }else{
      a.phi=0.0;//phi undefined
    }
  }else{
    std::cerr << "Unable to find polar angles for a Vector of zero length!\n"
	 << "Quitting..\n";
    exit(EXIT_FAILURE);
  }
  return a;
}

//This function rotates a Vector into coord.system specified by Euler angle (phi,theta,0),
//where (phi,theta) are polar angles.
Vector RotatePolar(Vector a,Polar_ ang){
  Vector b;
  double sin_th,sin_ph,cos_th,cos_ph;

  sin_th=sin(ang.theta); cos_th=cos(ang.theta);
  sin_ph=sin(ang.phi); cos_ph=cos(ang.phi);
  b.x=a.x*sin_ph*sin_th+a.y*cos_ph*sin_th+a.z*cos_th;
  b.y=-a.x*cos_ph+a.y*sin_ph;
  b.z=-a.x*sin_ph*cos_th-a.y*cos_ph*cos_th+a.z*sin_th;
  return b;
}

//This function rotates back a vector transformed with RotatePolar.
Vector InversePolar(Vector a,Polar_ ang){
  Vector b;
  double sin_th,sin_ph,cos_th,cos_ph;
  
  sin_th=sin(ang.theta); cos_th=cos(ang.theta);
  sin_ph=sin(ang.phi); cos_ph=cos(ang.phi);
  b.x=a.x*sin_ph*sin_th-a.y*cos_ph-a.z*cos_th*sin_ph;
  b.y=a.x*sin_th*cos_ph+a.y*sin_ph-a.z*cos_th*cos_ph;
  b.z=a.x*cos_th+a.z*sin_th;
  return b;
}

// Rotation around z-axis:
Vector RotateZ(Vector a,double phi){
  Vector b;
  b.x=a.x*cos(phi)+a.y*sin(phi);
  b.y=-a.x*sin(phi)+a.y*cos(phi);
  b.z=a.z;
  return b;
}

// Rotation around y-axis:
Vector RotateY(Vector a,double the){
  Vector b;
  b.x=a.x*cos(the)+a.z*sin(the);
  b.y=a.y;
  b.z=-a.x*sin(the)+a.z*cos(the);
  return b;
}

// Rotation around x-axis:
Vector RotateX(Vector a,double psi){
  Vector b;
  b.x=a.x;
  b.y=a.y*cos(psi)+a.z*sin(psi);
  b.z=-a.y*sin(psi)+a.z*cos(psi);
  return b;
}

// This function rotates a matrix of 3 unit vectors thru 3 z-y-z angles (the origin stays the same).
// Cf. El Hassan JMB 1995, 251, 648-664.
coord RotateMat(coord const & crd, double a1, double a2, double a3) {
  coord crd_new;
  // x
  crd_new.a = RotateZ(crd.a, a3);
  crd_new.a = RotateY(crd_new.a, a2);
  crd_new.a = RotateZ(crd_new.a, a1);
  // y
  crd_new.b = RotateZ(crd.b, a3);
  crd_new.b = RotateY(crd_new.b, a2);
  crd_new.b = RotateZ(crd_new.b, a1);
  // z
  crd_new.c = RotateZ(crd.c, a3);
  crd_new.c = RotateY(crd_new.c, a2);
  crd_new.c = RotateZ(crd_new.c, a1);
  // origin is simply copied:
  crd_new.orig = crd.orig;
  
  return crd_new;
}

//This function calculates atom's coordinates given 2 bond angles and 1 bond length relative
//to 3 atoms whose new (and old) coordinates are known.
Vector AtomByLAA(tetrad tetr){

  Vector d;
  Vector a,b,c,n;
  double l3;
  double ang_1,ang_3;
  double sin_3,cos_3;
  const double offset = -4e-6;
  
  //Reassign the tetrad:
  if (tetr.a != NULL && tetr.b != NULL && tetr.c != NULL) {
    a = *tetr.a;
    b = *tetr.b;
    c = *tetr.c;
    //Offset all 0 coord. you are building FROM by a small amount:
    if (a.x == 0.0) a.x += offset; if (a.y == 0.0) a.x -= offset; if (a.z == 0.0) a.x += 2*offset;
    if (b.x == 0.0) a.x -= 2*offset; if (b.y == 0.0) a.x += 3*offset; if (b.z == 0.0) a.x -= 3*offset;
    if (c.x == 0.0) a.x += 3*offset; if (c.y == 0.0) a.x -= 4*offset; if (c.z == 0.0) a.x += 4*offset;
  }else{//note that this only catches NULL pointers, not stale ones!!
    std::cerr<<"Some of the atom pointers are invalid in AtomByLAA! (a b c): "<<tetr.a<<","<<tetr.b<<","<<tetr.c<<std::endl;
    exit(EXIT_FAILURE);
  }
  
  l3=tetr.l3;
  ang_1=tetr.ang_1;ang_3=tetr.ang_3;

  //Check if ab || bc, in which case we can't build d_new (not enough eq's):
  if ((fabs(Bangle(a,b,c)+1)<EPS) || (fabs(Bangle(a,b,c)-1)<EPS)) {//check for parallel vectors
    std::cerr << "Can't build atom based on (anti)parallel Vector tetrads!\n"
	      << "Quitting...\n";
    exit(EXIT_FAILURE);
  }
  if (l3==0) {
    std::cerr << "Can't work with Vectors of 0 length(LAA)!\n"
	 << "Quitting..\n";
    exit(EXIT_FAILURE);
  }
  /*
  //debug
  cout<<"AtomByLAA:----------------------------------"<<endl;
  cout<<" a:"<<a.x<<"  "<<a.y<<"  "<<a.z<<endl;
  cout<<" b:"<<b.x<<"  "<<b.y<<"  "<<b.z<<endl;
  cout<<" c:"<<c.x<<"  "<<c.y<<"  "<<c.z<<endl;
  cout<<"l1: "<<tetr.l1<<"  l1: "<<Blength(a,b)<<endl;
  cout<<"l2: "<<tetr.l2<<"  l2: "<<Blength(b,c)<<endl;
  cout<<"abc: "<<tetr.ang_1<<" abc: "<<acos(Bangle(a,b,c))<<endl;
  */
  cos_3=cos(ang_3);
  sin_3=sin(ang_3);
  a -= b;
  c -= b;
  a.Normalize();
  c.Normalize();
  d = a;
  n = Vector::Cross(a,c);
  n.Normalize();
  d.Rotate(n,ang_1);
  d *= sin_3;
  d += cos_3 * n;
  //  d.Normalize();
  d *= l3;
  d += b;
  /*
  //debug
  cout<<" d:"<<d.x<<"  "<<d.y<<"  "<<d.z<<endl;
  cout<<"l3: "<<tetr.l3<<"  l3: "<<Blength(c,d)<<endl;
  cout<<"Dih.angle: "<<Dihedral(a,b,c,d)<<" chi:"<<tetr.chi<<endl;
  cout<<"bcd: "<<tetr.ang_3<<" bcd: "<<acos(Bangle(b,c,d))<<endl;
  cout<<endl;
  */
  return d;
}


//This function returns a  root of the quadratic eq'n a1*x^2+a2*x+a3=0.
//The root is positive or negative depending on hand_flag.
double Quadr(double a1,double a2,double a3,int hand_flag){
  if (a1!=0) {
    if (hand_flag>0) return (-a2+sqrt(fabs(a2*a2-4*a1*a3)))/(2*a1);
    if (hand_flag<0) return (-a2-sqrt(fabs(a2*a2-4*a1*a3)))/(2*a1);
  }else{
    std::cerr << "This eq'n is NOT quadratic: a1=0!\n"
	 << "Quitting..\n";
    exit(EXIT_FAILURE);
  }
  return -1;
}


//This function calculates atom's coordinates given 2 bond angles and 1 bond length relative
//to 3 atoms whose new (and old) coordinates are known.
Vector AtomByLAD(tetrad tetr){
  Vector d,n;
  //tetrad expansion:
  Vector a,b,c;
  double l1,l2,l3,l3_full;
  double chi,ang_3;
  double cos_3,sin_3;
  const double offset = -1e-5;
  
  //Reassign the tetrad:
  if (tetr.a != NULL && tetr.b != NULL && tetr.c != NULL) {
    a = *tetr.a;
    b = *tetr.b;
    c = *tetr.c;
    //Offset all 0 coord. you are building FROM by a small amount:
    if (a.x == 0.0) a.x += offset; if (a.y == 0.0) a.x -= offset; if (a.z == 0.0) a.x += 2*offset;
    if (b.x == 0.0) a.x -= 2*offset; if (b.y == 0.0) a.x += 3*offset; if (b.z == 0.0) a.x -= 3*offset;
    if (c.x == 0.0) a.x += 3*offset; if (c.y == 0.0) a.x -= 4*offset; if (c.z == 0.0) a.x += 4*offset;
  }else{//note that this only catches NULL pointers, not stale ones!!
    std::cerr<<"Some of the atom pointers are invalid in AtomByLAD! (a b c): "<<tetr.a<<","<<tetr.b<<","<<tetr.c<<std::endl;
    exit(EXIT_FAILURE);
  }
  
  l1=tetr.l1;l2=tetr.l2;l3=tetr.l3;l3_full=l3;
  chi=tetr.chi;ang_3=tetr.ang_3;
  
  //Check if bc||cd:
  if (fabs(ang_3)<EPS || fabs(ang_3-PI)<EPS) {
      std::cerr << "Dihedral angle undefined for bc||cd!\n"
		<< "Quitting...\n";
    exit(EXIT_FAILURE);
  }
  //Check if ab || bc, in which case we can't build d_new (not enough eq's):
  if ((fabs(Bangle(a,b,c)+1)<EPS) || (fabs(Bangle(a,b,c)-1)<EPS)) {
    //check for parallel vectors
    std::cerr << "Can't build atom based on (anti)parallel Vector tetrads!\n"
	      << "Quitting...\n";
    exit(EXIT_FAILURE);
  }
  if (l1==0 || l2==0 || l3==0) {
    std::cerr << "Can't work with Vectors of 0 length(LAD)!\n"
	      << "Quitting..\n";
    exit(EXIT_FAILURE);
  }
  

  //Project l3 onto the plane perp.to the dihedral axis:
  cos_3=cos(ang_3);sin_3=sin(ang_3);
  l3=l3*sin_3;

  n = c - b; // n is the torsion bond Vector
  n.Normalize();
  d = a - b; // start d out in the direction of a
  d.ProjectNormal(n); // make it perp to n
  d.Normalize(); // set the length
  d *= l3; //l3 is actually l3*sin(alpha3)
  d += n * (-l3_full*cos_3); // add the component parallel to n
  d += c - b; // translate to c, w/ b as the new origin
  d.Rotate(n,chi);
  d += b; // move origin back to zero
  
  return d;
}

/*  //This function calculates *cos* of bond angle. */
/*  //It terminates if the angle is undefined (one of the vectors is 0 length) */
/*  double Bangle(Vector a,Vector b,Vector c){ */
/*    double tmp; */
/*    a -= b; */
/*    c -= b; */
/*    assert(a.Length()!=0); */
/*    assert(c.Length()!=0); */
/*    tmp=(a * c)/(a.Length()*c.Length()); */
/*    //debug */
/*    //printf("Dot:%lf\n",Dot(a,c)); */
/*    //printf("tmp: %lf\n",tmp); */
/*    return tmp; */
/*  } */

/*  //This function calculates bond length. It returns 0 length rather than terminates. */
/*  double Blength(const Vector &a,const Vector &b){ */
/*    return sqrt((a.x - b.x)*(a.x - b.x) + */
/*  	      (a.y - b.y)*(a.y - b.y) + */
/*  	      (a.z - b.z)*(a.z - b.z)); */
/*  } */

double SqrDist(Vector a, Vector b){
  Vector diff;

  diff.x = a.x - b.x;
  diff.y = a.y - b.y;
  diff.z = a.z - b.z;

  return (diff.x*diff.x + diff.y*diff.y +
	  diff.z*diff.z);
}

int WithinSqrDist(Vector a, Vector b, double dist){
double dx,dy,dz,tot;
int result;

 result=1;
 dx=a.x-b.x;
 tot=dx*dx;
 if (tot<dist)
   {
     dy=a.y-b.y;
     tot+=dy*dy;
     if (tot<dist)
       {
	 dz=a.z-b.z;
	 tot+=dz*dz;
	 if (tot>dist)
	   result=0;
       }
     else
       result=0;
  }
 else
   result=0;
 return result;
}

double Theta2d(Vector a, Vector b, Vector R){
double t,b_a_size,o;
Vector ab_perp;
 
 // take a right triangle with the origin, b,
 // and the projection of b onto a as vertices
 // let o be the side opposite the angle between a and b
 // let b_a_size be the adjcent side
 // the atan2 of o/a gives the angle from a to b
 // na is the normalized Vector a

 a.Normalize();
 // b_a_size is the size of the projection of b onto a
 b_a_size = a * b;
 b.ProjectNormal(a);
 o=b.Length();
 // but is o positive or negative?
 ab_perp = Vector::Cross(a,b);
 if(ab_perp * R < 0){
   o=-o;
 }
 t=atan2(o,b_a_size);
 return t;
} 

//This function calculates *cos* of a dihedral angle between
//2 vectors x1-x2 and x4-x3.
double Dihedral(Vector x1, Vector x2, Vector x3, Vector x4){
// This function returns the dihedral angle between x1,x2,x3,x4
Vector x21_planar,x34_planar,r23;
double l1,l2,l3;
double angle;

  //Find Vector lengths:
  l1=Blength(x1,x2);
  l2=Blength(x2,x3);
  l3=Blength(x3,x4);
  
  if (l1==0 || l2==0 || l3==0) {
    std::cerr << "Can't find dihedral angle for Vectors of zero length!\n"
	      << "Quitting...\n";
    exit(EXIT_FAILURE);
  }
  //Check if x1-x2 is || to x1-x3:
  if ((fabs(Bangle(x1,x2,x3)+1)<EPS) || (fabs(Bangle(x1,x2,x3)-1)<EPS) ||
      (fabs(Bangle(x2,x3,x4)+1)<EPS) || (fabs(Bangle(x2,x3,x4)-1)<EPS)) {
    std::cerr << "Dihedral angle undefined for (anti)parallel Vectors!\n"
	      << "Quitting...\n";
    exit(EXIT_FAILURE);
  }
  
  // Set r23 to be the axis of rotation
  // The Vector from x2 to x3 
  r23 = x3 - x2;
  // x21_planar is the projection of the Vector from x2 to x1
  // into the plane normal to r23
  x21_planar = x1 - x2;
  x21_planar.ProjectNormal(r23);
  // x34_planar is the projection of the Vector from x3 to x4
  // into the plane normal to r23
  x34_planar = x4 - x3;
  x34_planar.ProjectNormal(r23);
  
  //cos_chi=cos(Theta2d(x21_planar,x34_planar,r23));
  //return cos_chi;
  angle=Theta2d(x21_planar,x34_planar,r23);
  return angle;
}


Vector ApplyRotation(Vector v, Vector R, double theta){
//  This function rotates the vector v around the axis R by theta radians.
//  The use of static variables here means we don't have to recompute the
//  rotation matrix each time we call the function, if R and theta stay
//  the same.
static Vector pos,A1,A2,A3,last_r(0,0,0);
static double cost,sint,d,Rx2,Ry2,Rz2,last_theta=0;

 if((last_theta!=theta)||(last_r.x!=R.x)||
    (last_r.y!=R.y)||(last_r.z!=R.z)){
   last_theta=theta;
   last_r=R;
   cost=cos(theta);
   sint=-sin(theta);
   //  This matrix was originally computed by Ed Thayer, and double-checked
   //  by me.  It is the product of a basis transformation into a basis in which
   //  R is one of the coordinate axes, a simple rotation around R, and
   //  a basis transformation back again.
   Rx2=R.x*R.x;  Ry2=R.y*R.y;  Rz2=R.z*R.z;
   d=Rx2+Ry2;
   A1.x=Rx2+(Ry2+Rx2*Rz2)*cost/d;
   A1.y=R.x*R.y*(1-cost)+R.z*sint;
   A1.z=R.x*R.z*(1-cost)-R.y*sint;
   A2.x=R.x*R.y*(1-cost)-R.z*sint;
   A2.y=Ry2+(Rx2+Ry2*Rz2)*cost/d;
   A2.z=R.y*R.z*(1-cost)+R.x*sint;
   A3.x=R.x*R.z*(1-cost)+R.y*sint;
   A3.y=R.y*R.z*(1-cost)-R.x*sint;
   A3.z=Rz2+d*cost;
 }
 //  This is just an obscure way of writing the product of matrix A
 //  with Vector v.
 pos.x=Dot(v,A1);
 pos.y=Dot(v,A2);
 pos.z=Dot(v,A3);
 return pos;
}

tetrad AssignLAA(Vector *_a, Vector *_b, Vector *_c, Vector *_d){
  tetrad tetr;
  int hand_flag;
  Vector a=*_a, b=*_b, c=*_c, d=*_d, n;

  tetr.a = _a;
  tetr.b = _b;
  tetr.c = _c;

  tetr.l1 = 0;
  tetr.l2 = 0;
  tetr.l3 = Blength(b,d);
  tetr.chi = 4;// this field is not used

  a -= b;
  c -= b;
  d -= b;

  a.Normalize();
  c.Normalize();
  d *= 1/tetr.l3;
  n = Vector::Cross(a,c);
  n.Normalize();
  tetr.ang_3 = acos(d * n);
  d.ProjectNormal(n);
  d.Normalize();
  tetr.ang_1 = acos(d * a);
  hand_flag = ((n * Vector::Cross(a,d)) > 0);
  if(hand_flag == 0){
    tetr.ang_1*=(-1);
  }

  return tetr;
}

tetrad AssignLAD(Vector *_a, Vector *_b, Vector *_c, Vector *_d){
  tetrad tetr;
  Vector a=*_a, b=*_b, c=*_c, d=*_d;

  tetr.a = _a;
  tetr.b = _b;
  tetr.c = _c;

  tetr.chi = Dihedral(a,b,c,d);
  tetr.ang_1 = 4;// this field is not used
  tetr.ang_3 = acos(Bangle(b,c,d));
  tetr.l1 = Blength(a,b);
  tetr.l2 = Blength(b,c);
  tetr.l3 = Blength(c,d);

  return tetr;
}



