/* matrix.c */

/*****************************************************************
        Copyright by Rockefeller University,
can not be reproduced or distributed without written permission of
copyright holder.  Version of April 2006.

Written by Alexandre Morozov (contact person: morozov@edsb.rockefeller.edu),
and Eric Siggia.

The program DNABEND implements an algorithm for computing elastic energies
for DNA molecules of arbitrary length and sequence. DNA can be constrained
to follow an arbitrary 3D curve, such as a left-handed superhelix found
in nucleosomal core particles.

The DNABEND method and its applications are described in the following
paper:

******************************************************************/

#include "matrix.h"

Matrix3::Matrix3(){
  for(int r=0;r<3;r++){
    for(int c=0;c<3;c++){
      data.data[r][c]=0;
    }
  }
  diagonalized=0;
}

Matrix3::Matrix3(const Matrix3 &oldmatrix){
  for(int r=0;r<3;r++){
    for(int c=0;c<3;c++){
      data.data[r][c]=oldmatrix.data.data[r][c];
      // Set up the identity matrix.
    }
  }
  diagonalized=0;
}


Matrix3::Matrix3(const matrix_type &oldmatrix_type){
    for(int r=0;r<3;r++){
	for(int c=0;c<3;c++){
	    data.data[r][c]=oldmatrix_type.data[r][c];
	    // Set up the identity matrix.
	}
    }
    diagonalized=0;
}

Matrix3::Matrix3(const Vector &v1, const Vector &v2, const Vector &v3){
  data.data[0][0] = v1.x;
  data.data[1][0] = v1.y;
  data.data[2][0] = v1.z;
  data.data[0][1] = v2.x;
  data.data[1][1] = v2.y;
  data.data[2][1] = v2.z;
  data.data[0][2] = v3.x;
  data.data[1][2] = v3.y;
  data.data[2][2] = v3.z;
}

void Matrix3::loadvecs(Vector &v1, Vector &v2, Vector &v3) {
  v1.x = data.data[0][0];
  v1.y = data.data[1][0];
  v1.z = data.data[2][0];
  v2.x = data.data[0][1];
  v2.y = data.data[1][1];
  v2.z = data.data[2][1];
  v3.x = data.data[0][2];
  v3.y = data.data[1][2];
  v3.z = data.data[2][2];
}

//Not 'const Matrix3&' since we also need to multiply
//matrix into itself sometimes.
void Matrix3::RegMultiply(Matrix3 &B){
    data = Matrix_Multiply(data,B.data);
}

void Matrix3::TransMultiply(Matrix3 &B){
    data = Transpose_Multiply(data,B.data);
}

double Matrix3::value(int row, int col){
  assert(row>=0 && row<3);
  assert(col>=0 && col<3);
  return data.data[row][col];
}


void Matrix3::set(int row, int col, double val){
  assert(row>=0 && row<3);
  assert(col>=0 && col<3);
  data.data[row][col]=val;
  diagonalized=0;
}

void Matrix3::transpose(){
  double temp;

  for(int r=0;r<3;r++){
    for(int c=r;c<3;c++){
      temp=data.data[r][c];
      data.data[r][c]=data.data[c][r];
      data.data[c][r]=temp;
    }
  }
  diagonalized=0;
}

double Matrix3::Evalue(int index){
  if(!diagonalized){
    Jacobi();
  }
  return(eigenvalues[index]);
}

Vector Matrix3::Evector(int index){
  if(!diagonalized){
    Jacobi();
  }
  return(eigenvectors[index]);
}

int Matrix3::symmetric(){

  return ((data.data[0][1]==data.data[1][0])&&(data.data[0][2]==data.data[2][0])&&
	  (data.data[1][2]==data.data[2][1]));
}

// Note the Jacobi procedure is for symetric matrices only, and will exit, if a
// non-symmetric matrix is passed in.
void Matrix3::Jacobi(){
  int r,c;
  double eps=1e-10;
  matrix_type rot,V,D;

  assert(symmetric());
  // First initialize V, which will hold our eigenvectors, to I.
  // And set D, which will be diagonalized, to the current matrix initially.
  for(r=0;r<3;r++){
    for(c=0;c<3;c++){
      if(r==c){
	V.data[r][c]=1;
      }else{
	V.data[r][c]=0;
      }
      D.data[r][c]=data.data[r][c];
    }
  }
  // While the off diagonal elements are still large...
  while(Off(D)>eps){
    for(r=0;r<3;r++){
      for(c=r+1;c<3;c++){
	// Zero them independently.
	// Although each step unbdoes the previous zeroing, the sum of the off-diagonals
	// still converges on zero => Thus the Jacobi method.
	// schur_decomp is just the rotation in the r-c plane that zeroes element r,c.
	Schur(r,c,D,rot);
	// V = V * Schur.
	V=Matrix_Multiply(V,rot);
	// D = Schur(Transpose) * D * Schur
	D=Matrix_Multiply(D,rot);
       	D=Transpose_Multiply(rot,D);
      }
    }
  }
  eigenvalues[0]=D.data[0][0];
  eigenvalues[1]=D.data[1][1];
  eigenvalues[2]=D.data[2][2];
  eigenvectors[0].x=V.data[0][0];
  eigenvectors[0].y=V.data[1][0];
  eigenvectors[0].z=V.data[2][0];
  eigenvectors[1].x=V.data[0][1];
  eigenvectors[1].y=V.data[1][1];
  eigenvectors[1].z=V.data[2][1];
  eigenvectors[2].x=V.data[0][2];
  eigenvectors[2].y=V.data[1][2];
  eigenvectors[2].z=V.data[2][2];
  diagonalized=1;
}

// Compute the rotation in the r-c plane needed to zero element A(r,c).
void Matrix3::Schur(int p, int q, matrix_type &D, matrix_type &rot){
  double t,T;
  double s,c;

  // Initialize our rotation matrix to I.
  for(int row=0;row<3;row++){
    for(int col=0;col<3;col++){
      if(row==col){
	rot.data[row][col]=1;
      }else{
	rot.data[row][col]=0;
      }
    }
  }

  // See "Matrix Calculations" by Golub & Van Loan for details on the Schur
  // decomposition, or refer to Numerical recipies.
  T=(D.data[q][q]-D.data[p][p])/(2*D.data[p][q]);
  t=1/(fabs(T)+sqrt(1+T*T));
  if(T<0){
    t=-t;
  }
  c=1/sqrt(1+t*t);
  s=t*c;
  rot.data[p][p]=c;
  rot.data[q][q]=c;
  rot.data[p][q]=s;
  rot.data[q][p]=-s;
}

double Matrix3::Off(matrix_type A){
  double sum=0;
  int r,c;

  for(r=0;r<3;r++){
    for(c=0;c<3;c++){
      if(!r==c){
	sum+=A.data[r][c]*A.data[r][c];
      }
    }
  }
  return sum;
}

// Multiply A on the left of B.
matrix_type Matrix3::Matrix_Multiply(matrix_type A, matrix_type B){
  matrix_type P;

  for(int r=0;r<3;r++){
    for(int c=0;c<3;c++){
      P.data[r][c]=A.data[r][0]*B.data[0][c]+A.data[r][1]*B.data[1][c]+A.data[r][2]*B.data[2][c];
    }
  }
  return P;
}

// Multiply the transpose of A on the left of B.
matrix_type Matrix3::Transpose_Multiply(matrix_type A, matrix_type B){
  matrix_type P;

  for(int r=0;r<3;r++){
    for(int c=0;c<3;c++){
      P.data[r][c]=A.data[0][r]*B.data[0][c]+A.data[1][r]*B.data[1][c]+A.data[2][r]*B.data[2][c];
    }
  }
  return P;
}

// Multiply this matrix into a vector: Ax, return resulting vector.
Vector Matrix3::VecMultiply(Vector &v) {
  Vector b(0.0,0.0,0.0);
  b.x = data.data[0][0]*v.x + data.data[0][1]*v.y + data.data[0][2]*v.z;
  b.y = data.data[1][0]*v.x + data.data[1][1]*v.y + data.data[1][2]*v.z;
  b.z = data.data[2][0]*v.x + data.data[2][1]*v.y + data.data[2][2]*v.z;
  return b;
}

// Multiply all elements of this matrix by a number:
void Matrix3::multiply(double num) {
  int r,c;
  for(r=0;r<3;r++){
    for(c=0;c<3;c++){
      data.data[r][c] *= num;
    }
  }
}

// Subtract input matrix from this matrix, element by element:
void Matrix3::subtract(Matrix3 &m) {
  int r,c;
  for(r=0;r<3;r++){
    for(c=0;c<3;c++){
      data.data[r][c] -= m.data.data[r][c];
    }
  }
}

void Matrix3::Print_data(matrix_type M){
  int r,c;

  for(r=0;r<3;r++){
    std::cout<<"[  ";
    for(c=0;c<3;c++){
      std::cout<<" "<<M.data[r][c]<<" ";
    }
    std::cout<<"]\n";
  }
  std::cout<<std::endl;
}

void Matrix3::Print(){
  Print_data(data);
}

