/* minimize.c */

/*****************************************************************
        Copyright by Rockefeller University,
can not be reproduced or distributed without written permission of
the copyright holder.  Version 1.1 of April 2006.

Written by Alexandre Morozov (contact person: morozov@edsb.rockefeller.edu),
and Eric Siggia.

The program DNABEND implements an algorithm for computing elastic energies
for DNA molecules of arbitrary length and sequence. DNA can be constrained
to follow an arbitrary 3D curve, such as a left-handed superhelix found
in nucleosomal core particles.

The DNABEND method and its applications are described in the following
paper:

******************************************************************/

#include "minimize.h"

// This is a minimizer class to carry out explicit minimization of the nucleosome energy
// function using GSL tools.

Minimizer::Minimizer() {
  DIM = Lseq = 0;
  x = NULL;
  x0 = NULL;
  ind = NULL;
  disp0 = NULL;
  F_SH = NULL;
  F_WO = NULL;
  dispWO = NULL;
  //xyz_ideal = NULL;
  //xyz_wt = NULL;
  A = NULL;
  b = NULL;
  Ewo_vec = NULL;
}

Minimizer::~Minimizer() {
  if (x != NULL) gsl_vector_free(x);
  if (x0 != NULL) gsl_vector_free(x0);
  if (ind != NULL) delete [] ind;
  if (disp0 != NULL) delete [] disp0;
  if (F_SH != NULL) gsl_matrix_free(F_SH);
  if (F_WO != NULL) gsl_matrix_free(F_WO);
  if (dispWO != NULL) gsl_vector_free(dispWO);
  //if (xyz_ideal != NULL) delete xyz_ideal;
  //if (xyz_wt != NULL) delete xyz_wt;
  if (A != NULL) gsl_matrix_free(A);
  if (b != NULL) gsl_vector_free(b);
  if (Ewo_vec != NULL) gsl_vector_free(Ewo_vec);
}

// seq: input seq;
// prms: WO means, stdevs & force constants;
// xyz_ideal: ideal set of frames + WO prms;
// xyz_wt: wt set of frames + WO prms (eg from 1kx5).
Minimizer::Minimizer(Sequence::Sequence *seq, DNAparams::DNAparams &prms,
		     Geom::Geom &xyz_ideal, Geom::Geom &xyz_wt,
		     struct Options *opt, std::string const &pdbout, std::string const &geomout,
		     std::string const &runmode) {
  // Check runmode:
  if (runmode != "silent" && runmode != "verbose") {
    std::cerr << "ERROR in Minimizer::Minimizer => invalid runmode: "<<runmode<<"!\n";
    exit(EXIT_FAILURE);
  }
  // Set weights:
  F0 = opt->w_el;
  F6 = opt->w_sh;
  F3 = opt->w_sum;
  F7 = opt->w_dev;
  // Set dimensions using input seq:
  Lseq = seq->Length()-1; // seq length **!! in dinucs !!**
  DIM = Lseq * Nprm;
  // Check seq:
  if (Lseq <= 0) {
    std::cerr << "ERROR in Minimizer::Minimizer => invalid input seq!\n";
    exit(EXIT_FAILURE);
  }
  // Check dims:
  if (Lseq != xyz_ideal.frames.size() - 1 || Lseq != xyz_wt.frames.size() - 1) {
    std::cerr << "ERROR in Minimizer::Minimizer => invalid dims of ideal/wt frames!\n";
    exit(EXIT_FAILURE);
  }
  // Allocate memory for the components of Ewo:
  Ewo_vec = gsl_vector_calloc(Lseq);
  // Allocate memory for the linear system of equations:
  A = gsl_matrix_calloc(DIM,DIM);
  b = gsl_vector_calloc(DIM);
  // Allocate memory for x (solution of the linear system of equations) & x0 (initial guess = wt):
  x = gsl_vector_calloc(DIM);
  x0 = gsl_vector_calloc(DIM); // local copy of initial prms
  // Build dinuc index lookup table from seq:
  ind = new int[Lseq];
  for (unsigned int i=0; i<Lseq; ++i) {
    char ch1 = seq->CharAt(i);
    char ch2 = seq->CharAt(i+1);
    int ind_cur = get_basestep_index(ch1,ch2);
    ind[i] = ind_cur;
  }
  // Re-write ideal frames with wt frames because nucleosomal tails are not superhelical:
  //ImportTails(xyz_ideal,xyz_wt,17,131);
  // Get WO prms from a set of ideal frames:
  gsl_vector *v = xyz_ideal.GetGeom();
  //BStable bst_ideal_test(v,seq);
  //bst_ideal_test.WriteBStable("WOprms_ideal.dat");
  disp0 = xyz_ideal.global_disp(); // this is simply ideal radius-vectors in the global frame
  // !! Translate ideal radius-vectors to the wt origin:
  Vector offset = (disp0[0] - xyz_wt.frames.at(0).orig);
  for (unsigned int i=0; i<xyz_ideal.frames.size(); ++i) {
    disp0[i] -= offset;
  }
  // Global differential displacements for the ideal frames:
  Vector *diff0 = xyz_ideal.disp();
  // Transform WO force constants & equil. displacements into the global frame using ideal superhelix:
  F_WO = gsl_matrix_calloc(DIM,DIM);
  dispWO = gsl_vector_calloc(DIM);
  //LocalizeWO(xyz_wt,prms); // this sets x0, x[=x0], F_WO & dispWO to their local values
  //debug
  //std::cout << "E_WO local: "<<E_WO()<<std::endl;
  GlobalizeWO(xyz_ideal,xyz_wt,prms); // this updates x & x0 and makes F_WO & dispWO global
  // Compute curvature coeffs imposed by the superhelix:
  F_SH = gsl_matrix_calloc(DIM,DIM); // note that deviations from 0.0 are penalized, so there is no offset similar to dispWO
  gsl_matrix_memcpy(F_SH,xyz_ideal.GetCurv() );
  //debug
  //std::cout << "E_WO global: "<<E_WO()<<std::endl;
  //exit(0);
  // Set A & b in Ax = b:
  //setWOcoeff();
  //setWOcoeff(v); // now with the superhelix :)
  setWOcoeff(diff0,v); // now with the superhelical offsets :)
  setDisplacementConstr();
  //setIdealConstr(v);
  //setSumConstr();
  // Either solve Ax = b or minimize the energy using conjugate grad:
  //RunG(); // conj grad - needs updating!!
  Solve(); // linear eqs
  UnshiftX(diff0,v); // add ideal values back to the solution (required for making PDBs etc.)
  // This simply writes global radius-vectors (stored in x before [shift, slide, rise] are reconstructed):
  //PDB("junk.pdb",xyz_ideal.frames.at(0).orig);
  ReconstructWO(xyz_ideal); // get WO prms from minimized global displacements using ideal frames
  BStable bst_fin(getresult(),seq); // table of final WO prms from x (angles in Deg!)
  if (geomout != "none") bst_fin.WriteBStable(geomout); // write final set of WO prms
  Geom xyz_fin(bst_fin,xyz_ideal.frames.at(0) ); // final set of xyz frames from the table of WO prms
  if (pdbout != "none") xyz_fin.PDBunitvecs(pdbout); // write final PDB
  // Compute and save energies:
  Ewo = E_WO(prms,1.0); // elastic (WO)
  if (runmode == "verbose") {
    printf("E_WO = %.3f\n",Ewo);
    for (unsigned int i=0; i<Lseq; ++i) {
      printf("%.3f ",gsl_vector_get(Ewo_vec,i) );
    }
    printf("\n");
  }
  //double rms = xyz_fin.RMSD(xyz_wt); // compare with the wt frames
  rms = xyz_fin.RMSD(xyz_ideal); // compare with the ideal frames
  if (runmode == "verbose") printf("Econstr = %.3f\n",rms * (Lseq + 1) );
  Etotal = F0 * Ewo + F6 * rms * (Lseq + 1); // rms * (Lseq + 1) is the constraint energy wrt the ideal superhelix
  //Etotal = F0 * Ewo + F6 * rms;
  if (runmode == "verbose") {
    printf("Etot(F0=%.3f,F6=%.g) = %.3f\n",F0,F6,Etotal); // total weighted energy
    std::cout << "RMSD(fin,ideal) = " << rms << std::endl; // RMSD to fixed ideal
  }
  /****************/
  gsl_vector_free(v);
  delete [] diff0;
}

// This function returns value of the elastic (WO) energy.
double Minimizer::getEwo(void) {
  return Ewo;
}

// This function returns value of the total (weighted) energy.
double Minimizer::getEtotal(void) {
  return Etotal;
}

// This function returns value of the rms to ideal frames.
double Minimizer::getRMS(void) {
  return rms;
}

// Import wt tails into ideal tails by rewriting ideal frames at
// [0,start-1] and [end-1,Lseq-1].
void Minimizer::ImportTails(Geom::Geom &xyz_ideal, Geom::Geom &xyz_wt,
			    unsigned int start, unsigned int end) {
  for (unsigned int i1=0; i1<start; ++i1) { // head
    xyz_ideal.frames.at(i1) = xyz_wt.frames.at(i1);
  }
  for (unsigned int i2=end-1; i2<Lseq; ++i2) { // tail
    xyz_ideal.frames.at(i2) = xyz_wt.frames.at(i2);
  }
  // Update WO prms in xyz_ideal:
  xyz_ideal.TranslateGeom();
}

// Set F_WO & dispWO according to WO tables:
void Minimizer::LocalizeWO(Geom::Geom &xyz_wt, DNAparams::DNAparams &prms) {
  for (unsigned int i=0; i<Lseq; ++i) {
    int indx = ind[i];
    for (unsigned int j1=0; j1<Nprm; ++j1) {
      int linind1 = i*Nprm + j1;
      gsl_vector_set(dispWO,linind1,prms.means_bs[indx][j1]);
      for (unsigned int j2=0; j2<Nprm; ++j2) {
	int linind2 = i*Nprm + j2;
	gsl_matrix_set(F_WO,linind1,linind2,prms.Fij[indx][j1][j2]);
      }
    }
  }
  // Get initial WO prms from the wt BStable (stored in xyz_wt):
  for (unsigned int i=0; i<Lseq; ++i) {
    for (unsigned int j=0; j<Nprm; ++j) {
      int linind = i*Nprm + j;
      gsl_vector_set(x0,linind,xyz_wt.getm(i,j) );
    }
  }
  gsl_vector_memcpy(x,x0); // copy of initial prms: x = x0
}

// This function transforms WO force constants & equil. displacements into the global frame:
// NOTE: initial guess (ie local displacements from input Bstable) is transformed into the global
// frame using its own set of unit vecs. This is done to improve the initial guess by avoiding
// cumulative errors that occur when wt coords are rebuilt using ideal superhelical frames. 
void Minimizer::GlobalizeWO(Geom::Geom &ideal, Geom::Geom &wt, DNAparams::DNAparams &prms) {
  // Check input frames:
  if (ideal.frames.size()-1 != ideal.frames_mst.size() ) { // ideal
    std::cerr << "ERROR in GlobalizeWO: mismatch in ideal input frames!Exiting ..\n";
    exit(EXIT_FAILURE);
  }
  if (wt.frames.size()-1 != wt.frames_mst.size() ) { // wt
    std::cerr << "ERROR in GlobalizeWO: mismatch in wt input frames!Exiting ..\n";
    exit(EXIT_FAILURE);
  }
  // Set up the *global* coeffs for WO potential using a set of input MST frames:
  for (unsigned int i=0; i<Lseq; ++i) {
    int indx = ind[i];
    // Construct 4 3x3 matrices with original WO coeffs & then transform them into the global frame:
    Matrix3 AA,AD,DA,DD; // angle-angle, angle-diplacement, displacement-angle etc.
    for (unsigned int j1=0; j1<3; ++j1) {
      for (unsigned int j2=0; j2<3; ++j2) {
	AA.set(j1, j2, prms.Fij[indx][j1][j2]);
      }
    }
    for (unsigned int j1=0; j1<3; ++j1) {
      for (unsigned int j2=3; j2<6; ++j2) {
	AD.set(j1, j2-3, prms.Fij[indx][j1][j2]);
      }
    }
    for (unsigned int j1=3; j1<6; ++j1) {
      for (unsigned int j2=0; j2<3; ++j2) {
	DA.set(j1-3, j2, prms.Fij[indx][j1][j2]);
      }
    }
    for (unsigned int j1=3; j1<6; ++j1) {
      for (unsigned int j2=3; j2<6; ++j2) {
	DD.set(j1-3, j2-3, prms.Fij[indx][j1][j2]);
      }
    }
    // MST frames:
    Matrix3 R_DA(ideal.frames_mst.at(i).a, ideal.frames_mst.at(i).b, ideal.frames_mst.at(i).c);
    Matrix3 R_DD(R_DA),Rt(R_DA),R_WO(R_DA);
    Rt.transpose();
    // Matrix multiplication:
    AD.RegMultiply(Rt); // FR^{-1}
    R_DA.RegMultiply(DA); // RF
    DD.RegMultiply(Rt);
    R_DD.RegMultiply(DD); // RFR^{-1}
    //debug
    //AD.Print();
    //exit(0);
    // find RHS in the global frame:
    Vector WO(prms.means_bs[indx][3],prms.means_bs[indx][4],prms.means_bs[indx][5]); // vector of local WO prms
    WO = R_WO.VecMultiply(WO);
    // vector of initial local prms stored in x0:
    Matrix3 R_wt(wt.frames_mst.at(i).a, wt.frames_mst.at(i).b, wt.frames_mst.at(i).c);
    Vector x_glob(gsl_vector_get(x0,i*Nprm+3),gsl_vector_get(x0,i*Nprm+4),gsl_vector_get(x0,i*Nprm+5) );
    //x_glob = R_WO.VecMultiply(x_glob);
    x_glob = R_wt.VecMultiply(x_glob);
    //debug
    //printf("%d: x_glob = [%.2f  %.2f  %.2f]\n",i+1,x_glob.x,x_glob.y,x_glob.z);
    //printf("%d: WO = [%.2f  %.2f  %.2f]\n",i+1,WO.x,WO.y,WO.z);
    //Vector diff = x_glob - WO;
    //printf("%d: |diff| = %.2f\n",i+1,diff.Length() );
    // Insert 4 matrices into A, and update the RHS:
    for (unsigned int j1=0; j1<Nprm; ++j1) {
      int linind1 = i*Nprm + j1;
      // Force constants:
      for (unsigned int j2=0; j2<Nprm; ++j2) {
	int linind2 = i*Nprm + j2;
	double coeff;
	if (j1<3 && j2<3) { // AA
	  coeff = AA.value(j1,j2);
	}else if (j1<3 && j2>=3) { // AD
	  coeff = AD.value(j1,j2-3);
	}else if (j1>=3 && j2<3) { // DA
	  coeff = R_DA.value(j1-3,j2);
	}else if (j1>=3 && j2>=3) { // DD
	  coeff = R_DD.value(j1-3,j2-3);
	}
	gsl_matrix_set(F_WO,linind1,linind2,coeff);
      }
      // Equil prms:
      double WO_prm,glob_prm;
      if (j1<3) {
	WO_prm = prms.means_bs[indx][j1];
	glob_prm = gsl_vector_get(x0,linind1);
      }else if (j1==3) {
	WO_prm = WO.x;
	glob_prm = x_glob.x;
      }else if (j1==4) {
	WO_prm = WO.y;
	glob_prm = x_glob.y;
      }else if (j1==5) {
	WO_prm = WO.z;
	glob_prm = x_glob.z;
      }
      gsl_vector_set(x0,linind1,glob_prm);
      gsl_vector_set(dispWO,linind1,WO_prm);
    }
  }
  // Since this is done before minimization, all values in x0 are initial global values, and thus
  // should be copied into x as well:
  gsl_vector_memcpy(x,x0); // copy of initial global prms: x = x0
}

// This function computes WO energy given a set of WO prms stored in x.
// Instead of using internally stored globalized force constants etc. it
// employs local WO coeffs stored in external tables.
double Minimizer::E_WO(DNAparams::DNAparams &prms) {
  double E = 0.0; // total energy of the conformation
  
  for (unsigned int i=0; i<Lseq; ++i) {
    double Edinuc = 0.0;
    int indx = ind[i];
    for (unsigned int j1=0; j1<Nprm; ++j1) {
      int linind1 = i*Nprm + j1;
      for (unsigned int j2=0; j2<Nprm; ++j2) {
	int linind2 = i*Nprm + j2;
	double dE = F0 * 0.5 * prms.Fij[indx][j1][j2] * 
	  (getx(linind1) - prms.means_bs[indx][j1]) * (getx(linind2) - prms.means_bs[indx][j2]);
	E += dE;
	Edinuc += dE;
	//debug
	//if ((j1==1 && j2==4) || (j1==4 && j2==1) ) {
	//  printf("Fij = %.3f, d1 = %.3f, d2 = %.3f, dE = %.3f\n",prms.Fij[indx][j1][j2],
	//	 getx(linind1) - prms.means_bs[indx][j1],getx(linind2) - prms.means_bs[indx][j2],dE);
	//}
      }
    }
    gsl_vector_set(Ewo_vec,i,Edinuc);
  }
  
  return E;
}

// This function computes WO energy given a set of WO prms stored in x.
// Instead of using internally stored globalized force constants etc. it
// employs local WO coeffs stored in external tables.
// User provides overall weight for the energy.
double Minimizer::E_WO(DNAparams::DNAparams &prms, const double weight) {
  double E = 0.0; // total energy of the conformation
  
  for (unsigned int i=0; i<Lseq; ++i) {
    double Edinuc = 0.0;
    int indx = ind[i];
    for (unsigned int j1=0; j1<Nprm; ++j1) {
      int linind1 = i*Nprm + j1;
      for (unsigned int j2=0; j2<Nprm; ++j2) {
	int linind2 = i*Nprm + j2;
	double dE = weight * 0.5 * prms.Fij[indx][j1][j2] * 
	  (getx(linind1) - prms.means_bs[indx][j1]) * (getx(linind2) - prms.means_bs[indx][j2]);
	E += dE;
	Edinuc += dE;
      }
    }
    gsl_vector_set(Ewo_vec,i,Edinuc);
  }
  
  return E;
}

// This function computes WO energy given a set of WO prms stored in x.
double Minimizer::E_WO(void) {
  double E = 0.0; // total WO energy of the conformation
  
  for (unsigned int i=0; i<Lseq; ++i) {
    double Edinuc = 0.0;
    for (unsigned int j1=0; j1<Nprm; ++j1) {
      int linind1 = i*Nprm + j1;
      for (unsigned int j2=0; j2<Nprm; ++j2) {
	int linind2 = i*Nprm + j2;
	double dE = 0.5 * gsl_matrix_get(F_WO,linind1,linind2) * (getx(linind1) - gsl_vector_get(dispWO,linind1) )
	  * (getx(linind2) - gsl_vector_get(dispWO,linind2) );
	E += dE;
	Edinuc += dE;
      }
    }
    gsl_vector_set(Ewo_vec,i,Edinuc);
  }
  
  return E;
}

// This function computes derivatives of the WO energy.
// Note that memory for g is allocated outside.
void Minimizer::E_WO_df(gsl_vector *g) {
  assert(x->size == g->size); // check dims
  for (unsigned int i=0; i<Lseq; ++i) {
    for (unsigned int j1=0; j1<Nprm; ++j1) {
      int linind1 = i*Nprm + j1;
      double tmp = 0.0;
      for (unsigned int j2=0; j2<Nprm; ++j2) {
	int linind2 = i*Nprm + j2;
	tmp += gsl_matrix_get(F_WO,linind1,linind2) * (getx(linind2) - gsl_vector_get(dispWO,linind2) );
      }
      gsl_vector_set(g,linind1,tmp);
    }
  }
}

// This function adds constraints to Ax = b to control sums of WO angle prms.
void Minimizer::setSumConstr(void) {
  double WOmean;
  
  for (unsigned int k=0; k<3; ++k) {
    if (k==0) {
      WOmean = Omega_0;
    }else if (k==1) {
      WOmean = Roll_0;
    }else if (k==2) {
      WOmean = Tilt_0;
    }
    for (unsigned int i1=0; i1<Lseq; ++i1) {
      int linind_1 = i1*Nprm+k; // j=0,1,2: twist,roll,tilt
      double bupdated = gsl_vector_get(b,linind_1);
      for (unsigned int i2=0; i2<Lseq; ++i2) {
	int linind_2 = i2*Nprm+k; // j=0,1,2: twist,roll,tilt
	double Aupdated = 2 * F3 + gsl_matrix_get(A,linind_1,linind_2);
	gsl_matrix_set(A,linind_1,linind_2,Aupdated);
	bupdated += 2 * F3 * WOmean;
      }
      gsl_vector_set(b,linind_1,bupdated);
    }
  }
}

// This function adds local constraints to Ax = b (angles only!).
// The constraints are based on the ideal set of frames.
// This function adds Bstable constraints to Ax = b.
void Minimizer::setIdealConstr(gsl_vector *bst_ideal) {
  double Fcur;
  double scaling_factors[Nprm] = {0, 1, 3, 0, 0, 0}; // WO prm dependent scaling factors for the overall weight
  
  for (unsigned int i=0; i<Lseq; ++i) {
    for (unsigned int j=0; j<Nprm; ++j) {
      if (scaling_factors[j] == 0) continue;
      Fcur = F7 * scaling_factors[j];
      int linind = i*Nprm + j;
      double Aold = gsl_matrix_get(A,linind,linind);
      gsl_matrix_set(A,linind,linind,2*Fcur + Aold);
      double bold = gsl_vector_get(b,linind);
      gsl_vector_set(b,linind,2*Fcur*gsl_vector_get(bst_ideal,linind) + bold);
    }
  }
}

// This function adds WO prms to Ax = b.
void Minimizer::setWOcoeff(void) {
  // Set up coeffs for WO potential:
  for (unsigned int i=0; i<Lseq; ++i) {
    for (unsigned int j1=0; j1<Nprm; ++j1) {
      int linind1 = i*Nprm + j1;
      double tmp = 0.0;
      for (unsigned int j2=0; j2<Nprm; ++j2) {
	int linind2 = i*Nprm + j2;
	double Aold = gsl_matrix_get(A,linind1,linind2);
	gsl_matrix_set(A,linind1,linind2,F0*gsl_matrix_get(F_WO,linind1,linind2) + Aold);
	tmp += F0 * gsl_matrix_get(F_WO,linind1,linind2) * gsl_vector_get(dispWO,linind2);
      }
      double bold = gsl_vector_get(b,linind1);
      gsl_vector_set(b,linind1,tmp + bold);
    }
  }
  /*
  //debug
  //printf("======= Printing b =======\n");
  //gsl_vector_fprintf(stdout, b, "%6.3f");
  printf("======= Printing A =======\n");
  //gsl_matrix_fprintf(stdout, A, "%6.3f"); // writes it all in one column
  for (unsigned int i=0; i<DIM; ++i) {
    printf("i=%3d: ",i);
    for (unsigned int j=0; j<DIM; ++j) {
      if (gsl_matrix_get(A,i,j) == 0.0) continue;
      printf("%8.5f ",gsl_matrix_get(A,i,j) );
    }
    printf("\n");
  }
  */
}

// This function adds WO prms + superhelical roll & tilt to Ax = b.
void Minimizer::setWOcoeff(gsl_vector *bst_ideal) {
  // Set up coeffs for WO potential:
  for (unsigned int i=0; i<Lseq; ++i) {
    for (unsigned int j1=0; j1<Nprm; ++j1) {
      int linind1 = i*Nprm + j1;
      double tmp = 0.0;
      for (unsigned int j2=0; j2<Nprm; ++j2) {
	int linind2 = i*Nprm + j2;
	double Aold = gsl_matrix_get(A,linind1,linind2);
	gsl_matrix_set(A,linind1,linind2,F0*gsl_matrix_get(F_WO,linind1,linind2) + Aold);
	tmp += F0 * gsl_matrix_get(F_WO,linind1,linind2) * gsl_vector_get(dispWO,linind2);
	//if (j2==1 || j2==2 || j2==3 || j2==4) { // superhelical roll & tilt are added in
	if (j2==1 || j2==2) {
	  tmp += F0 * gsl_matrix_get(F_WO,linind1,linind2) * gsl_vector_get(bst_ideal,linind2);
	}
      }
      double bold = gsl_vector_get(b,linind1);
      gsl_vector_set(b,linind1,tmp + bold);
    }
  }
}

// This function sets up coeffs in Ax = b corresponding to the WO energy & redefines all degrees of freedom
// to mean deviations from the ideal superhelix.
// diff0: global differential displacements in the ideal frame;
// bst_ideal: ideal WO prms.
void Minimizer::setWOcoeff(Vector *diff0, gsl_vector *bst_ideal) {
  // Set up coeffs for WO potential:
  for (unsigned int i=0; i<Lseq; ++i) {
    for (unsigned int j1=0; j1<Nprm; ++j1) {
      int linind1 = i*Nprm + j1;
      double tmp = 0.0;
      for (unsigned int j2=0; j2<Nprm; ++j2) {
	int linind2 = i*Nprm + j2;
	double Aold = gsl_matrix_get(A,linind1,linind2);
	gsl_matrix_set(A,linind1,linind2,F0*gsl_matrix_get(F_WO,linind1,linind2) + Aold);
	double prm_id;
	if (j2==0 || j2==1 || j2==2) {
	  prm_id = gsl_vector_get(bst_ideal,linind2);
	}else if (j2==3) {
	  prm_id = diff0[i].x;
	}else if (j2==4) {
	  prm_id = diff0[i].y;
	}else if (j2==5) {
	  prm_id = diff0[i].z;
	}
	tmp += F0 * gsl_matrix_get(F_WO,linind1,linind2) * (gsl_vector_get(dispWO,linind2) - prm_id);
	// !!!
	//if (j2==1 || j2==2) {
	//  tmp += F0 * gsl_matrix_get(F_WO,linind1,linind2) * gsl_vector_get(bst_ideal,linind2);
	//}
      }
      double bold = gsl_vector_get(b,linind1);
      gsl_vector_set(b,linind1,tmp + bold);
    }
  }
}

// This function adds dislpacement constraints to Ax = b.
void Minimizer::setDisplacementConstr(void) {
  // Add superhelix-imposed force constant coeffs to the potential:
  for (unsigned int i1=0; i1<Lseq; ++i1) {
    for (unsigned int j1=0; j1<Nprm; ++j1) {
      int linind1 = i1*Nprm + j1;
      for (unsigned int i2=0; i2<Lseq; ++i2) {
	for (unsigned int j2=0; j2<Nprm; ++j2) {
	  int linind2 = i2*Nprm + j2;
	  double Aold = gsl_matrix_get(A,linind1,linind2);
	  //gsl_matrix_set(A,linind1,linind2,2*F6*gsl_matrix_get(F_SH,linind1,linind2) + Aold); // this is wrong if F_SH is not symmetric!! (thanks to Kenneth Evans, 10/07)
	  gsl_matrix_set(A,linind1,linind2,F6*(gsl_matrix_get(F_SH,linind1,linind2) + gsl_matrix_get(F_SH,linind2,linind1) ) + Aold);
	}
      }
    }
  }
}

// This function adds the ideal values to unshift the solution stored in x.
// Resulting values are suitable for reconstructing local WO prms, rebuilding
// xyz, etc.
void Minimizer::UnshiftX(Vector *diff0, gsl_vector *bst_ideal) {
  for (unsigned int i=0; i<Lseq; ++i) {
    for (unsigned int j=0; j<Nprm; ++j) {
      int linind = i*Nprm + j;
      double prm_id;
      if (j==0 || j==1 || j==2) {
	prm_id = gsl_vector_get(bst_ideal,linind);
      }else if (j==3) {
	prm_id = diff0[i].x;
      }else if (j==4) {
	prm_id = diff0[i].y;
      }else if (j==5) {
	prm_id = diff0[i].z;
      }
      double oldx = gsl_vector_get(x,linind);
      gsl_vector_set(x,linind,oldx + prm_id);
    }
  }
}

// This function computes constraint energy in the global frame starting from the globalized WO params
// and using *absolute* displacements in energy penalties.
double Minimizer::E_constr(void) {
  double E = 0.0; // total constraint energy of the conformation
  // Compute global constraint energy:
  for (unsigned int i=1; i<Lseq+1; ++i) { // there are Lseq+1 frames, but 1st one is fixed!
    //Vector vg = xyz_wt.frames.at(0).orig; // 1st frame in wt struc (stays fixed during minimization)
    Vector vg = disp0[0];
    for (unsigned int j=0; j<i; ++j) {
      Vector vg_cur(getx(j*Nprm+3),getx(j*Nprm+4),getx(j*Nprm+5) ); // must be globalized!!
      vg += vg_cur;
    }
    E += (vg - disp0[i]) * (vg - disp0[i]);
    //debug
    //fprintf(stdout,"ATOM   %4d %4s %3s %1s %3d    %8.3f%8.3f%8.3f  1.00  0.00\n",
    //    i+1,"O  ","ALA","I",i+1,vg.x,vg.y,vg.z);
    //fprintf(stdout,"ATOM   %4d %4s %3s %1s %3d    %8.3f%8.3f%8.3f  1.00  0.00\n",
    //  i+1,"O  ","ALA","I",i+1,disp0[i].x,disp0[i].y,disp0[i].z);
  }
  //debug
  //exit(0);
  return E;
}

void Minimizer::E_constr_df(gsl_vector *g) {
  assert(x->size == g->size); // check dims
  gsl_vector_set_all(g,0.0);
  for (unsigned int k=0; k<Lseq; ++k) {
    for (unsigned int j2=3; j2<Nprm; ++j2) {
      int linind = k*Nprm + j2;
      double diff_g = 0.0;
      for (unsigned int i=k; i<Lseq; ++i) {
	//Vector vg = xyz_wt.frames.at(0).orig;
	Vector vg = disp0[0];
	for (unsigned int j=0; j<i; ++j) { // build \vec{r}_i
	  Vector vg_cur(getx(j*Nprm+3),getx(j*Nprm+4),getx(j*Nprm+5) ); // must be globalized!!
	  vg += vg_cur;
	}
	if (j2==3) {
	  diff_g += (vg - disp0[i]).x; // disp0[i+1] ??
	}else if (j2==4) {
	  diff_g += (vg - disp0[i]).y; // disp0[i+1] ??
	}else if (j2==5) {
	  diff_g += (vg - disp0[i]).z; // disp0[i+1] ??
	}
      }
      gsl_vector_set(g,linind,2*diff_g);
    }
  }
}

/*
// This function adds displacement constraints to Ax = b.
void Minimizer::setDisplacementConstr(void) {
  for (unsigned int k=0; k<Lseq; ++k) {
    for (unsigned int j=3; j<Nprm; ++j) {
      int linind1 = k*Nprm + j;
      double tmp = 0.0;
      for (unsigned int i=k; i<Lseq; ++i) {
	for (unsigned int i2=0; i2<i; ++i2) {
	  int linind2 = i2*Nprm+j;
	  double Aold = gsl_matrix_get(A,linind1,linind2);
	  gsl_matrix_set(A,linind1,linind2,2*F6 + Aold);
	}
	if (j==3) {
	  tmp += (disp0[i] - xyz_wt.frames.at(0).orig).x;
	}else if (j==4) {
	  tmp += (disp0[i] - xyz_wt.frames.at(0).orig).y;
	}else if (j==5) {
	  tmp += (disp0[i] - xyz_wt.frames.at(0).orig).z;
	}
      }
      double bold = gsl_vector_get(b,linind1);
      gsl_vector_set(b,linind1,2*F6*tmp + bold);
    }
  }
}
*/

// This is an input function to GSL C routines, hence not a class method.
double Etot(const gsl_vector *v, void *params) {
  M_params *prm = (M_params *) params;
  Minimizer *curM = prm->M;
  //DNAparams *DNAprms = prm->DNA;
  // Weights of both terms:
  double F0 = curM->F0;
  double F6 = curM->F6;
  /*
  //debug
  for (unsigned int i=0; i<curM->Lseq; ++i) {
    for (unsigned int j=0; j<curM->Nprm; ++j) {
      int linind = curM->Nprm * i + j;
      if (fabs(gsl_vector_get(v,linind) - curM->getx(linind) ) > 1e-01) { // DoF changed!
	printf("Changed %d %d: from %.3f to %.3f\n",i,j,gsl_vector_get(v,linind),curM->getx(linind) );
      }
    }
  }
  */
  // set new WO prms & recompute energies:
  curM->setresult(v);
  double E1 = curM->E_WO();
  double E2 = curM->E_constr();
  double Etot = (F0*E1 + F6*E2);
  //debug
  //printf("E1 = %.3f, E2 = %.3f\n",F0*E1,F6*E2);
  return Etot; // total weighted E
}

void Etot_df(const gsl_vector *v, void *params, gsl_vector *df) {
  M_params *prm = (M_params *) params;
  Minimizer *curM = prm->M;
  //DNAparams *DNAprms = prm->DNA;
  // Weights of both terms:
  double F0 = curM->F0;
  double F6 = curM->F6;
  // set new WO prms & recompute energy derivs:
  curM->setresult(v);
  gsl_vector *df1 = gsl_vector_calloc(df->size);
  gsl_vector *df2 = gsl_vector_calloc(df->size);
  // WO E:
  curM->E_WO_df(df1);
  gsl_vector_scale(df1,F0); // multiply by F0
  // Constr E:
  curM->E_constr_df(df2);
  gsl_vector_scale(df2,F6); // multiply by F6
  // Set df:
  gsl_vector_memcpy(df,df1);
  gsl_vector_add(df,df2);
  gsl_vector_free(df1);
  gsl_vector_free(df2);
}

void Etot_fdf(const gsl_vector *v, void *params, double *f, gsl_vector *df) {
  *f = Etot(v, params); 
  Etot_df(v, params, df);
}

// Reconstruct [shift, slide, rise] from the global incremental displacements using
// ideal frames.
void Minimizer::ReconstructWO(Geom::Geom &xyz_ideal) {
  //debug
  //for (unsigned int i=0; i<DIM; ++i) {
  //  printf("x0[%d] = %.2f\n",i,gsl_vector_get(x0,i) );
  //}
  //gsl_vector_memcpy(x,x0); // !!!!!!!!
  
  // Make a copy of x; replace all displacements with 0.0:
  gsl_vector *xloc = gsl_vector_calloc(x->size);
  gsl_vector_memcpy(xloc,x);
  for (unsigned int i=0; i<Lseq; ++i) {
    gsl_vector_set(xloc,i*Nprm+3,0.0); // 0 displacements
    gsl_vector_set(xloc,i*Nprm+4,0.0);
    gsl_vector_set(xloc,i*Nprm+5,0.0);
  }
  //Geom xyz_cur(xloc); // get unit vecs from the current set of WO *angles*
  for (unsigned int i=0; i<Lseq; ++i) {
    Vector xglob(getx(i*Nprm+3),getx(i*Nprm+4),getx(i*Nprm+5) );
    //Matrix3 Tmst(xyz_cur.frames_mst.at(i).a,xyz_cur.frames_mst.at(i).b,xyz_cur.frames_mst.at(i).c); // vectors become matrix columns
    // Use the same (ideal) frames to un-transform the vectors!!
    Matrix3 Tmst(xyz_ideal.frames_mst.at(i).a,xyz_ideal.frames_mst.at(i).b,xyz_ideal.frames_mst.at(i).c);
    Tmst.transpose();
    Vector a = Tmst.VecMultiply(xglob);
    setx(i*Nprm+3,a.x); // shift
    setx(i*Nprm+4,a.y); // slide
    setx(i*Nprm+5,a.z); // rise
  }
  //debug
  //for (unsigned int i=0; i<DIM; ++i) {
  //  printf("x[%d] = %.2f\n",i,gsl_vector_get(x,i) );
  //}
  gsl_vector_free(xloc);
}

// orig is the initial vector from which PDB will be built.
void Minimizer::PDB(const std::string filename, Vector orig) {
  FILE *fp = fopen(filename.c_str(),"w");
  
  if (fp == NULL) {
    printf("Error reading file %s\n",filename.c_str() );
    exit(1);
  }
  
  //Vector vg = xyz_wt.frames.at(0).orig;
  Vector vg = orig;
  fprintf(fp,"ATOM   %4d %4s %3s %1s %3d    %8.3f%8.3f%8.3f  1.00  0.00\n",
	  1,"O  ","ALA","I",1,vg.x,vg.y,vg.z);
  for (unsigned int i=0; i<Lseq; ++i) {
    Vector xloc(getx(i*Nprm+3),getx(i*Nprm+4),getx(i*Nprm+5) );
    vg += xloc;
    fprintf(fp,"ATOM   %4d %4s %3s %1s %3d    %8.3f%8.3f%8.3f  1.00  0.00\n",
	    i+2,"O  ","ALA","I",i+2,vg.x,vg.y,vg.z);
  }
  fprintf(fp,"TER\n");
  fclose(fp);
}

// This function solves Ax = b (A & b must be set!)
void Minimizer::Solve(void) {
  int s;
  gsl_permutation *p = gsl_permutation_alloc(DIM);
  gsl_linalg_LU_decomp(A, p, &s);
  gsl_linalg_LU_solve(A, p, b, x);
  //debug
  //printf ("x = \n");
  //gsl_vector_fprintf (stdout, x, "%g");
  
  gsl_permutation_free (p);
}

// This is a central function that runs the conjugate gradient minimizer.
int Minimizer::RunG(void) {
  size_t np = DIM; // number of DoF
  size_t iter = 0;
  int status;
  
  const gsl_multimin_fdfminimizer_type *T =
    gsl_multimin_fdfminimizer_conjugate_fr;
  
  gsl_multimin_fdfminimizer *s = gsl_multimin_fdfminimizer_alloc(T, np);
  
  // Starting point:
  gsl_vector_memcpy(x,x0); // copy of initial prms: x = x0
  
  // Initialize function:
  gsl_multimin_function_fdf F;
  M_params par;
  par.M = this;
  //par.DNA = &DNAprms;
  par.DNA = NULL;
  F.f = &Etot;
  F.df = &Etot_df;
  F.fdf = &Etot_fdf;
  F.n = np;
  F.params = (void *) &par;
  
  gsl_multimin_fdfminimizer_set (s, &F, x, 0.01, 1e-4);
  
  // Iterate:
  do {
    iter++;
    status = gsl_multimin_fdfminimizer_iterate(s);
    
    if (status)
      break;
    
    status = gsl_multimin_test_gradient (s->gradient, 1e-3);
    
    if (status == GSL_SUCCESS)
      printf ("Minimum found at:\n");
    
    printf ("%5d  %10.5f\n", iter, s->f);
    
  } while (status == GSL_CONTINUE && iter < 300);
  
  gsl_multimin_fdfminimizer_free (s);
  return status;
}

// This is a central function that runs the simplex minimizer.
int Minimizer::Run(void) {
  size_t np = DIM; // number of DoF
  int status;
  double size;
  
  const gsl_multimin_fminimizer_type *T = 
    gsl_multimin_fminimizer_nmsimplex;
  
  gsl_multimin_function F;
  
  // Initial vertex size vector:
  gsl_vector *ss = gsl_vector_alloc(np);
  gsl_vector_set_all(ss, 1.0); // Set all step sizes to 1
  
  // Starting point:
  gsl_vector_memcpy(x,x0); // copy of initial prms: x0 => x
  
  // Initialize function:
  M_params par;
  par.M = this;
  //par.DNA = &DNAprms;
  par.DNA = NULL;
  F.f = &Etot;
  F.n = np;
  F.params = (void *) &par;
  // Initialize method:
  gsl_multimin_fminimizer *s = gsl_multimin_fminimizer_alloc(T,np);
  gsl_multimin_fminimizer_set(s, &F, x, ss);
  
  // Iterate:
  size_t iter=0;
  do {
    iter++;
    status = gsl_multimin_fminimizer_iterate(s);
    
    if (status) 
      break;
    
    size = gsl_multimin_fminimizer_size(s);
    status = gsl_multimin_test_size(size, 1e-2);
    
    if (status == GSL_SUCCESS) {
      printf ("converged to minimum at\n");
    }
    
    printf ("%5d ", iter);
    //for (size_t i = 0; i < np; i++) {
    //  printf ("%10.3e ", gsl_vector_get (s->x, i));
    //}
    printf ("f() = %7.3f size = %.3f\n", s->fval, size);
  } while (status == GSL_CONTINUE && iter < 10000);
  
  gsl_vector_free(ss);
  gsl_multimin_fminimizer_free(s);
  
  return status;
}

double Minimizer::getx(int ind) {
  return gsl_vector_get(x,ind);
}

void Minimizer::setx(int ind, double value) {
  gsl_vector_set(x,ind,value);
}

gsl_vector *Minimizer::getresult(void) {
  return x;
}

void Minimizer::setresult(const gsl_vector *v) {
  assert(x->size == v->size);
  gsl_vector_memcpy(x,v);
}

