/* vector.c */

/*****************************************************************
        Copyright by Rockefeller University,
can not be reproduced or distributed without written permission of
the copyright holder.  Version 1.1 of April 2006.

Written by Alexandre Morozov (contact person: morozov@edsb.rockefeller.edu),
and Eric Siggia.

The program DNABEND implements an algorithm for computing elastic energies
for DNA molecules of arbitrary length and sequence. DNA can be constrained
to follow an arbitrary 3D curve, such as a left-handed superhelix found
in nucleosomal core particles.

The DNABEND method and its applications are described in the following
paper:

******************************************************************/

#include "vector.h"
#include <assert.h>

// Default values for this constructor are 0,0,0.
Vector::Vector(double x_in, double y_in, double z_in){
  x=x_in;
  y=y_in;
  z=z_in;
}

// Copy constructor.
Vector::Vector(const Vector &v){
  x=v.x;
  y=v.y;
  z=v.z;
}

// Returns the cross product of vectors b and v.
Vector Vector::Cross(const Vector &b, const Vector &v){
  Vector temp;
  
  temp.x=b.y*v.z-b.z*v.y;
  temp.y=b.z*v.x-b.x*v.z;
  temp.z=b.x*v.y-b.y*v.x;

  return temp;
}

// Returns the dot product of this Vector and v.
double Vector::operator*(const Vector &v){
  double sum;

  sum=x*v.x;
  sum+=y*v.y;
  sum+=z*v.z;

  return sum;
}

// Normalizes this Vector.
Vector &Vector::Normalize(){
  double sum=0;

  sum+=x*x;
  sum+=y*y;
  sum+=z*z;
  sum=sqrt(sum);
  assert(sum!=0);
  x/=sum;
  y/=sum;
  z/=sum;

  return *this;
}

// Flips the direction of this Vector.
Vector &Vector::Negate(){

  x=-x;
  y=-y;
  z=-z;

  return *this;
}

// Returns true if the distance to Vector v from this Vector is less
// than the square root of dist.
int Vector::WithinSqrDist(const Vector &v, double dist){
  double sum,diff;

  diff=x-v.x;
  sum=diff*diff;
  if(sum>dist) // If the x distance is already too big, we can avoid
    return 0;  // unecessary calculations.
  diff=y-v.y;
  sum+=diff*diff;
  if(sum>dist) // If the x+y distance is too big...
    return 0;
  diff=z-v.z;
  sum+=diff*diff;
  if(sum>dist)
    return 0;

  return 1; // The total square distance is less than dist.
}

// Returns the length of this Vector
double Vector::Length(){
  double sum=0;

  sum+=x*x;
  sum+=y*y;
  sum+=z*z;

  return sqrt(sum);
}

Vector Vector::operator+(const Vector &v){
  Vector temp;

  temp.x=x+v.x;
  temp.y=y+v.y;
  temp.z=z+v.z;

  return temp;
}

Vector Vector::operator-(const Vector &v){
  Vector temp;

  temp.x=x-v.x;
  temp.y=y-v.y;
  temp.z=z-v.z;

  return temp;
}

Vector operator*(double c, const Vector &v){
  Vector temp;

  temp.x=c*v.x;
  temp.y=c*v.y;
  temp.z=c*v.z;

  return temp;
}

Vector Vector::operator*(double c){
  Vector temp;

  temp.x=c*x;
  temp.y=c*y;
  temp.z=c*z;

  return temp;
}

void Vector::operator+=(const Vector &v){

  x+=v.x;
  y+=v.y;
  z+=v.z;
}

void Vector::operator-=(const Vector &v){

  x-=v.x;
  y-=v.y;
  z-=v.z;
}

void Vector::operator*=(double c){
  x*=c;
  y*=c;
  z*=c;
}

void Vector::operator/=(double c){
  
  if (fabs(c)<1e-12) {//c shouldn't be too small..
    std::cerr << "Illegal division in Vector /= operation!\nExiting..\n";
    exit(EXIT_FAILURE);
  }
  x/=c;
  y/=c;
  z/=c;
}

std::ostream &operator<<(std::ostream &out, Vector v){
  out << "(" << v.x << "," << v.y << "," << v.z << ")";
  //out << v.x << " \t" << v.y << " \t" << v.z;
  return out;
}

std::istream &operator>>(std::istream &in, Vector v){
  double a,b,c;

  in >> a >> b >> c;
  v.x = a;
  v.y = b;
  v.z = c;
  return in;
}

void Vector::Rotate(Vector R, double theta){
//  This function rotates the Vector v around the axis R by theta radians.
//  The use of static variables here means we don't have to recompute the
//  rotation matrix each time we call the function, if R and theta stay
//  the same.
//NOTE: Vector R must be normalized!!
 static Vector pos,A1,A2,A3,last_r(0,0,0);
 static double cost,sint,d,Rx2,Ry2,Rz2,last_theta=0;
 Vector temp(*this);

 if((last_theta!=theta)||(last_r.x!=R.x)||
    (last_r.y!=R.y)||(last_r.z!=R.z)){
   last_theta=theta;
   last_r=R;
   cost=cos(theta);
   sint=-sin(theta);
   //  This matrix was originally computed by Ed Thayer, and double-checked
   //  by me.  It is the product of a basis transformation into a basis in which
   //  R is one of the coordinate axes, a simple rotation around R, and
   //  a basis transformation back again.
   Rx2=R.x*R.x;  Ry2=R.y*R.y;  Rz2=R.z*R.z;
   d=Rx2+Ry2;
   if(Rx2+Ry2>0){
     A1.x=Rx2+(Ry2+Rx2*Rz2)*cost/d;
     A1.y=R.x*R.y*(1-cost)+R.z*sint;
     A1.z=R.x*R.z*(1-cost)-R.y*sint;
     A2.x=R.x*R.y*(1-cost)-R.z*sint;
     A2.y=Ry2+(Rx2+Ry2*Rz2)*cost/d;
     A2.z=R.y*R.z*(1-cost)+R.x*sint;
     A3.x=R.x*R.z*(1-cost)+R.y*sint;
     A3.y=R.y*R.z*(1-cost)-R.x*sint;
     A3.z=Rz2+d*cost;
   }else{
     A1.x = cost     ; A1.y = R.z*sint; A1.z = 0;
     A2.x = -R.z*sint; A2.y = cost    ; A2.z = 0;
     A3.x =         0; A3.y =        0; A3.z = Rz2;
   }
 }
 //  This is just an obscure way of writing the product of matrix A
 //  with Vector v.
 x = A1 * temp;
 y = A2 * temp;
 z = A3 * temp;
}

void Vector::ProjectNormal(Vector r){
  Vector a(*this);
  Vector pr;
 
  // norm_r is the unit Vector in the direction of r
  // just in case its not already normalized 
  r.Normalize();
 
  // pr is the projection of a onto r
  pr = r;
  pr *= a*r;
 
  // Next, we subtract the projection onto r from the original Vector,
  // yielding the component normal to r.
  a -= pr;

  *this = a;
}

void Vector::set(double a, double b, double c) {
  x=a;
  y=b;
  z=c;
}

