/* xyzframe.c */

/*****************************************************************
        Copyright by Rockefeller University,
can not be reproduced or distributed without written permission of
the copyright holder.  Version 1.1 of April 2006.

Written by Alexandre Morozov (contact person: morozov@edsb.rockefeller.edu),
and Eric Siggia.

The program DNABEND implements an algorithm for computing elastic energies
for DNA molecules of arbitrary length and sequence. DNA can be constrained
to follow an arbitrary 3D curve, such as a left-handed superhelix found
in nucleosomal core particles.

The DNABEND method and its applications are described in the following
paper:

******************************************************************/

#include "xyzframe.h"

// Read WO prms from a BStable, build frames from a default 1st frame (taken from 1kx5).
Geom::Geom(BStable::BStable &bst) {
  // Set up a local copy of WO prms:
  Lseq = (unsigned int) bst.length(); // prm table length (ie seq length **in dinucs**)
  m = gsl_matrix_calloc(Lseq,Nprm);
  for (unsigned int k=0; k<Lseq; ++k) {
    for (unsigned int j=0; j<Nprm; ++j) {
      gsl_matrix_set(m,k,j,bst.get(k,j) );
    }
  }
  derivs = NULL;
  F_curv = NULL;
  // Init. 1st frame with PDB coords (from a Rosetta run on 1kx5):
  crd_init = InitFirstFrame();
  // Find all the other frames:
  ReconstructGeom(crd_init);
}

// Read WO prms from a BStable, build frames from an input 1st frame.
Geom::Geom(BStable::BStable &bst, coord crd) {
  // Set up a local copy of WO prms:
  Lseq = (unsigned int) bst.length(); // prm table length (ie seq length **in dinucs**)
  m = gsl_matrix_calloc(Lseq,Nprm);
  for (unsigned int k=0; k<Lseq; ++k) {
    for (unsigned int j=0; j<Nprm; ++j) {
      gsl_matrix_set(m,k,j,bst.get(k,j) );
    }
  }
  derivs = NULL;
  F_curv = NULL;
  // Init. 1st frame with user-unput coords:
  crd_init = crd;
  // Find all the other frames:
  ReconstructGeom(crd_init);
}

// Read WO prms from a gsl_matrix, build frames. 
Geom::Geom(gsl_matrix *bst_mat, coord crd) {
  // Set up a local copy of WO prms:
  Lseq = bst_mat->size1; // prm table length (ie seq length **in dinucs**)
  m = gsl_matrix_calloc(Lseq,Nprm);
  gsl_matrix_memcpy(m,bst_mat);
  derivs = NULL;
  F_curv = NULL;
  // Init. 1st frame with PDB coords (from a Rosetta run on 1kx5):
  //crd_init = InitFirstFrame();
  crd_init = crd;
  // Find all the other frames:
  ReconstructGeom(crd_init);
}

// Read WO prms from a gsl_vector, build frames. 
Geom::Geom(const gsl_vector *bst_vec, coord crd) {
  // Check dims of the input vector:
  if (bst_vec->size/Nprm <= 0) {
    std::cerr << "ERROR in Geom::Geom => invalid input vector of WO prms!\n";
    exit(EXIT_FAILURE);
  }
  if (bst_vec->size%Nprm != 0) { // not a multiple of 6!!
    std::cerr << "ERROR in Geom::Geom => invalid dims of the input vector of WO prms!\n";
    exit(EXIT_FAILURE);
  }
  // Set up a local copy of WO prms:
  Lseq = bst_vec->size/Nprm;  // prm table length (ie seq length **in dinucs**)
  m = gsl_matrix_calloc(Lseq,Nprm);
  for (unsigned int k=0; k<Lseq; ++k) {
    for (unsigned int j=0; j<Nprm; ++j) {
      int linind = k*Nprm + j;
      gsl_matrix_set(m,k,j,gsl_vector_get(bst_vec,linind) );
    }
  }
  derivs = NULL;
  F_curv = NULL;
  // Init. 1st frame with PDB coords (from a Rosetta run on 1kx5):
  //crd_init = InitFirstFrame();
  crd_init = crd;
  // Find all the other frames:
  ReconstructGeom(crd_init);
}

// Read frames from 2 files: 1 for bps & 1 for mst.
Geom::Geom(std::string const & fbp, std::string const & fmst) {
  // Read basepair frames:
  ReadFrames(fbp,frames);
  // Read MST frames:
  ReadFrames(fmst,frames_mst);
  // Check results:
  if (frames.size() != frames_mst.size() + 1) {
    std::cerr << "Unequal sets of BP and MST frames!\nExiting ..\n";
    exit(EXIT_FAILURE);
  }
  Lseq = frames_mst.size();
  crd_init = frames.at(0);
  derivs = NULL;
  F_curv = NULL;
  // Compute WO prms & store them in the internal table:
  m = gsl_matrix_calloc(Lseq,Nprm);
  TranslateGeom();
}

// Read a set of frames from a PDB file.
void Geom::ReadFrames(std::string const & filename, std::vector<coord> &frm) {
  using std::cout;
  using std::cerr;
  using std::endl;
  using std::ifstream;
  
  ifstream file(filename.c_str() );
  
  char buffer[512];
  
  if (file.fail()) {
    cerr << "Could not open file: "<<filename<<" for reading!\nExiting..\n";
    exit(EXIT_FAILURE);
  }
  
  // Read in frames:
  int cnt = 0; // ATOM line counter
  double x,y,z,b1,b2;
  int atomnum,resnum;
  char atom[5],atomname[5],resname[4],chainID[2];
  while (!file.eof() ) {
    file.getline(buffer,512,'\n');
    if (file.fail() && !file.eof() ) file.clear();
    if (sscanf(buffer,"%s %d %s %s %s %d %lf %lf %lf %lf %lf",
	       atom,&atomnum,atomname,resname,chainID,&resnum,&x,&y,&z,&b1,&b2) == 11) {
      if (cnt%4==0) { // frames MUST come in order: [origin, n1, n2, n3]
	coord Tcur;
	Tcur.orig.x = x;
	Tcur.orig.y = y;
	Tcur.orig.z = z;
	frm.push_back(Tcur);
      }else if (cnt%4==1) {
	unsigned int fs = frm.size();
	frm.at(fs-1).a.x = x;
	frm.at(fs-1).a.y = y;
	frm.at(fs-1).a.z = z;
      }else if (cnt%4==2) {
	unsigned int fs = frm.size();
	frm.at(fs-1).b.x = x;
	frm.at(fs-1).b.y = y;
	frm.at(fs-1).b.z = z;
      }else if (cnt%4==3) {
	unsigned int fs = frm.size();
	frm.at(fs-1).c.x = x;
	frm.at(fs-1).c.y = y;
	frm.at(fs-1).c.z = z;
      }
      ++cnt;
    }
  }
  file.close();
  
  // Subtract off the origin to make unit vectors:
  for (unsigned int i=0; i<frm.size(); ++i) {
    frm.at(i).a -= frm.at(i).orig;
    frm.at(i).b -= frm.at(i).orig;
    frm.at(i).c -= frm.at(i).orig;
  }
  
  // Check frames:
  if (cnt%4 != 0) { // expect the number of lines read to be a multiple of 4: [origin, n1, n2, n3]
    cerr << "Unexpected number of lines in the frames PDB file!\nExiting..\n";
    exit(EXIT_FAILURE);
  }
  fprintf(stderr,"Read %d frames from file: %s ..\n",cnt/4,filename.c_str() );
}

Geom::~Geom() {
  if (m != NULL) gsl_matrix_free(m);
  if (derivs != NULL) {
    for (unsigned int j=0; j<Lseq; ++j) {
      delete [] derivs[j];
    }
    delete [] derivs;
  }
  if (F_curv != NULL) gsl_matrix_free(F_curv);
}

// Return a single value from the local copy of BStable:
double Geom::getm(unsigned int i1, unsigned int i2) {
  assert(i1<Lseq && i2<Nprm);
  if (m==NULL) {
    std::cerr << "Table of WO prms not set!\nExiting ..\n";
    exit(EXIT_FAILURE);
  }
  return gsl_matrix_get(m,i1,i2);
}

// Compare 2 sets of bp+mst frames (diagnostic function, ie for comparison of ideal superhelix &
// native frames):
void Geom::CompareFrames(Geom::Geom &gm, const std::string mode) {
  // Check dims:
  if (frames.size() != gm.frames.size() ||
      frames_mst.size() != gm.frames_mst.size() ) {
    std::cerr << "ERROR: cannot compare frames with different dimensions!\nExiting ..\n";
    exit(EXIT_FAILURE);
  }
  // Compute & print out the dot product between unit vectors, and distance between origins:
  if (mode=="bp") {
    for (unsigned int i=0; i<frames.size(); ++i) {
      double ndot = frames.at(i).a * gm.frames.at(i).a;
      double bdot = frames.at(i).b * gm.frames.at(i).b;
      double tdot = frames.at(i).c * gm.frames.at(i).c;
      if (ndot > 1.0) ndot = 1.0; // otherwise acos fails
      if (bdot > 1.0) bdot = 1.0; // otherwise acos fails
      if (tdot > 1.0) tdot = 1.0; // otherwise acos fails
      Vector or12 = frames.at(i).orig - gm.frames.at(i).orig;
      fprintf(stdout,"Frame %3d: (n, b, t) = (%5.2f, %5.2f, %5.2f) orig_dist = %6.2f\n",
	      i+1,RadtoDeg(acos(ndot)),RadtoDeg(acos(bdot)),RadtoDeg(acos(tdot)),or12.Length() );
    }
  }else if (mode=="mst") {
    for (unsigned int i=0; i<frames_mst.size(); ++i) {
      double ndot = frames_mst.at(i).a * gm.frames_mst.at(i).a;
      double bdot = frames_mst.at(i).b * gm.frames_mst.at(i).b;
      double tdot = frames_mst.at(i).c * gm.frames_mst.at(i).c;
      if (ndot > 1.0) ndot = 1.0; // otherwise acos fails
      if (bdot > 1.0) bdot = 1.0; // otherwise acos fails
      if (tdot > 1.0) tdot = 1.0; // otherwise acos fails
      Vector or12 = frames_mst.at(i).orig - gm.frames_mst.at(i).orig;
      fprintf(stdout,"Frame %3d: (n, b, t) = (%5.2f, %5.2f, %5.2f) orig_dist = %6.2f\n",
	      i+1,RadtoDeg(acos(ndot)),RadtoDeg(acos(bdot)),RadtoDeg(acos(tdot)),or12.Length() );
    }
  }else{
    std::cerr << "Unknown mode: "<<mode<<" in Geom::CompareFrames!\nExiting ..\n";
    exit(EXIT_FAILURE);
  }
}

// Init. 1st frame with PDB coords (from a Rosetta run on 1kx5):
coord Geom::InitFirstFrame(void) {
  Vector xinit(0.805358,-0.556527,-0.204146);
  Vector yinit(0.157778,-0.130716,0.978785);
  Vector zinit(-0.571405,-0.820482,-0.0174659);
  Vector orinit(22.2795,138.506,19.3309);
  coord crd_init;
  crd_init.a = xinit;
  crd_init.b = yinit;
  crd_init.c = zinit;
  crd_init.orig = orinit;
  return crd_init;
}

void Geom::ReconstructGeom(coord & crd) {
  ReconstructGeom(crd,Lseq);
}

// This function reconstructs basestep triad geometries from a set of values of WO prms.
// bst contains Lseq*6 table of WO prms, while crd is the coord system of the 1st basepair
// (in some global PDB frame).
void Geom::ReconstructGeom(coord & crd, unsigned int Find) {
  
  frames.clear();
  frames_mst.clear();
  
  frames.push_back(crd); // save 1st frame
  
  //debug
  //double cumul_twist = 0.0;
  //int l = bst.length(); // prm table length
  for (unsigned int k=0; k<Find; ++k) { // 1st frame is already known
    // Compute Gamma & phi:
    double twi = DegtoRad(gsl_matrix_get(m,k,0) ); // twist roll tilt (in rad!!)
    double r = DegtoRad(gsl_matrix_get(m,k,1) );
    double t = DegtoRad(gsl_matrix_get(m,k,2) );
    //debug
    //twi = DegtoRad(34.69599);
    //cumul_twist += twi;
    //r = DegtoRad(4.53) * cos(cumul_twist);
    //t = DegtoRad(4.53) * sin(cumul_twist);
    double Gamma = sqrt(r*r + t*t);
    double phi = atan(t/r);
    if (r < 0.0) phi += PI; // phi > PI/2 if r < 0
    //debug
    //std::cout << "G = "<<Gamma<<"; phi = "<<phi<<"; twi = "<<twi<<std::endl;
    // previous frame:
    std::vector<coord>::reverse_iterator p = frames.rbegin();
    //debug
    //std::cout << "Rotating frame with origin: " << p->orig << std::endl;
    // Construct [T_{i}^{g}]:
    Matrix3 Tig1(p->a,p->b,p->c); // vectors become matrix columns
    Matrix3 Tig2(p->a,p->b,p->c); // need 2 copies because we'll rotate them differently
    // Init. unit vector frame (in which rotations happen):
    coord funit;
    funit.a.set(1.0,0.0,0.0);
    funit.b.set(0.0,1.0,0.0);
    funit.c.set(0.0,0.0,1.0);
    funit.orig.set(0.0,0.0,0.0);
    // Get MST and T_{i+1}^{i} in the local frame:
    coord mst = RotateMat(funit, -0.5*twi + phi, 0.5*Gamma, -phi);
    coord Tnext = RotateMat(funit, -0.5*twi + phi, Gamma, -0.5*twi - phi);
    // Construct T_{MST} and T_{i+1}^{i} in the global frame:
    Matrix3 Tmst(mst.a,mst.b,mst.c); // vectors become matrix columns
    Matrix3 Tij(Tnext.a,Tnext.b,Tnext.c); // vectors become matrix columns
    //Tmst.transpose(); // actually you dont need it at all
    //Tij.transpose(); // actually you dont need it at all
    Tig1.RegMultiply(Tij); // T_{i+1}^{i} in the global frame
    Tig2.RegMultiply(Tmst); // T_{MST} in the global frame
    Tig1.loadvecs(Tnext.a,Tnext.b,Tnext.c); // load matrix into 3 vectors by columns
    Tig2.loadvecs(mst.a,mst.b,mst.c); // load matrix into 3 vectors by columns
    //debug
    //std::cout << "====================\n";
    //Tig.Print();
    //std::cout << "====================\n";
    // Compute roll-tilt axis:
    //Vector rt = (r/Gamma)*mst.b + (t/Gamma)*mst.a;
    //double rtz = Vector::Cross(rt,mst.b) * mst.c;
    //if (phi*rtz < 0) {
    //  std::cout << "product = " << rtz << "\n";
    //  std::cout << "phi = " << phi << "\n";
    //}
    //////////////
    double shift = gsl_matrix_get(m,k,3);
    double slide = gsl_matrix_get(m,k,4);
    double rise = gsl_matrix_get(m,k,5);
    //debug
    //shift = 0.0;
    //slide = 0.0;
    Tnext.orig = p->orig + shift*mst.a + slide*mst.b + rise*mst.c;
    frames.push_back(Tnext);
    // MST frame origin should be halfway in between frames T_{i} and T_{i+1}, but we make it identical
    // to that for frame T_{i} because it's not used anywhere anyway:
    mst.orig = p->orig;
    //mst.orig = 0.5 * (Tnext.orig + p->orig); // this is correct but creates technical problems !!!
    frames_mst.push_back(mst); // note that MST origin is the same as that for frame T_{i}!
  }
  // debug
  //Write("bp");
  //Write("mst");
}

// This function builds & stores derivatives of 3x3 MST frame matrices wrt [Twist,Roll,Tilt].
// There are Lseq x 3*Lseq matrix derivatives arranged in a 2D array.
void Geom::BuildDerivs(void) {
  // Allocate memory:
  derivs = new Matrix3*[Lseq];
  for (unsigned int j=0; j<Lseq; ++j) {
    derivs[j] = new Matrix3[3*(j+1)]; // allocate triangular 2D table for all derivs
  }
  /*
  // Numeric:
  const double angle_diff = 0.5;
  for (unsigned int k1=0; k1<Lseq; ++k1) { // j: over all sets of WO prms
    //debug
    printf("k1 = %d ..\n",k1);
    for (unsigned int k2=0; k2<=k1; ++k2) { // j' <= j
      for (unsigned int ind=0; ind<3; ++ind) { // over alpha [Twist,Roll,Tilt] = [0,1,2]
	int linind = k2*3 + ind;
	derivs[k1][linind] = NumDeriv(k1, k2, ind, angle_diff);
      }
    }
  }
  */
  // Analytical:
  // Reconstruct T_k (k = 0 .. Lseq) & T_k^{MST} (k = 1 .. Lseq) from frames and frames_mst:
  std::vector<Matrix3> T;
  std::vector<Matrix3> T_MST;
  Matrix3 R0(frames.at(0).a,frames.at(0).b,frames.at(0).c);
  T.push_back(R0);
  for (unsigned int i=1; i<=Lseq; ++i) {
    // Reconstruct T_i (i = 1 .. Lseq):
    Matrix3 Rcur(frames.at(i).a,frames.at(i).b,frames.at(i).c);
    Matrix3 Rprev(frames.at(i-1).a,frames.at(i-1).b,frames.at(i-1).c);
    //debug
    //Matrix3 J(Rprev);
    //=========
    Rprev.transpose();
    //debug
    //J.RegMultiply(Rprev);
    //J.Print();
    //=========
    Rprev.RegMultiply(Rcur); // R_{i-1}^{-1} R_{i} = T_{i}
    T.push_back(Rprev);
    // Reconstruct T_i^{MST} (i = 1 .. Lseq):
    Matrix3 Rprev_MST(frames.at(i-1).a,frames.at(i-1).b,frames.at(i-1).c);
    Matrix3 Rcur_MST(frames_mst.at(i-1).a,frames_mst.at(i-1).b,frames_mst.at(i-1).c);
    Rprev_MST.transpose();
    Rprev_MST.RegMultiply(Rcur_MST); // R_{i-1}^{-1} R_{i}^{MST}
    T_MST.push_back(Rprev_MST);
  }
  // Now compute all derivaties by making appropriate matrix products:
  for (unsigned int k1=0; k1<Lseq; ++k1) { // j: over all sets of WO prms
    //debug
    printf("k1 = %d ..\n",k1);
    for (unsigned int k2=0; k2<=k1; ++k2) { // j' <= j
      // Compute twist etc.:
      double twi = DegtoRad(gsl_matrix_get(m,k2,0) ); // twist roll tilt (in rad!!)
      double r = DegtoRad(gsl_matrix_get(m,k2,1) );
      double t = DegtoRad(gsl_matrix_get(m,k2,2) );
      for (unsigned int ind=0; ind<3; ++ind) { // over alpha = (0,1,2) [Twist,Roll,Tilt]
	int linind = k2*3 + ind;
	Matrix3 R(T.at(0) ); // T_0
	if (k2==k1) { // see corresponding formulas
	  for (unsigned int k=1; k<=k1; ++k) {
	    R.RegMultiply(T.at(k) );
	  }
	  Matrix3 D = Tmst_deriv(twi,r,t,ind);
	  //Matrix3 D = Tmst(twi,r,t); // check if R^{MST} is restored
	  R.RegMultiply(D);
	  //debug
	  //if (k1==50) {
	  //R.Print();
	  //Matrix3 J(frames_mst.at(k1).a,frames_mst.at(k1).b,frames_mst.at(k1).c);
	  //J.Print();
	  //exit(0);
	  //}
	}else{
	  for (unsigned int k=1; k<=k2; ++k) {
	    R.RegMultiply(T.at(k) );
	  }
	  Matrix3 TD = T_deriv(twi,r,t,ind);
	  //Matrix3 TD = Geom::T(twi,r,t); // check if R^{MST} is restored
	  R.RegMultiply(TD);
	  for (unsigned int k=k2+1; k<k1; ++k) {
	    R.RegMultiply(T.at(k+1) );
	  }
	  R.RegMultiply(T_MST.at(k1) );
	  //debug
	  //if (k1==100 && k2==0) {
	  //  R.Print();
	  //  Matrix3 J(frames_mst.at(k1).a,frames_mst.at(k1).b,frames_mst.at(k1).c);
	  //  J.Print();
	  //  exit(0);
	  //}
	}
	derivs[k1][linind] = R;
      }
    }
  }
}

// Return a specific derivs Matrix3.  
Matrix3 Geom::GetDeriv(unsigned int bs_ind, unsigned int WO_ind, unsigned int angle_ind) {
  assert(derivs != NULL); // derivs built?
  // Check ranges:
  assert(bs_ind >= 0 && bs_ind < Lseq);
  assert(angle_ind >= 0 && angle_ind < 3);
  assert(WO_ind <= bs_ind);
  int linind = WO_ind * 3 + angle_ind;
  return derivs[bs_ind][linind]; // bs_ind = j; linind = WO_ind * 3 + angle_ind = j' * 3 + alpha
}

// This function builds a numerical derivative of R^i wrt to a given angle.
// This is for comparison with the analytical expressions used in Geom::BuildDerivs.
// NOTE: give diff in degrees!!
Matrix3 Geom::NumDeriv(unsigned int bs_ind, unsigned int WO_ind, unsigned int angle_ind, double diff) {
  assert(bs_ind >= 0 && bs_ind < Lseq);
  assert(angle_ind >= 0 && angle_ind < 3);
  assert(WO_ind <= bs_ind);
  // Save all frames (to restor them later):
  std::vector<coord> frames_cpy; // basepair frames: origin & 3 unit vecs
  std::vector<coord> frames_mst_cpy; // MST frames: origin & 3 unit vecs
  for (unsigned int i=0; i<frames.size(); ++i) {
    frames_cpy.push_back(frames.at(i) );
  }
  for (unsigned int i=0; i<frames_mst.size(); ++i) {
    frames_mst_cpy.push_back(frames_mst.at(i) );
  }
  // Save the old MST frame [ deriv of which you wish to compute ]:
  coord init = frames_mst.at(bs_ind);
  // Add diff to the BStable:
  double angle_old = gsl_matrix_get(m,WO_ind,angle_ind);
  gsl_matrix_set(m,WO_ind,angle_ind,angle_old + diff);
  ReconstructGeom(crd_init,Lseq); // ??
  coord fin = frames_mst.at(bs_ind);
  Matrix3 T1(init.a,init.b,init.c);
  Matrix3 T2(fin.a,fin.b,fin.c);
  T2.subtract(T1);
  T2.multiply(DegtoRad(1.0)/diff);
  // Restore frames & BStable:
  gsl_matrix_set(m,WO_ind,angle_ind,angle_old);
  frames.clear();
  frames_mst.clear();
  for (unsigned int i=0; i<frames_cpy.size(); ++i) {
    frames.push_back(frames_cpy.at(i) );
  }
  for (unsigned int i=0; i<frames_mst_cpy.size(); ++i) {
    frames_mst.push_back(frames_mst_cpy.at(i) );
  }
  return T2;
}

// This function computes a table of connectivity coefficients b^{ab}_{jj'} given a set of
// pre-computed frame derivatives + a set of ideal displacements.
// NOTE: new memory allocated inside!!
gsl_matrix *Geom::ConnectivityCoeff(void) {
  assert(derivs != NULL); // derivs built?
  gsl_matrix *b = gsl_matrix_calloc(3*Lseq,3*Lseq);
  for (unsigned int i1=0; i1<Lseq; ++i1) { // j
    Vector d_id(gsl_matrix_get(m,i1,3),gsl_matrix_get(m,i1,4),gsl_matrix_get(m,i1,5) );
    //debug
    //printf("d_id = [%.3f %.3f %.3f]; |d_id| = %.3f\n",d_id.x,d_id.y,d_id.z,d_id.Length() );
    for (unsigned int i2=0; i2<=i1; ++i2) { // j'
      for (unsigned int j2=0; j2<3; ++j2) { // alpha
	int linind2 = i2*3 + j2;
	Matrix3 mm = derivs[i1][linind2];
	Vector v = mm.VecMultiply(d_id);
	//debug
	//printf("v = [%.3f %.3f %.3f]; |v| = %.3f\n",v.x,v.y,v.z,v.Length() );
	gsl_matrix_set(b,i1*3+0,linind2,v.x); // 1st index: [j,beta]
	gsl_matrix_set(b,i1*3+1,linind2,v.y); // 2nd index: [j',alpha]
	gsl_matrix_set(b,i1*3+2,linind2,v.z);
	//debug
	//printf("linind1 = %d, linind2 = %d: b = %.3f\n",i1*3+0,linind2,gsl_matrix_get(b,i1*3+0,linind2) );
	//printf("linind1 = %d, linind2 = %d: b = %.3f\n",i1*3+1,linind2,gsl_matrix_get(b,i1*3+1,linind2) );
	//printf("linind1 = %d, linind2 = %d: b = %.3f\n",i1*3+2,linind2,gsl_matrix_get(b,i1*3+2,linind2) );
      }
    }
  }
  //debug
  //for (unsigned int i1=0; i1<3*Lseq; ++i1) {
  //for (unsigned int i2=0; i2<3*Lseq; ++i2) {
  //  printf("linind1 = %d, linind2 = %d: b = %.3f\n",i1,i2,gsl_matrix_get(b,i1,i2) );
  //}
  //}
  //exit(0);
  return b;
}

// This aux.function writes connectivity matrix in binary format.
void Geom::WriteConnectivityCoeff(const std::string filename) {
  // Get connectivities:
  gsl_matrix *b = ConnectivityCoeff(); // 1st index: [j,beta]; 2nd index: [j',alpha]
  
  FILE *fp = fopen(filename.c_str(),"w");
  if (fp == NULL) {
    printf("Error writing to file: %s\n",filename.c_str() );
    exit(1);
  }
  
  if (gsl_matrix_fwrite(fp, b) != 0) {
    fprintf(stderr,"ERROR writing connectivity matrix in binary format!\n");
    exit(1);
  }
  
  fclose(fp);
  gsl_matrix_free(b);
}

// This function computes & stores curvature coefficients imposed by the superhelix.
// NOTE: the curvature coeffs are computed for the quadratic potential corresp. to
// the displacement constraint.
void Geom::ComputeCurv(void) {
  // Allocate memory:
  F_curv = gsl_matrix_calloc(Lseq*Nprm,Lseq*Nprm);
  // Get connectivities:
  gsl_matrix *b = ConnectivityCoeff(); // 1st index: [j,beta]; 2nd index: [j',alpha]
  // Set superhelix-induced force constants:
  for (unsigned int i1=0; i1<Lseq; ++i1) {
    printf("Computing curvature for i1=%d\n",i1);
    for (unsigned int i2=0; i2<Lseq; ++i2) {
      for (unsigned int j1=0; j1<3; ++j1) {
	int linind1 = i1*Nprm + j1;
	for (unsigned int j2=0; j2<3; ++j2) {
	  int linind2 = i2*Nprm + j2;
	  // G^{j1 j2}_{i1 i2}:
	  if (j1==j2) {
	    int maxindG;
	    if (i1 > i2) {maxindG = i1;} else {maxindG = i2;}
	    int B_i1_i2 = Lseq + 1 - maxindG;
	    gsl_matrix_set(F_curv,linind1+3,linind2+3,B_i1_i2);
	  }
	  // H^{j1 j2}_{i1 i2}:
	  double Htmp = 0.0;
	  for (unsigned int j=i1; j<Lseq; ++j) {
	    for (unsigned int l=i2; l<Lseq; ++l) {
	      int maxindH;
	      if (j > l) {maxindH = j;} else {maxindH = l;}
	      int Bjl = Lseq + 1 - maxindH;
	      for (unsigned int beta=0; beta<3; ++beta) {
		double b_j_i1_j1_beta = gsl_matrix_get(b,j*3+beta,i1*3+j1);
		double b_l_i2_j2_beta = gsl_matrix_get(b,l*3+beta,i2*3+j2);
		Htmp += (Bjl * b_j_i1_j1_beta * b_l_i2_j2_beta);
	      }
	    }
	  }
	  gsl_matrix_set(F_curv,linind1,linind2,Htmp);
	  //debug
	  //printf("i1=%d, i2=%d, j1=%d, j2=%d: G = %e, H = %e\n",i1,i2,j1,j2,
	  // gsl_matrix_get(F_curv,linind1+3,linind2+3), gsl_matrix_get(F_curv,linind1,linind2) );
	  // F^{j1 j2}_{i1 i2}:
	  double Ftmp = 0.0;
	  for (unsigned int l=i2; l<Lseq; ++l) {
	    int maxindF;
	    if (i1 > l) {maxindF = i1;} else {maxindF = l;}
	    int B_i1_l = Lseq + 1 - maxindF;
	    double b_l_i2_j2_j1 = gsl_matrix_get(b,l*3+j1,i2*3+j2);
	    Ftmp += (B_i1_l * b_l_i2_j2_j1);
	  }
	  gsl_matrix_set(F_curv,linind1+3,linind2,Ftmp);
	}
      }
    }
  }
  gsl_matrix_free(b);
}

// The following 3 functions are the user's interface to curvatures, allowing him
// to write pre-computed curvatures to a file, get the pointer to curvatures,
// or read them from a file.

void Geom::WriteCurv(const std::string filename) {
  if (F_curv==NULL) {
    std::cerr << "Error in WriteCurv: curvature coeffs are not pre-computed!\nExiting ..\n";
    exit(EXIT_FAILURE);
  }
  
  FILE *fp = fopen(filename.c_str(),"w");
  if (fp == NULL) {
    printf("Error writing to file: %s\n",filename.c_str() );
    exit(1);
  }
  /*
  // Text
  for (unsigned int i1=0; i1<Lseq*Nprm; ++i1) {
    for (unsigned int i2=0; i2<Lseq*Nprm; ++i2) {
      fprintf(fp,"%4d %4d %11.6f\n",
	      i1,i2,gsl_matrix_get(F_curv,i1,i2) );
    }
  }
  */
  // Binary
  if (gsl_matrix_fwrite(fp, F_curv) != 0) {
    fprintf(stderr,"ERROR writing curvature matrix in binary format!\n");
    exit(1);
  }
  
  fclose(fp);
}

void Geom::ReadCurv(const std::string filename) {
  //using std::ifstream;
  //int ind1,ind2;
  //double val;
  
  if (F_curv==NULL) {
    // Allocate memory if necessary:
    F_curv = gsl_matrix_calloc(Lseq*Nprm,Lseq*Nprm);
  }
  
  // Binary
  FILE *fp = fopen(filename.c_str(),"r");
  if (gsl_matrix_fread(fp,F_curv) != 0) {
    fprintf(stderr,"ERROR reading curvature matrix in binary format!\n");
    exit(1);
  }
  fclose(fp);
  
  /*
  // Text
  ifstream file(filename.c_str() );
  
  char buffer[1024];
  
  if (file.fail() ) {
    printf("Error reading file: %s\n",filename.c_str() );
    exit(1);
  }
  
  int cnt = 0;
  while (!file.eof() ) {
    file.getline(buffer,1024,'\n');
    if (file.fail() && !file.eof() ) file.clear();
    if (sscanf(buffer,"%d %d %lf",&ind1,&ind2,&val) == 3) {
      gsl_matrix_set(F_curv,ind1,ind2,val);
      cnt++;
    }
  }
  file.close();
  
  // Check the number of items read:
  int num_exp = (Lseq*Nprm)*(Lseq*Nprm);
  if (cnt != num_exp) {
    printf("Error reading file %s: %d tokens read, %d expected ..\n",filename.c_str(),cnt,num_exp);
    exit(1);
  }
  */
}

gsl_matrix *Geom::GetCurv(void) {
  return F_curv;
}

// Given a set of bp frames, this function will translate them back into
// WO prms, and update the "in-object" copy of WO prms.
// Used in one of the constructors where frames are read from files rather than
// computed from WO prms.
void Geom::TranslateGeom(void) {
  
  for (unsigned int i=0; i<Lseq; ++i) {
    coord M1 = frames.at(i);
    coord M2 = frames.at(i+1);
    
    // z_{i} and z_{j} are already translated to the origin => compute the angle between them:
    double Gamma = fabs(acos(M1.c * M2.c));
    // Get the Roll-Tilt axis:
    Vector rt = Vector::Cross(M1.c,M2.c);
    rt.Normalize();
    // Rotate around rt so that xy planes (and z-axes) of both triads become parallel:
    Vector xi = M1.a, yi = M1.b, zi = M1.c;
    Vector xj = M2.a, yj = M2.b, zj = M2.c;
    xi.Rotate(rt,Gamma/2.0);
    yi.Rotate(rt,Gamma/2.0);
    zi.Rotate(rt,Gamma/2.0);
    xj.Rotate(rt,-Gamma/2.0);
    yj.Rotate(rt,-Gamma/2.0);
    zj.Rotate(rt,-Gamma/2.0);
    
    // Find the mid-step triad (MST):
    coord MST;
    // a) MST origin:
    //MST.orig = 0.5 * (M1.orig + M2.orig); // this is correct but not used below anyway
    MST.orig = M1.orig;
    // b) MST unit vectors:
    MST.a = (xi + xj);
    MST.a.Normalize();
    MST.b = (yi + yj);
    MST.b.Normalize();
    MST.c = zi;
    //MST.c = (zi + zj);
    //MST.c.Normalize(); // zi and zj must coincide!
    
    // Finally, basestep WO parameters:
    // Angles:
    double twist = RadtoDeg(fabs(acos(yi * yj)) );
    Vector tmp = Vector::Cross(yi,yj);
    if (tmp*MST.c < 0.0) twist *= (-1);
    double phi = fabs(acos(rt * MST.b));
    tmp = Vector::Cross(rt,MST.b);
    if (tmp*MST.c < 0.0) phi *= (-1);
    double roll = RadtoDeg(Gamma*cos(phi) );
    double tilt = RadtoDeg(Gamma*sin(phi) );
    // Displacements:
    tmp = (M2.orig - M1.orig);
    double shift = tmp*MST.a;
    double slide = tmp*MST.b;
    double rise = tmp*MST.c;
    //debug
    //std::cout << "G = "<<Gamma<<"; phi = "<<phi<<"; twi = "<<twist<<std::endl;
    
    // Set "internal" matrix of WO prms:
    gsl_matrix_set(m,i,0,twist);
    gsl_matrix_set(m,i,1,roll);
    gsl_matrix_set(m,i,2,tilt);
    gsl_matrix_set(m,i,3,shift);
    gsl_matrix_set(m,i,4,slide);
    gsl_matrix_set(m,i,5,rise);
  }
}

// This aux. function returns internally stored WO prms in gsl_vector format.
// NOTE: new memory allocated inside!!
gsl_vector *Geom::GetGeom(void) {
  gsl_vector *WO = gsl_vector_calloc(Lseq*Nprm);
  for (unsigned int i=0; i<Lseq; ++i) {
    for (unsigned int j=0; j<Nprm; ++j) {
      unsigned int linind = i*Nprm + j;
      gsl_vector_set(WO,linind,gsl_matrix_get(m,i,j) );
    }
  }
  return WO;
}

// This function writes out frame coords:
void Geom::Write(const std::string mode) {
  int cnt = 0;
  std::vector<coord>::const_iterator p2;
  if (mode=="bp") {
    for (p2 = frames.begin(); p2 != frames.end(); p2++) {
      std::cout << "Frame " << cnt << ":\n";
      std::cout << "x <- c" << p2->a << std::endl;
      std::cout << "y <- c" << p2->b << std::endl;
      std::cout << "z <- c" << p2->c << std::endl;
      std::cout << "orig <- c" << p2->orig << std::endl;
      std::cout << "=======================\n";
      ++cnt;
    }
  }else if (mode=="mst") {
    for (p2 = frames_mst.begin(); p2 != frames_mst.end(); p2++) {
      std::cout << "MST frame " << cnt << ":\n";
      std::cout << "x <- c" << p2->a << std::endl;
      std::cout << "y <- c" << p2->b << std::endl;
      std::cout << "z <- c" << p2->c << std::endl;
      std::cout << "orig <- c" << p2->orig << std::endl;
      std::cout << "=======================\n";
      ++cnt;
    }
  }else{
    std::cerr << "Unknown mode: "<<mode<<" in Geom::Write!\nExiting ..\n";
    exit(EXIT_FAILURE);
  }
}

// This function writes out frame origins in PDB format
// (so they can be viewed by Rasmol).
void Geom::PDB(const std::string filename, const char mode) {
  FILE *fp = fopen(filename.c_str(),"w");
  
  if (fp == NULL) {
    printf("Error reading file %s\n",filename.c_str() );
    exit(1);
  }
  
  if (mode == 's') { // standard PDB ..
    for (unsigned int i=0; i<frames.size(); ++i) {
      fprintf(fp,"ATOM   %4d %4s %3s %1s %3d    %8.3f%8.3f%8.3f  1.00  0.00\n",
	      i+1,"O  ","ALA","I",i+1,frames.at(i).orig.x,frames.at(i).orig.y,frames.at(i).orig.z);
    }
  }else if (mode == 'l') { // increased precision PDB ..
    for (unsigned int i=0; i<frames.size(); ++i) {
      fprintf(fp,"ATOM   %4d %4s %3s %1s %3d    %10.5f%10.5f%10.5f  1.00  0.00\n",
	      i+1,"O  ","ALA","I",i+1,frames.at(i).orig.x,frames.at(i).orig.y,frames.at(i).orig.z);
    }
  }else{
    fprintf(stderr,"Unknown PDB mode: %c!\n",mode);
    exit(1);
  }
  fprintf(fp,"TER\n");
  fclose(fp);
}

// This function writes out frame origins & unit vectors in PDB format
// (so they can be viewed by Rasmol).
void Geom::PDBunitvecs(const std::string filename, const char mode) {
  FILE *fp = fopen(filename.c_str(),"w");
  const double scale = 1.0; // scaling factor for each unit vector (gives better Rasmol pix)
  
  if (fp == NULL) {
    printf("Error reading file %s\n",filename.c_str() );
    exit(1);
  }
  
  int atom_cnt = 1;
  for (unsigned int i=0; i<frames.size(); ++i) {
    // Unit vectors in the global frame:
    Vector v1 = frames.at(i).orig + scale * frames.at(i).a;
    Vector v2 = frames.at(i).orig + scale * frames.at(i).b;
    Vector v3 = frames.at(i).orig + scale * frames.at(i).c;
    if (mode == 's') { // standard PDB ..
      fprintf(fp,"ATOM   %4d %4s %3s %1s %3d    %8.3f%8.3f%8.3f  1.00  0.00\n",
	      atom_cnt++,"O  ","ALA","I",i+1,frames.at(i).orig.x,frames.at(i).orig.y,frames.at(i).orig.z);
      fprintf(fp,"ATOM   %4d %4s %3s %1s %3d    %8.3f%8.3f%8.3f  1.00  0.00\n",
	      atom_cnt++,"N1 ","ALA","I",i+1,v1.x,v1.y,v1.z);
      fprintf(fp,"ATOM   %4d %4s %3s %1s %3d    %8.3f%8.3f%8.3f  1.00  0.00\n",
	      atom_cnt++,"N2 ","ALA","I",i+1,v2.x,v2.y,v2.z);
      fprintf(fp,"ATOM   %4d %4s %3s %1s %3d    %8.3f%8.3f%8.3f  1.00  0.00\n",
	      atom_cnt++,"N3 ","ALA","I",i+1,v3.x,v3.y,v3.z);
    }else if (mode == 'l') { // increased precision PDB ..
      fprintf(fp,"ATOM   %4d %4s %3s %1s %3d    %10.5f%10.5f%10.5f  1.00  0.00\n",
	      atom_cnt++,"O  ","ALA","I",i+1,frames.at(i).orig.x,frames.at(i).orig.y,frames.at(i).orig.z);
      fprintf(fp,"ATOM   %4d %4s %3s %1s %3d    %10.5f%10.5f%10.5f  1.00  0.00\n",
	      atom_cnt++,"N1 ","ALA","I",i+1,v1.x,v1.y,v1.z);
      fprintf(fp,"ATOM   %4d %4s %3s %1s %3d    %10.5f%10.5f%10.5f  1.00  0.00\n",
	      atom_cnt++,"N2 ","ALA","I",i+1,v2.x,v2.y,v2.z);
      fprintf(fp,"ATOM   %4d %4s %3s %1s %3d    %10.5f%10.5f%10.5f  1.00  0.00\n",
	      atom_cnt++,"N3 ","ALA","I",i+1,v3.x,v3.y,v3.z);
    }else{
      fprintf(stderr,"Unknown PDB mode: %c!\n",mode);
      exit(1);
    }
  }
  fprintf(fp,"TER\n");
  fclose(fp);
}

// vector length: how many BP frames are stored
unsigned int Geom::length(void) {
  //debug
  //printf("bp=%d; mst=%d\n",frames.size(),frames_mst.size() );
  if (frames.size() != frames_mst.size() + 1) { // + 1 bc 1st BP is stored as well
    std::cerr << "Error in Geom::length => unequal number of BP frames and MST frames!\nExiting ..\n";
    exit(EXIT_FAILURE);
  }
  return frames.size();
}

// Compute RMSD with another set of frames (ie from pdb_seq):
double Geom::RMSD(Geom::Geom &gm) {
  if (length() != gm.length() ) {
    std::cerr << "Error in Geom::RMSD => unequal number of frames to compare!\nExiting ..\n";
    exit(EXIT_FAILURE);
  }
  double rmsd2 = 0.0;
  for (unsigned int i=0; i<length(); ++i) {
    Vector orig1 = frames.at(i).orig;
    Vector orig2 = gm.frames.at(i).orig;
    Vector diff = orig1 - orig2;
    double l = diff.Length();
    //debug
    //std::cout << "i="<<i<<"; rmsd = "<<l<<std::endl;
    rmsd2 += (l * l);
  }
  return sqrt(rmsd2/length() );
}

// Recompute frame origins using another set of frames
// (ie an idealized set of frames from a superhelix):
void Geom::Transform(Geom::Geom &gm, unsigned int begin, unsigned int end) {
  // Check dims:
  if (frames_mst.size() != gm.frames_mst.size() ||
      frames.size() != gm.frames.size() ) {
    std::cerr << "Unequal sets of BP and MST frames in Geom::Transform!\nExiting ..\n";
    exit(EXIT_FAILURE);
  }
  // Check if WO prms are available:
  if (m==NULL || Lseq != frames_mst.size() ) {
    std::cerr << "WO prms are not set in Geom::Transform!\nExiting ..\n";
    exit(EXIT_FAILURE);
  }
  // Check [begin,end]:
  if (begin<1 || end>=frames.size() ) {
    std::cerr << "[begin,end] interval out of bounds in Geom::Transform!\nExiting ..\n";
    exit(EXIT_FAILURE);
  }
  // Carry out the transformations of WO prms in this object using frames from gm:
  //Vector global_diff(0,0,0);
  for (unsigned int i=begin; i<=end; ++i) {
    double shift = gsl_matrix_get(m,i-1,3);
    double slide = gsl_matrix_get(m,i-1,4);
    double rise =  gsl_matrix_get(m,i-1,5);
    // Displacements in the local frame:
    Vector a(shift,slide,rise);
    // Displacements in the global frame [ MST unit vectors become matrix columns ]:
    Matrix3 Tmst(gm.frames_mst.at(i-1).a, gm.frames_mst.at(i-1).b, gm.frames_mst.at(i-1).c);
    Vector aglob = Tmst.VecMultiply(a);
    /*
    //debug
    printf("=====================\n");
    Matrix3 Tmst_self(frames_mst.at(i-1).a, frames_mst.at(i-1).b, frames_mst.at(i-1).c);
    Vector aglob_self = Tmst_self.VecMultiply(a);
    Vector diff = aglob_self - aglob;
    global_diff += diff;
    printf("Unitvecs : a=[%.2f %.2f %.2f] [%.2f %.2f %.2f]\n",frames_mst.at(i-1).a.x,frames_mst.at(i-1).a.y,frames_mst.at(i-1).a.z,
	   gm.frames_mst.at(i-1).a.x,gm.frames_mst.at(i-1).a.y,gm.frames_mst.at(i-1).a.z);
    printf ("Dot product angles: %.2f %.2f %.2f\n",
	    RadtoDeg(acos(gm.frames_mst.at(i-1).a * frames_mst.at(i-1).a) ),
	    RadtoDeg(acos(gm.frames_mst.at(i-1).b * frames_mst.at(i-1).b) ),
	    RadtoDeg(acos(gm.frames_mst.at(i-1).c * frames_mst.at(i-1).c) ) );
    Vector d1 = gm.frames_mst.at(i-1).a - frames_mst.at(i-1).a;
    Vector d2 = gm.frames_mst.at(i-1).b - frames_mst.at(i-1).b;
    Vector d3 = gm.frames_mst.at(i-1).c - frames_mst.at(i-1).c;
    printf("Unitvec diff : a=[%.2f %.2f %.2f]\n",d1.x,d1.y,d1.z);
    printf("Unitvec diff : b=[%.2f %.2f %.2f]\n",d2.x,d2.y,d2.z);
    printf("Unitvec diff : c=[%.2f %.2f %.2f]\n",d3.x,d3.y,d3.z);
    printf("Unitvec L : [%.2f %.2f %.2f]\n",d1.Length(),d2.Length(),d3.Length() );
    printf("Diff. between wt & ideal: [%.2f %.2f %.2f]\n",diff.x,diff.y,diff.z);
    //printf("L_local = %f, L_global = %f\n",a.Length(),aglob.Length() );
    */
    frames.at(i).orig = frames.at(i-1).orig + aglob;
    frames_mst.at(i-1).orig = frames.at(i-1).orig;
  }
  //debug
  //printf("Global_diff: [%.2f %.2f %.2f]\n",global_diff.x,global_diff.y,global_diff.z);
}

void Geom::Transform(Geom::Geom &gm) {
  Transform(gm, 1, frames.size()-1);
}

// This is an "external" function which is actually plugged into
// GSL numerical derivative routines.
double GetFrame(double val, void *params) {
  double res = 0.0;
  
  Frame_params *prm = (Frame_params *) params;
  Geom *curGM = prm->GM;
  double old_val = gsl_matrix_get(curGM->m,prm->c1,prm->c2);
  gsl_matrix_set(curGM->m,prm->c1,prm->c2,val);
  
  // Set up the first frame:
  if ((curGM->frames).size() == 0) {
    std::cerr << "Frames not initially set in GetFrame!\nExiting ..\n";
    exit(EXIT_FAILURE);
  }
  coord frame_init = (curGM->frames).at(0); // get frame 0
  curGM->ReconstructGeom(frame_init, prm->ind); // this gets new xyz frames for [0 .. ind]
  
  if (prm->crd == 0) { // x
    res = (curGM->frames).at(prm->ind).orig.x;
  }else if (prm->crd == 1) { // y
    res = (curGM->frames).at(prm->ind).orig.y;
  }else if (prm->crd == 2) { // z
    res = (curGM->frames).at(prm->ind).orig.z;
  }else{
    std::cerr << "Invalid value of crd index: "<<prm->crd<<" in GetFrame!\nExiting ..\n";
    exit(EXIT_FAILURE);
  }
  
  gsl_matrix_set(curGM->m,prm->c1,prm->c2,old_val); // restore old (native) value in m
  return res;
}

// This function returns a set of differential displacements used in building
// the constraints in the global frame. NOTE: new memory gets allocated inside!!
Vector *Geom::disp(void) {
  Vector *diff = new Vector[frames.size()-1];
  for (unsigned int i=0; i<frames.size()-1; ++i) {
    diff[i] = frames.at(i+1).orig - frames.at(i).orig;
  }
  return diff;
}

// This function returns a set of global radius vectors (ie frame origins).
// NOTE: new memory gets allocated inside!!
Vector *Geom::global_disp(void) {
  Vector *gdiff = new Vector[frames.size()];
  for (unsigned int i=0; i<frames.size(); ++i) {
    gdiff[i] = frames.at(i).orig;
  }
  return gdiff;
}

// This aux. function returns the 3x3 T_i^{i+1} matrix given (Omega_i,r_i,t_i).
Matrix3 Geom::T(double Omega, double r, double t) {
  double sinom = sin(Omega);
  double cosom = cos(Omega);
  double sinom2 = sin(0.5*Omega);
  double cosom2 = cos(0.5*Omega);
  Matrix3 m;
  m.set(0,0,cosom);
  m.set(0,1,-sinom);
  m.set(0,2,r*cosom2 + t*sinom2);
  m.set(1,0,sinom);
  m.set(1,1,cosom);
  m.set(1,2,r*sinom2 - t*cosom2);
  m.set(2,0,-r*cosom2 + t*sinom2);
  m.set(2,1,r*sinom2 + t*cosom2);
  m.set(2,2,1.0);
  return m;
}

// This aux. function returns the 3x3 T_i^{MST} matrix given (Omega_i,r_i,t_i).
Matrix3 Geom::Tmst(double Omega, double r, double t) {
  double sinom2 = sin(0.5*Omega);
  double cosom2 = cos(0.5*Omega);
  Matrix3 m;
  m.set(0,0,cosom2);
  m.set(0,1,-sinom2);
  m.set(0,2,0.5*r*cosom2 + 0.5*t*sinom2);
  m.set(1,0,sinom2);
  m.set(1,1,cosom2);
  m.set(1,2,0.5*r*sinom2 - 0.5*t*cosom2);
  m.set(2,0,-0.5*r);
  m.set(2,1,0.5*t);
  m.set(2,2,1.0);
  return m;
}

// This aux. function returns a derivative of the 3x3 T_i^{i+1} matrix given
// (Omega_i,r_i,t_i). The derivative is taken wrt any of the 3 input variables.
Matrix3 Geom::T_deriv(double Omega, double r, double t, int i) {
  assert(i==0 || i==1 || i==2);
  double sinom = sin(Omega);
  double cosom = cos(Omega);
  double sinom2 = sin(0.5*Omega);
  double cosom2 = cos(0.5*Omega);
  Matrix3 m;
  if (i==0) { // twist
    m.set(0,0,-sinom);
    m.set(0,1,-cosom);
    m.set(0,2,-0.5*r*sinom2 + 0.5*t*cosom2);
    m.set(1,0,cosom);
    m.set(1,1,-sinom);
    m.set(1,2,0.5*r*cosom2 + 0.5*t*sinom2);
    m.set(2,0,0.5*r*sinom2 + 0.5*t*cosom2);
    m.set(2,1,0.5*r*cosom2 - 0.5*t*sinom2);
    m.set(2,2,0.0);
  }else if (i==1) { // roll
    m.set(0,0,0.0);
    m.set(0,1,0.0);
    m.set(0,2,cosom2);
    m.set(1,0,0.0);
    m.set(1,1,0.0);
    m.set(1,2,sinom2);
    m.set(2,0,-cosom2);
    m.set(2,1,sinom2);
    m.set(2,2,0.0);
  }else if (i==2) { // tilt
    m.set(0,0,0.0);
    m.set(0,1,0.0);
    m.set(0,2,sinom2);
    m.set(1,0,0.0);
    m.set(1,1,0.0);
    m.set(1,2,-cosom2);
    m.set(2,0,sinom2);
    m.set(2,1,cosom2);
    m.set(2,2,0.0);
  }
  // Switch to derivatives wrt to deg rather than rad:
  m.multiply(DegtoRad(1.0) );
  return m;
}

// This aux. function returns a derivative of the 3x3 T_i^{MST} matrix given
// (Omega_i,r_i,t_i). The derivative is taken wrt any of the 3 input variables.
Matrix3 Geom::Tmst_deriv(double Omega, double r, double t, int i) {
  assert(i==0 || i==1 || i==2);
  double sinom2 = sin(0.5*Omega);
  double cosom2 = cos(0.5*Omega);
  Matrix3 m;
  if (i==0) { // twist
    m.set(0,0,-0.5*sinom2);
    m.set(0,1,-0.5*cosom2);
    m.set(0,2,-0.25*r*sinom2 + 0.25*t*cosom2);
    m.set(1,0,0.5*cosom2);
    m.set(1,1,-0.5*sinom2);
    m.set(1,2,0.25*r*cosom2 + 0.25*t*sinom2);
    m.set(2,0,0.0);
    m.set(2,1,0.0);
    m.set(2,2,0.0);
  }else if (i==1) { // roll
    m.set(0,0,0.0);
    m.set(0,1,0.0);
    m.set(0,2,0.5*cosom2);
    m.set(1,0,0.0);
    m.set(1,1,0.0);
    m.set(1,2,0.5*sinom2);
    m.set(2,0,-0.5);
    m.set(2,1,0.0);
    m.set(2,2,0.0);
  }else if (i==2) { // tilt
    m.set(0,0,0.0);
    m.set(0,1,0.0);
    m.set(0,2,0.5*sinom2);
    m.set(1,0,0.0);
    m.set(1,1,0.0);
    m.set(1,2,-0.5*cosom2);
    m.set(2,0,0.0);
    m.set(2,1,0.5);
    m.set(2,2,0.0);
  }
  // Switch to derivatives wrt to deg rather than rad:
  m.multiply(DegtoRad(1.0) );
  return m;
}

