/* dna_parameters.c */

/*****************************************************************
        Copyright by Rockefeller University,
can not be reproduced or distributed without written permission of
the copyright holder.  Version 1.1 of April 2006.

Written by Alexandre Morozov (contact person: morozov@edsb.rockefeller.edu),
and Eric Siggia.

The program DNABEND implements an algorithm for computing elastic energies
for DNA molecules of arbitrary length and sequence. DNA can be constrained
to follow an arbitrary 3D curve, such as a left-handed superhelix found
in nucleosomal core particles.

The DNABEND method and its applications are described in the following
paper:

******************************************************************/

#include "dna_parameters.h"

DNAparams::DNAparams(void) {
  // Allocate memory:
  means_bs = new double*[DIM1];
  stddevs_bs = new double*[DIM1];
  for (int i=0; i<DIM1; i++) {
    means_bs[i] = new double[DIM2];
    stddevs_bs[i] = new double[DIM2];
  }
  Fij = new double**[DIM1];
  for (int i=0; i<DIM1; i++) {
    Fij[i] = new double*[DIM2];
    for (int j=0; j<DIM2; j++) {
      Fij[i][j] = new double[DIM2];
    }
  }
  // Fill it up with DNA params:
  char *chome = getenv("HOME"); // $HOME env var
  if (chome==NULL) {
    std::cerr << "Could not determine value of the $HOME environment variable!\n";
    exit(EXIT_FAILURE);
  }
  std::string home(chome);
  //DBpath = home + "/rosetta_database";
  //DBpath = home + "/homology_tf/tf_survey/Nucleosome/eminim/data";
  DBpath = "./";
  means_bs_file = DBpath + "/Equil_AM.mean.dat";
  stddevs_bs_file = DBpath + "/Equil_AM.stddev.dat";
  Fij_file = DBpath + "/Fij_AM.dat";
  //Fij_file = DBpath + "/Fij_AM.dat.mod";
  ReadMomentsTable(means_bs_file,means_bs,DIM1,DIM2); // means of DNA basestep effective params
  ReadMomentsTable(stddevs_bs_file,stddevs_bs,DIM1,DIM2); // stddevs of DNA basestep effective params
  ReadFijTable(Fij_file,Fij); // Fij (effective basestep force constants) for DNA potential
}

DNAparams::~DNAparams() {
  for (int i=0; i<DIM1; i++) {
    delete [] means_bs[i];
    delete [] stddevs_bs[i];
  }
  delete [] means_bs;
  delete [] stddevs_bs;
  for (int j=0; j<DIM2; j++) {
    for (int i=0; i<DIM1; i++) {
      delete [] Fij[i][j];
    }
    delete [] Fij[j];
  }
  delete [] Fij;
}

void DNAparams::ReadMomentsTable (std::string const & filename, double **table,
				  int const dim1, int const dim2) {
  using std::cout;
  using std::cerr;
  using std::endl;
  using std::ifstream;
  
  ifstream file(filename.c_str() );
  
  char buffer[512];
  char basestep_name[3];
  int datapoints;
  
  if (file.fail()) {
    cerr << "Could not open file: "<<filename<<" for reading!\nExiting..\n";
    exit(EXIT_FAILURE);
  }
  
  int cnt = -1; // basestep counter
  while (!file.eof()) {
    file.getline(buffer,512,'\n');
    if (file.fail() && !file.eof() ) file.clear();
    if (cnt==-1) {
      ++cnt;
      continue; // skip header
    }
    if (sscanf(buffer,"%s %d %lf %lf %lf %lf %lf %lf",basestep_name,&datapoints,
	       &table[cnt][0],&table[cnt][1],&table[cnt][2],&table[cnt][3],&table[cnt][4],&table[cnt][5]) == 8) {
      //debug
      //printf("%4d\n",cnt);
      ++cnt;
    }
  }
  file.close();
  if (cnt != dim1) {
    cerr << "Read "<<cnt<< " lines instead of "<<dim1<<" in file "<<filename<<"!\n";
    exit(EXIT_FAILURE);
  }
  /*
  //debug
  for (int i=0; i<dim1; ++i) {
    for (int j=0; j<dim2; ++j) {
      printf("%6.3f ",table[i][j]);
      //cout << table[i][j] << "  ";
    }
    printf("\n");
    //cout << endl;
  }
  */
}

void DNAparams::ReadFijTable (std::string const & filename, double ***table) {
  using std::cout;
  using std::cerr;
  using std::endl;
  using std::ifstream;
  
  ifstream file(filename.c_str() );
  
  char buffer[512];
  char label[24];
  
  if (file.fail()) {
    cerr << "Could not open file: "<<filename<<" for reading!\nExiting..\n";
    exit(EXIT_FAILURE);
  }
  
  int cnt = -1; // basestep counter
  int prm_cnt = -1; // parameter counter
  while (!file.eof()) {
    file.getline(buffer,512,'\n');
    if (file.fail() && !file.eof() ) file.clear();
    if (strlen(buffer) == 2) {
      prm_cnt = 0;
      ++cnt;
    }else if (sscanf(buffer,"%s %lf %lf %lf %lf %lf %lf",label,
		     &table[cnt][0][prm_cnt],&table[cnt][1][prm_cnt],&table[cnt][2][prm_cnt],
		     &table[cnt][3][prm_cnt],&table[cnt][4][prm_cnt],&table[cnt][5][prm_cnt]) == 7) {
      ++prm_cnt;
    }
  }
  file.close();
  if (cnt != DIM1-1) {
    cerr << "Read "<<cnt<< " lines instead of "<<DIM1-1<<" in file "<<filename<<"!\n";
    exit(EXIT_FAILURE);
  }
  /*
  //debug
  int chosen_bstep = 1;
  for (int i=0; i<6; ++i) {
  for (int j=0; j<6; ++j) {
  cout << table[chosen_bstep][i][j] << "  ";
  }
  cout << endl;
  }
  */
}

